#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include "node.h"
#include "2DArray.h"
#include "Set.h"
#include "trans_utils.h"
#include "cluster.h"

extern int count1and2links(char *inputFile, TwoDArrayColln& countColl);




TwoDArrayColln::TwoDArrayColln(int numTuples, int nA, int *domSizes)
{
  _numAttrs = nA;
  int i, j;
  int *startEndVals = new int[_numAttrs+1];
  int sum = 0;
  int threshold = 1;

  _reductionFactors = new int[_numAttrs];
  for (i = 0; i < _numAttrs; i++)
    {
      startEndVals[i] = sum;
      sum += domSizes[i];
      if (domSizes[i] > MAX_DIM_SIZE)
	_reductionFactors[i] = (int)(domSizes[i]/MAX_DIM_SIZE)+1;
      else
	_reductionFactors[i] = 1;
    }
  startEndVals[_numAttrs] = sum;
  

#ifdef DEBUG
  cout << "TwoDArrayColln:StartEndVals ::: ";
  for (i = 0; i < _numAttrs+1; i++)
    cout << startEndVals[i] << " " ;
  cout << endl << endl;
#endif

  _indexToAttrPairs = new (TwoDArray*)[_numAttrs * (_numAttrs-1)/2];
  for (i = 0; i < _numAttrs; i++)
    for (j = i+1; j < _numAttrs; j++)
      {
	_indexToAttrPairs[_index(i, j)] = new TwoDArray(domSizes[i], domSizes[j], startEndVals[i], startEndVals[i+1]-1, startEndVals[j], startEndVals[j+1]-1);
	threshold = ((int)(ALPHA * (double)(numTuples/(domSizes[i]*domSizes[j]))) > MIN_THRESHOLD) ? (int)(ALPHA * (double)(numTuples/(domSizes[i]*domSizes[j]))) : MIN_THRESHOLD;
	_indexToAttrPairs[_index(i,j)]->SetThreshold(threshold);
	cout << "Threshold between " << i << " " << j << " is " << threshold << endl;
      }

  _attrValCounts = new int[sum];
  memset((void*)_attrValCounts, 0, sizeof(int)*sum);

  delete [] startEndVals;
}

// assuming here that the tuple has all k-fields;
// also that the entries in the first fields are
// smaller than the second and so on...
//void TwoDArrayColln::IncrementCounts(Set& tuple)
//{
//}


//distSet x "result" is a cluster on "attr x thruAttr"
//returns the size of the restrictingSet size;
//values in restrictingSet will be in the sorted order at the end;
int 
TwoDArrayColln::FindRestrictingSet(Set& distSet, int attr, int thruAttr, 
				   int *restrictingSet)
{
  int i,j;
  int beginThruAttr;
  int endThruAttr;
  int threshold;
  int size=0;
  TwoDArray *temp;
  bool flag;

  assert(attr != thruAttr);

  if (attr < thruAttr)
    {
      temp = _indexToAttrPairs[_index(attr, thruAttr)];
      beginThruAttr = temp->Begin2();
      endThruAttr = temp->End2();
    }
  else
    {
      temp = _indexToAttrPairs[_index(thruAttr, attr)];
      beginThruAttr = temp->Begin1();
      endThruAttr = temp->End1();
    }

  threshold = temp->Threshold();
  for (i = beginThruAttr; i <= endThruAttr; i++)
    {
      flag = TRUE;
      for (j = 0; j < distSet.size(); j++)
	{
	  if (attr < thruAttr)
	    {
	      if (temp->get(distSet.elements(j), i) < threshold)
		{
		  flag = FALSE;
		  break;
		}
	    }
	  else
	    {
	      if (temp->get(i, distSet.elements(j)) < threshold)
		{
		  flag = FALSE;
		  break;
		}
	    }
	}
      if (TRUE == flag)
	{
	  restrictingSet[size] = i; //will automatically be in the sorted order;
	  size++;
	  assert(size < MAX_CLUSTERSIZE);
	}
    }

#if 0
  if (size > 1)
    {
      cout <<"Distinguishing set is " << distSet << " and restSetSz = " << size << endl;
      for (i = 0; i < size; i++)
	cout << restrictingSet[i] << " ";
      cout << endl;
    }
#endif

  temp = NULL;
  return size;
}



/***
    Self-join on the reln (attr1, attr2); 
    then group by each pair of values in the same column;
    first=true if similarity between values of attr1, false if attr2;
    restrictingSet: sorted set of values on the attr for which similarity is being 
    computed 
    numRestrictingVals: size of the restrictingSet;
***/
void TwoDArrayColln::Join(int attr1, int attr2, bool First, SymTwoDArray& counts,
			  int *restrictingSet, int numRestrictingVals)
{
  int begin = counts.Begin();
  int end = counts.End();
  int i, j, k;
  int indx;
  int numStrongLinks=0;

  assert(attr1 < attr2);
  TwoDArray& temp = *_indexToAttrPairs[_index(attr1, attr2)];
  cout << "Inside Join: the threshold is " << temp.Threshold() << endl;
  if (NULL == restrictingSet) //all values should be used;
    {
      if (TRUE == First)
	// attr1 corresponds to the counts array;
	// or, similarity for values on attr1 using attr2;
	{
	  for (k = temp.Begin2(); k <= temp.End2(); k++)
	    for (i = begin; i <= end; i++)
	      {
		numStrongLinks=0; //#strong(i,*) links;
		if (temp.get(i,k) >= temp.Threshold())
		  {
		    numStrongLinks++;
		    for (j = i+1; j <= end; j++)
		      {
			if (temp.get(j,k) >= temp.Threshold())
			  {
			  //counts.increment(i,j,attr2, temp.get(i,k)*temp.get(j,k));
			    counts.increment(i, j, attr2, 1);
			    //cout << i << " " << j << endl;
			  }
		      }
		  }
		//counts.set(i, i, attr2, _attrValCounts[i]*_attrValCounts[i]);
		//counts.set(i, i, attr2, numStrongLinks);
	      }
	}
      else
	// attr2 corresponds to the counts array;
	// or, similarity for values on attr2 using attr1;
	{
	  for (k = temp.Begin1(); k <= temp.End1(); k++)
	    for (i = begin; i <= end; i++)
	      {
		numStrongLinks=0;
		if (temp.get(k,i) >= temp.Threshold())
		  {
		    numStrongLinks++;
		    for (j = i+1; j <= end; j++)
		      {
			if (temp.get(k,j) >= temp.Threshold())
			  //counts.increment(i,j,attr1, temp.get(k,i)*temp.get(k,j));
			  counts.increment(i, j, attr1, 1);
		      }
		  }
		//counts.set(i, i, attr1, _attrValCounts[i]*_attrValCounts[i]);
		//counts.set(i, i, attr1, numStrongLinks);
	      }
	}
    }
  else //restrictingSet is active.
    {
      indx=0;
      if (TRUE == First)
	// attr1 corresponds to the counts array;
	// or, similarity for values on attr1 using restrictingSet vals on attr2;
	{
	  for (k = temp.Begin2(); k <= temp.End2(); k++)
	    {
	      assert(indx < numRestrictingVals);
	      if (restrictingSet[indx] != k)
		continue;
	      for (i = begin; i <= end; i++)
		{
		  //numStrongLinks=0;
		  if (temp.get(i,k) >= temp.Threshold())
		    {
		      numStrongLinks++;
		      for (j = i+1; j <= end; j++)
			{
			  if (temp.get(j,k) >= temp.Threshold())
			    counts.increment(i, j, attr2, 1);
			  //counts.increment(i, j, attr2, temp.get(i,k)*temp.get(j,k));
			}
		      //counts.set(i, i, attr2, numStrongLinks);
		      //counts.set(i, i, attr2, _attrValCounts[i]*_attrValCounts[i]);
		    }
		}
	      indx++;
	      if (numRestrictingVals == indx)
		break;
	    }
	}
      else
	// attr2 corresponds to the counts array;
	// or, similarity for values on attr2 using attr1;
	{
	  indx = 0;
	  for (k = temp.Begin1(); k <= temp.End1(); k++)
	    {
	      assert(indx < numRestrictingVals);

	      if (restrictingSet[indx] != k)
		continue;
	      for (i = begin; i <= end; i++)
		{
		  //numStrongLinks=0;
		  if (temp.get(k,i) >= temp.Threshold())
		    {
		      numStrongLinks++;
		      for (j = i+1; j <= end; j++)
			{
			  if (temp.get(k,j) >= temp.Threshold())
			    counts.increment(i, j, attr1, 1);
			    //counts.increment(i, j, attr1, temp.get(k,i)*temp.get(k,j));
			}
		    }
		  //counts.set(i, i, attr1, numStrongLinks);
		  //counts.set(i, i, attr1, _attrValCounts[i]*_attrValCounts[i]);
		}
	      indx++;
	      if (indx == numRestrictingVals)
		break;
	    }
	}
    }
  /***
  ofstream ofile("/tmp/link-strengths");
  ofile << "TwoDArray Info between " << attr1 << " and " << attr2 << endl;
  ofile << temp << endl;
  ofile << "After join of " << attr1 << " , " << attr2 << endl;
  ofile << counts; 
  ofile.close();
  ****/
}




/*

 */


SummarizedInfo::SummarizedInfo(int numTuples, int nA, int *domSz)
{
  int i=0;
  int totNumAttrVals=0;

  _numTuples = numTuples;
  _numAttrs = nA;
  _domSizes = new int[_numAttrs];
  _startAttrVals = new int[_numAttrs];
  assert((_domSizes != NULL) && (_startAttrVals != NULL));

#ifdef DEBUG  
  cout << "Sum Info: startAttrVals ";
#endif
  for (i = 0; i < _numAttrs; i++)
    {
      _domSizes[i] = domSz[i];
      _startAttrVals[i] = totNumAttrVals;
      totNumAttrVals += _domSizes[i];
#ifdef DEBUG      
      cout  << _startAttrVals[i] << " " << _domSizes[i] << " : ";
#endif
    }
#ifdef DEBUG
  cout << endl;
#endif
  //_info = new (TwoDArray*)[_numAttrs*(_numAttrs-1)/2];
  _info = new TwoDArrayColln(_numTuples,_numAttrs,_domSizes);
  _attrInfo = new SymTwoDArrayColln(_numAttrs);
}

SummarizedInfo::~SummarizedInfo()
{
  delete [] _domSizes;
  delete [] _startAttrVals;
  
  delete _info;
  delete _attrInfo;
}

/****
void 
SummarizedInfo::Summarize(char *inputFile)
{
// read tuple by tuple and call IncrementCounts;
  int xNumber=0;
  Set xactIter(0);
  char buffer[TRANS_BLOCK_SIZE];

  SetupXactScan(inputFile, buffer);
  while (getNextXact(xactIter, xNumber, buffer) != -1)
    {
      _info->IncrementCounts(xactIter);
    }
  xactIter.Nullify();
}
****/

double
SummarizedInfo::Threshold(int attr, int thruAttr)
{
  int i;
  double threshold = 0.0;
  //double prob = ((double)(_numTuples)/_domSizes[attr]) * ((double)(_numTuples)/_domSizes[attr]);
  double prob = ((double)(_numTuples)/_domSizes[attr]);

  if (-1 == thruAttr)
    {
      //threshold = prob * _numAttrs;
      for (i = 0; i < _numAttrs; i++)
	if (i != attr)
	  threshold += prob/_domSizes[i];
    }
  else
    {
      //threshold = prob;
      threshold = prob/_domSizes[thruAttr];
    }
  if (threshold > 1.0)
    return (ALPHA * threshold);
  else
    return 1.5; // a hack to avoid zero thresholds;
}


void
SummarizedInfo::Summarize(char *inputFile)
{
  count1and2links(inputFile, *_info);
}

/***
void
SummarizedInfo::InitAttrInfo(int attr)
{
  int sum=0;
  int i;
  for (i = 0; i < attr-1; i++)
    {
      sum += _domSizes[i];
    }
  _attrInfo->InitAttrInfo(_domSizes[attr], sum, sum+_domSizes[attr]-1, attr, _numAttrs);
}
***/


void 
SummarizedInfo::ComputeAttrInfo(int attr, int thruAttr)
{
  int sum=0;
  int i;
  for (i = 0; i < attr; i++)
    sum += _domSizes[i];

  cout << "Computing Attr Info " << attr << " " << thruAttr << endl;
  if (NULL == _attrInfo->GetAttrInfo(attr))
    {
      _attrInfo->InitAttrInfo(_domSizes[attr], sum, sum+_domSizes[attr]-1, 
			      attr, _numAttrs);
#if 1
      cout << "Created SymTwoDarray^^^^^^ " << sum << " " 
	   << sum+_domSizes[attr]-1 << endl;
#endif
    }
  if (-1 == thruAttr)
    {
      for (i = 0; i < attr; i++)
	_info->Join(i, attr, FALSE, *(_attrInfo->GetAttrInfo(attr)));
      for (i = attr+1; i < _numAttrs; i++)
	_info->Join(attr, i, TRUE, *(_attrInfo->GetAttrInfo(attr)));
    }
  else
    {
      assert(thruAttr != attr);
      if (thruAttr < attr)
	_info->Join(i, attr, FALSE, *(_attrInfo->GetAttrInfo(attr)));
      else
	_info->Join(attr, i, TRUE, *(_attrInfo->GetAttrInfo(attr)));
    }
  cout << "Cactus2: Joined Attributes " << attr << " " << thruAttr << endl;
}

void
SummarizedInfo::CloseAttrInfo(int attr)
{
  // for now, do nothing; 
  // but, we can delete the info for attr;
  _attrInfo->CloseAttrInfo(attr);
}


//both set1 and set2 consist of attr values of the same ``attr'';
bool
SummarizedInfo::ValidateCluster(int *set1, int size1, int* set2, int size2,
				int element1, int element2, int attr, int thruAttr)
{
  SymTwoDArray& attrInfo = *(AttrInfo(attr));
  //double threshold = Threshold(attr, thruAttr);
  double threshold = 0;
  int i;
  int strength = 0;

  for (i = 0; i < size2; i++)
    {
      assert(element1 != set2[i]);
      strength += attrInfo.get(element1, set2[i], thruAttr);
    }
  if ((double)strength/((double)size2) <= threshold)
    return FALSE;

  strength = 0;
  for (i = 0; i < size1; i++)
    {
      assert(element2 != set1[i]);
      strength += attrInfo.get(element2, set1[i], thruAttr);
    }
  if ((double)strength/((double)size1) <= threshold)
    return FALSE;
  return TRUE;
}




//return the size of the cluster, -1 if no cluster was formed.
int
SummarizedInfo::FindSiblingSet(int attr, int thruAttr, Set& distSet, 
			       int *restrictingSet)
{
#ifdef DEBUG
  //cout << "SummarizedInfo::SetUpExtendDistSet" << endl;
#endif

  SymTwoDArray& attrInfo = *(AttrInfo(attr));
  // a hack to speed things up
  if (attrInfo.get(distSet.elements(0), distSet.elements(1), thruAttr) < 1)
    return -1;

  //TwoDArray& attrPairInfo = *(_info->AttrPairInfo(attr, thruAttr));

  int restSetSz=0;
  bool First=TRUE;

  assert(attr != thruAttr);
  if (attr > thruAttr)
    First = FALSE;

  //find the restricting set;
  restSetSz = _info->FindRestrictingSet(distSet,attr,thruAttr,restrictingSet);
  assert(restSetSz < MAX_CLUSTERSIZE);

  if (restSetSz <= 1) //there can be no clique with this distSet;
    {
      return -1;
    }

#if 0
  int i;
  //cout << "In SumInfo:FindSiblingSet :: " << distSet << ":";
  cout << "The distinguishing set is ";
  for (i = 0; i < distSet.size(); i++)
    cout << distSet.elements(i) << " ";
  cout << endl;
  cout << "Connection Count is " << attrInfo.get(distSet.elements(0), distSet.elements(1), thruAttr) << endl;
  cout << "The sibling  set is ";
  for (i = 0; i < restSetSz; i++)
    cout << restrictingSet[i] << " ";
  cout << endl;
#endif
  
  /***
  //clear attrInfo and join again with restrictingSet;
  attrInfo.ClearCounts(thruAttr);
  if (attr < thruAttr)
    _info->Join(attr, thruAttr, First, attrInfo, restrictingSet, restSetSz);
  else
    _info->Join(thruAttr, attr, First, attrInfo, restrictingSet, restSetSz);
  ***/

  return restSetSz;
}


/****
int
SummarizedInfo::CliqueStrength(int *set, int size, int attr, int thruAttr)
{
  int i,j;
  TwoDArray& temp = *(_info->AttrPairInfo(attr, thruAttr));
  int beginThru, endThru;
  int strength = 0;
  bool flag=TRUE;
  
  if (attr < thruAttr)
    {
      beginThru = temp.Begin2();
      endThru = temp.End2();
    }
  else
    {
      beginThru = temp.Begin1();
      endThru = temp.End1();
    }
  
  for (i = beginThru; i <= endThru; i++)
    {
      flag = TRUE;
      for (j = 0; j < size; j++)
	{
	  if (temp.get(i, set[j]) < temp.Threshold())
	    {
	      flag = FALSE;
	      break;
	    }
	}
      if (TRUE == flag)
	strength++;
    }
#ifdef DEBUG
  cout << "Clique Strength " << strength << endl;
#endif  
  return strength;
}
**********/

// output operators;
ostream& 
operator<<(ostream& of, TwoDArray& obj)
{
  int i, j;

  of << "Dimensions of the 2-D array : " << obj._dimn1 << " " << obj._dimn2 << endl;
  of << "Attribute value ranges (dimn1 then dimn2) " << obj._begin1 << " " 
     << obj._end1 << " :: " << obj._begin2 << " " << obj._end2 << endl;
  of << " Minimum Threshold " << obj._threshold << endl;

  for (i = obj._begin1; i <= obj._end1; i++)
    for (j = obj._begin2; j <= obj._end2; j++)
      of << "<" << i << " " << j << "> : " << obj.get(i,j) << endl;

  return of;
}

ostream& 
operator<<(ostream& of, SymTwoDArray& obj)
{
  of << "Dimension " << obj._dimn << " " << obj._begin << "  " << obj._end << endl;
  int i,j,k;

  for (i = obj._begin; i <= obj._end; i++)
    for (j = i; j <= obj._end; j++)
      {
	of << i << " " << j << " : "; 
	for (k = 0; k < obj._numAttrs; k++)
	  of << obj.get(i,j,k) << " ";
	of << endl;
      }

  of << endl;
  return of;
}


ostream& 
operator<<(ostream& of, TwoDArrayColln& obj)
{
  int i, j;
  for (i = 0; i < obj._numAttrs; i++)
    for (j = i+1; j < obj._numAttrs; j++)
      {
	of << "TwoDArray for attribute pair : " << "("<<i<<","<<j<<")"<< endl;
	of << *(obj._indexToAttrPairs[obj._index(i,j)]) << endl;
      }
  return of;
}

ostream& 
operator<<(ostream& of, SymTwoDArrayColln& obj)
{
  int i;
  for (i = 0; i < obj._numAttrs; i++)
    {
      of << "SymTwoDArray for attribute " << i << endl;
      of << *(obj._attrInfoArr[i]) << endl << endl;
    }
  return of;
}

ostream& 
operator<<(ostream& of, SummarizedInfo& obj)
{
  of << *(obj._info) << endl;
  of << *(obj._attrInfo) << endl;
  return of;
}
