#ifndef __TWODARR_H__
#define __TWODARR_H__

#include <assert.h>
#include <string.h>

class TwoDArray 
{
private:
  int _dimn1;
  int _dimn2;
  int _begin1;
  int _begin2;
  int _end1;
  int _end2;
  int *_array;

  int _threshold;

  int _index(int i, int j)
    {
      return ((i-_begin1)*_dimn2 + (j-_begin2));
    }
public:
  TwoDArray(int d1, int d2, int b1, int e1, int b2, int e2) 
    {
      _dimn1 = d1;
      _dimn2 = d2;
      _begin1 = b1;
      _begin2 = b2;
      _end1 = e1;
      _end2 = e2;
      _array = new int[_dimn1 * _dimn2];
#ifdef DEBUG
      cout << "TwoDArray::TwoDArray " << endl;
      cout << _dimn1 << " " << _dimn2 << " " << _begin1 << " " << _end1 
	   << " " << _begin2 << " " << _end2 << endl;
#endif
      memset((void*)_array, 0, _dimn1*_dimn2*sizeof(int));
      assert(_array != NULL);
      
      _threshold = 0;
    }

  ~TwoDArray()
    {
      delete [] _array;
    }

  int Begin1() { return _begin1;}
  int Begin2() { return _begin2;}
  int End1() { return _end1;}
  int End2() { return _end2;}

  int Threshold() { return _threshold;}
  void SetThreshold(int val) 
    {
#ifdef DEBUG
      cout << "Setting Threshold : " << val << " for " << _begin1 << " " 
	   << _end1 << " " << _begin2 << " " << _end2 << endl;
#endif
      _threshold = val;
    }

  int get(int i, int j)
    {
      //cout << i << " " << _begin1 << " " << _dimn1 << " ";
      assert((i-_begin1) < _dimn1);
      assert((j-_begin2) < _dimn2);
      return _array[_index(i,j)];
    }
  void set(int i, int j, int val)
    {
      _array[_index(i,j)] = val;
    }
  void increment(int i, int j)
    {
      //cout << i << " " << _begin1 << " " << _dimn1 << " " << endl;
      assert((i-_begin1) < _dimn1);
      //cout << j << " " << _begin2 << " " << _dimn2 << " " << endl;
      assert((j-_begin2) < _dimn2);
      _array[_index(i,j)]++;
    }

  friend ostream& operator<<(ostream& of, TwoDArray& obj);
};

// an array for storing number of links between
// all pairs (a1, a2) of the same attribute.
class SymTwoDArray
{
private:
  int _numAttrs;
  int _attr; // attribute for which this instance summarizes info;
  int _dimn;
  int _begin;
  int _end;
  int *_array;

  int _index(int i, int j) 
    {
      assert(i <= j);
      return ((i-_begin)*_dimn  - (i-_begin)*((i-_begin)-1)/2 + j-i);
    }

  int _indexC(int i, int j, int thruAttr)
    {
      assert(i <= j);
      assert(thruAttr < _numAttrs);
      return (thruAttr*_sizePerAttr() + _index(i,j));
    }

  int _sizePerAttr() { return _dimn*(_dimn+1)/2; }
  int _size() { return _numAttrs*_dimn*(_dimn+1)/2; }

public:
  SymTwoDArray(int d, int b, int e, int attr, int nA) 
    {
      _dimn = d;
      _begin = b;
      _end = e;
      _attr = attr;
      _numAttrs = nA;
      _array = new int[_size()];
      assert(_array != NULL);
      memset((void*)_array, 0, _size()*sizeof(int));
    }

  ~SymTwoDArray()
    {
      delete [] _array;
    }

  void ClearCounts(int thruAttr)
    {      
      memset((void*)(_array+_indexC(_begin, _begin, thruAttr)), 0, _sizePerAttr()*sizeof(int));
    }

  int Begin() {return _begin;}
  int End() {return _end;}
  int Dimn() {return _dimn;}

  //double Threshold() { return _threshold;}

  // i < j;
  inline int get_fast(int i, int j, int thruAttr)
    {
      assert(i <= j);
      assert(j-_begin < _dimn);
      if (-1 == thruAttr)
	return (_array[_indexC(i,j,_attr)]);
      else
	return (_array[_indexC(i,j,thruAttr)]);
    }

  // not necessary that i < j;
  inline int get(int i, int j, int thruAttr)
    {
      //assert (i != j);
      return ((i <= j) ? get_fast(i,j,thruAttr) : get_fast(j,i,thruAttr));
    }

  void set(int i, int j, int thruAttr, int val)
    {
      assert(i <= j);
      assert(j-_begin < _dimn);
      _array[_indexC(i,j,thruAttr)] = val;
      _array[_indexC(i,j,_attr)] = val; // aggregated component;
    }

  void increment(int i, int j, int thruAttr, int ct=1)
    {
      assert(i < j);
      _array[_indexC(i,j,thruAttr)] += ct;
      _array[_indexC(i,j,_attr)] += ct; // aggregated component;
    }
  
  friend ostream& operator<<(ostream& of, SymTwoDArray& obj);
};

class TwoDArrayColln
{
private:
  int _numAttrs;
  TwoDArray  **_indexToAttrPairs;
  int *_attrValCounts;

  //the following flags are necessary for 
  //reducing the domain size;
  bool _reducedDomain;
  int *_reductionFactors;

  int _index(int i, int j)
    {
      assert(i < j);
      assert(j < _numAttrs);
      return (i*(_numAttrs-1) - (i-1)*i/2 + j-i-1);
    }

  int _map(int attrVal, int attr)
    {
      return (int)(attrVal/_reductionFactors[attr]);
    }

  int _min(int val1, int val2) {return (val1 < val2) ? val1 : val2; }

public:
  TwoDArrayColln(int nT, int nA, int *domSizes);
  ~TwoDArrayColln()
    {
      int i,j;
      for (i = 0; i < _numAttrs; i++)
	for (j = i+1; j < _numAttrs; j++)
	  delete _indexToAttrPairs[_index(i,j)];

      delete [] _indexToAttrPairs;
      delete [] _attrValCounts;
      delete [] _reductionFactors;
    }

  TwoDArray* AttrPairInfo(int attr1, int attr2)
    {
      if (attr1 < attr2)
	return _indexToAttrPairs[_index(attr1, attr2)];
      else
	return _indexToAttrPairs[_index(attr2, attr1)];
    }

  int Count(int attr1, int val1, int attr2, int val2)
    {
      return (_indexToAttrPairs[_index(attr1,attr2)]->get(val1, val2));
    }

  int Threshold(int attr1, int attr2)
    {
      return _indexToAttrPairs[_index(attr1,attr2)]->Threshold();
    }

  void IncrementCounts(Set& tuple)
    {
      //int sz = tuple.size();
      int sz = _numAttrs;
      int i,j;
      
      for (i = 0; i < sz; i++)
	{
	  //_attrValCounts[tuple.elements(i)]++;
	  for (j = i+1; j < sz; j++)
	    _indexToAttrPairs[_index(i,j)]->increment(tuple.elements(i), 
						      tuple.elements(j));
	    //_indexToAttrPairs[_index(i,j)]->increment(_map(tuple.elements(i), i), 
	    //				      _map(tuple.elements(j), j));
	}
    }

  void ExpandedCounts();

  void Join(int attr1, int attr2, bool First, SymTwoDArray& counts,
	    int *restrictingSet=NULL, int numRestrictingVals=-1);

  int FindRestrictingSet(Set& distSet, int attr, int thruAttr, 
			 int *restrictingSet);

  friend ostream& operator<<(ostream& of, TwoDArrayColln& obj);
};


class SymTwoDArrayColln
{
private:
  int _numAttrs;
  SymTwoDArray **_attrInfoArr;

public:
  SymTwoDArrayColln(int nA)
    {
      _numAttrs = nA;
      _attrInfoArr = new (SymTwoDArray*)[_numAttrs];
      memset((void*)_attrInfoArr, 0, _numAttrs*sizeof(SymTwoDArray*));
    }

  ~SymTwoDArrayColln()
    {
      int i;
      for (i = 0; i < _numAttrs; i++)
	if (NULL != _attrInfoArr[i]) delete _attrInfoArr[i];
      delete [] _attrInfoArr;
    }

  /***
  void DeleteSymTwoDArray(int attr, int thruAttr)
    {
      if (NULL != _attrInfoArr[i])
	_attrInfoArr[i]->DeleteSymTwoDArray(thruAttr);
    }
  ***/

  void InitAttrInfo(int domSize, int b, int e, int attr, int numAttrs)
    {
      assert(NULL == _attrInfoArr[attr]);
      _attrInfoArr[attr] = new SymTwoDArray(domSize, b, e, attr, numAttrs);
    }

  SymTwoDArray* GetAttrInfo(int attr) { return _attrInfoArr[attr];}

  void CloseAttrInfo(int attr)
    {
      assert(NULL != _attrInfoArr[attr]);
      if (NULL != _attrInfoArr[attr]) 
	{
	  delete _attrInfoArr[attr]; 
	  _attrInfoArr[attr] = NULL;
	}
    }

  /***
  int get(int i, int j, int attr)
    {
      assert(_attrInfoArr[attr] != NULL);
      return (_attrInfoArr[attr].get(i,j));
    }

  void set(int i, int j, int thruAttr, int attr, int val)
    {
      assert(_attrInfoArr[attr] != NULL);
      return (_attrInfoArr[attr].set(i,j,thruAttr,val));
    }
  ***/

  friend ostream& operator<<(ostream& of, SymTwoDArrayColln& obj);
};


// class for maintaining pairwise counts;
// contains all sufficient clustering information.
class SummarizedInfo
{
private: 
  int _numTuples;
  int _numAttrs;
  int *_domSizes;
  int *_startAttrVals;
  TwoDArrayColln *_info;
  SymTwoDArrayColln *_attrInfo;
 
public:
  SummarizedInfo(int numTuples, int nA, int *domSz);
  ~SummarizedInfo();

  int StartAttrVal(int attr) { return _startAttrVals[attr];}
  int DomSize(int attr) { return _domSizes[attr];}
  SymTwoDArray* AttrInfo(int attr) { return _attrInfo->GetAttrInfo(attr);}

  TwoDArrayColln* TwoDArrayCollnInfo() { return _info;}
  TwoDArray* TwoDArrayInfo(int attr, int thruAttr)
    {
      return _info->AttrPairInfo(attr, thruAttr);
    }

  double Threshold(int attr, int thruAttr=-1);

  //void SummarizePair(char *inputFile, int attr1, int attr2);
  void Summarize(char *inputFile);
  void ComputeAttrInfo(int attr, int thruAttr=-1);
  //void InitAttrInfo(int attr);
  void CloseAttrInfo(int attr);


  bool ValidateCluster(int *set1, int size1, int *set2, int size2, 
		       int element1, int element2, int attr, int thruAttr);


  //int CliqueStrength(int *set, int size, int attr, int thruAttr);

  int FindSiblingSet(int attr, int thruAttr, Set& distSet, int *siblingSet);


  
  void WriteTwoDArray(int attr1, int attr2);
  void ReadTwoDArray(int attr1, int attr2);

  friend ostream& operator<<(ostream& of, SummarizedInfo& obj);
};


#endif






