#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include "node.h"
#include "2DArray.h"
#include "Set.h"
#include "trans_utils.h"
#include "cluster.h"
#include "timeutil.h"

void OutputSortedAttrVals(int *sortedAttrVals, int *breaks, Graph& graph, int attr, 
			  int basin);



void Multiply(SymTwoDArray& attrInfo, Graph& graph1, Graph& graph2, 
	      int attr)
{
  int dimn = attrInfo.Dimn();
  int basins = graph1.NumBasins(attr);
  assert(dimn == graph1.DomainSz(attr));

  int i, j, b;
  double val = 0.0;
  int begin = attrInfo.Begin();
  int end = attrInfo.End();

  for (b = 0; b < basins; b++)
    {
      for (i = begin; i <= end; i++)
	{			
	  val = 0.0;
	  for (j = begin; j < i; j++)
	  //for (j = begin; j <= i; j++)
	    {
	      val += ((double)(attrInfo.get_fast(j,i,-1)))*graph1.GetWt(j,attr,b);
	    }
	  for (j = i+1; j <= end; j++)
	    {
	      val += ((double)(attrInfo.get_fast(i,j,-1)))*graph1.GetWt(j,attr,b);
	    }
	  graph2.AddWtToNode(i, attr, b, val);
	}
    }
}


int count1and2links(char *inputFile,
		    TwoDArrayColln& countColln
#ifdef ASCII
		    ,int numItems, int minSupport
#endif
		    )
{
  int large2Itemsets = 0;
  

#ifdef ASCII
    // Potentially very wasteful but we only need one buffer and set at a time.
    int strSize = MAX_TRANS_BYTES(numItems);
    char buf[strSize];
    
    xfile.getline(buf, strSize);

    while (xfile) {
      Set transaction(buf);
      xfile.getline(buf, strSize);
#else
      int xNumber = 0;
      char buffer[TRANS_BLOCK_SIZE];
      Set transaction(0);

      SetupXactScan(inputFile, buffer);
      while (getNextXact(transaction, xNumber, buffer) != -1) 
	{
#endif  //SAMPLING
	  countColln.IncrementCounts(transaction);
	}
      transaction.Nullify();
      return large2Itemsets;
}


TwoDArrayColln* PreProcess(char *inputFile, int numTuples, int numAttrs, 
			   int *domSizes)
{
  TwoDArrayColln* countColln = new TwoDArrayColln(numTuples, numAttrs, domSizes);
  count1and2links(inputFile, *countColln);
  return countColln;
}



void ClusterAttrCactus1(SymTwoDArray& attrInfo, Graph& graph1, Graph& graph2, 
			int attr, int numIterns)
{
  int i;
  for (i = 0; i < numIterns; i++)
    {
      Multiply(attrInfo, graph1, graph2, attr);
      graph2.OrthoNormalizeAttr(attr);
      graph1.AssignAttr(graph2, attr); 
      //graph1.TestOrthoNormality(attr);
    }
  //cout << graph1 << endl;
}


void PostProcessCactus1(TwoDArrayColln& countColln, SymTwoDArray& attrInfo,
			Graph& graph, int attr)
{
  /**** YET TO BE DONE;
  int *sortedAttrVals = new int[graph.DomainSz(attr)];
  int *breaks = new int[graph.DomainSz(attr)];
  int numPartns;

  graph.Sort(attr, NON_PRINC_BASIN, sortedAttrVals);
  numPartns = graph.Partition(attr, NON_PRINC_BASIN, sortedAttrVals, breaks);

#ifdef DEBUG
  OutputSortedAttrVals(sortedAttrVals, breaks, graph, attr, NON_PRINC_BASIN);
#endif

  ClusterColln *clusters = new ClusterColln(graph.NumAttrs());
  clusters->CreateClustersForAttr(attr, numPartns, sortedAttrVals, breaks);

  delete [] sortedAttrVals;
  delete [] breaks;

  clusters->AdjustClusters(attr, graph, attrInfo);
  
  cout << *(clusters->GetAttrClusters(attr));
  ******/
}

void OutputSortedAttrVals(int *sortedAttrVals, int *breaks, Graph& graph, int attr, 
			  int basin)
{
  int i;
  int breakNum = 0;
  cout << "Partition number 0" << endl;
  for (i = 0; i < graph.DomainSz(attr); i++)
    {
      if (i == breaks[breakNum])
	{
	  cout << endl << "Partition number " << breakNum+1 << endl;
	  breakNum++;
	}
      cout << sortedAttrVals[i] << " : " << graph.GetWt(sortedAttrVals[i], attr, basin) << endl; 
    }
}


void 
Cactus1(char *inputFile, int numTuples, int numAttrs, int *domSizes,
	Graph& graph1, Graph& graph2, int numIterns)
{
  int i, j;
  int sum=0;

  TwoDArrayColln *countColln;
  countColln = PreProcess(inputFile, numTuples, numAttrs, domSizes);

  for (i = 0; i < 1; i++)
    //for (i = 0; i < numAttrs; i++)
    {
      SymTwoDArray attrInfo(domSizes[i], sum, sum+domSizes[i]-1, i, numAttrs);

      cout << "Created Sym Two D array^^^^^^ " << sum << " " << sum+domSizes[i]-1 << endl;
      for (j = 0; j < i; j++)
	countColln->Join(j, i, FALSE, attrInfo);
      for (j = i+1; j < numAttrs; j++)
	countColln->Join(i, j, TRUE, attrInfo);

      cout << "Joined attributes&&&&&& " << endl;
      ClusterAttrCactus1(attrInfo, graph1, graph2, i, numIterns);
      
      PostProcessCactus1(*countColln, attrInfo, graph1, i);
      sum += domSizes[i];
    }
}



void PostProcessCactus2(SummarizedInfo& summary, ClusterColln& colln)
{

}


void
Cactus2(char *inputFile, int numTuples, int numAttrs, int *domSizes,
				int distSetSz, char *outputFile)
{
  int i;
  Timer timer1("Total Clustering Time");
  Timer timer21("Inter-attribute Summary Construction Time");
  Timer timer22("Intra-attribute Summary Construction Time");
  Timer timer3("Clustering Time");
  Timer timer4("Second Scan Time");
  timer1.Start();

  timer21.Start();
  SummarizedInfo *summary = new SummarizedInfo(numTuples, numAttrs, domSizes);
  summary->Summarize(inputFile);
  timer21.Stop();

  ClusterColln *allAttrClusters;
  allAttrClusters = new ClusterColln(numAttrs);

  for (i = 0; i < numAttrs; i++)
    {
      cout << " Processing Attribute " << i << endl;
      timer22.Start();
      summary->ComputeAttrInfo(i);
      timer22.Stop();

      cout << "Clustering attribute " << i << endl;
      timer3.Start();
      allAttrClusters->InitAttrClusters(i);
      (allAttrClusters->GetAttrClusters(i))->ClusterAttrCactus2(*summary, distSetSz); 
      timer3.Stop();

      summary->CloseAttrInfo(i);
      cout << " Clustered Attribute " << i << endl << endl;;
    }
  PostProcessCactus2(*summary, *allAttrClusters);

  allAttrClusters->SetBaseClusNums();
  cout << *allAttrClusters << endl;

  /***
      timer4.Start();
      SubspaceClusters *subspaceInfo = new SubspaceClusters(numAttrs,*allAttrClusters);
      subspaceInfo->Driver(*allAttrClusters, *(summary->TwoDArrayCollnInfo()),
		       inputFile);
		       timer4.Stop();
		       delete subspaceInfo;
  ***/

  timer1.Stop();

  ofstream ofile(outputFile, ios::app);
  ofile << timer21 << endl;
  ofile << timer22 << endl;
  ofile << timer3 << endl;
  ofile << timer4 << endl;
  ofile << timer1 << endl;
  ofile << endl;
  ofile.close();

  delete summary;
  delete allAttrClusters;
}
