#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <assert.h>

const int MAX_NAME_LENGTH = 300;

class Assignment
{
private:
  char **_strings;
  int _numStrings;
  int _maxSize;

public:
  Assignment(int maxSize)
    {
      _strings = new (char*)[maxSize];
      _numStrings = 0;
      _maxSize = maxSize;
    }

  ~Assignment()
    {
      int i;
      for (i = 0; i < _numStrings; i++)
	delete [] _strings[i];
      delete [] _strings;
    }

  int AssignString(char *str)
    {
      int i;
      for (i = 0; i < _numStrings; i++)
	{
	  if (strcmp(str, _strings[i]) == 0)
	    return i;
	}
      _strings[_numStrings] = new char[strlen(str)+1];
      //cout << strlen(str) << " ";
      strcpy(_strings[_numStrings], str);
      _numStrings++;
      assert(_numStrings < _maxSize);
      return (_numStrings-1);
    }
  
  friend ofstream& operator<<(ofstream& ofile, Assignment& assn);
  
};

ofstream& 
operator<<(ofstream& ofile, Assignment& assn)
{
  int i;
  for (i = 0; i < assn._numStrings; i++)
    {
      ofile << i << "\t" << assn._strings[i] << endl;
    }
  return ofile;
}



main(int argc, char **argv)
{
  if (argc < 2)
    {
      cout << "Usage: <infile> <numLines> <numAttrs> <outfile>" << endl;
      exit(1);
    }

  //input parameters
  char inputFileName[MAX_NAME_LENGTH];
  char outputFileName[MAX_NAME_LENGTH];
  int maxCardinality;
  int numAttrs;

  strcpy(inputFileName, argv[1]);
  maxCardinality = atoi(argv[2]);
  numAttrs = atoi(argv[3]);
  strcpy(outputFileName, argv[4]);

  //parameters for processing
  ifstream ifile; 
  ofstream ofile, tmpOfile;
  
  int numTuples = 0;
  int i = 0;
  Assignment **mapping = new (Assignment*)[numAttrs];
  //  char **tuple = new (char*)[numAttrs];
  char *tuple = new char[numAttrs];
  int *cardinalities = new int[numAttrs];
  char ch;
  char *str = new char[2];
  int mapNum;

  str[1] = '\0';

  for (i = 0; i < numAttrs; i++)
    {
      cardinalities[i] = 0;
      //tuple[i] = new char[100]; // a hack; come back and clean;
      mapping[i] = new Assignment(maxCardinality);
    }

  ifile.open(inputFileName);
  do
    {
      i = 0;
      while (i < numAttrs)
	{
	  ifile >> tuple[i];
	  if (NULL == ifile)
	    break;
	  cout << tuple[i];
	  if (i < numAttrs-1)
	    {
	      ifile >> ch;
	      cout << ch; 
	    }
	  str[0] = tuple[i];
	  mapNum = mapping[i]->AssignString(str);
	  if (mapNum > cardinalities[i])
	    cardinalities[i] = mapNum;
	  i++;
	}
      cout << endl;
    } while (ifile != NULL);
  ifile.close();

  for (i = 0; i < numAttrs; i++)
    cout << cardinalities[i]+1 << " ";
  cout << endl;

  // variables for writing the transformed file;
  int tupSz=numAttrs+3;
  int *arr = new int[tupSz];
  int *beginArr = new int[numAttrs];
  int begin=0;

  cardinalities[0] = 0; // a hack
  beginArr[0] = 0;
  for (i = 1; i < numAttrs; i++)
    {
      beginArr[i] = begin;
      begin += cardinalities[i]+1;
    }


  tmpOfile.open("/tmp/mushr.mapping");
  for (i = 0; i < numAttrs; i++)
    {
      tmpOfile << " Attribute " << i << " " << beginArr[i] << endl;
      tmpOfile << *(mapping[i]) << endl;
    }
  tmpOfile.close();




  // output the formatted file;
  ofile.open(outputFileName);
  ifile.open(inputFileName);
  

  do 
    {
      i = 0;
      arr[0] = numTuples;
      arr[1] = numAttrs-1;
      while (i < numAttrs)
	{
	  ifile >> tuple[i];
	  if (i < numAttrs-1)
	    {
	      ifile >> ch;
	    }
	  if (i > 0) //not writing the edibility value
	    {
	      str[0] = tuple[i]; // yet another hack;
	      mapNum = mapping[i]->AssignString(str);
	      assert(mapNum <= cardinalities[i]);
	      arr[2+i-1] = beginArr[i]+mapNum;	  
	    }
	  i++;
	}
      if (NULL == ifile)
	break;
      arr[tupSz-1-1] = -1;
      numTuples++;
      ofile.write((char*)arr, (tupSz-1)*sizeof(int));
      //cout << arr[0] << "\t" << arr[1] << "\t" << arr[2] << "\t" << arr[3] 
      //   << "\t" << arr[4] << "\t" << arr[5] << "\t" << arr[6] << endl;
    } while (ifile != NULL);

  ifile.close();
  ofile.close();

  cout << "NumTuples :" << numTuples << endl;

  for (i = 0; i < numAttrs; i++)
    {
      delete mapping[i];
      //delete tuple[i];
    }
  delete [] mapping;
  delete [] tuple;
  delete [] cardinalities;
  delete [] arr;
  delete [] beginArr;
}

