#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <assert.h>

const int MAX_NAME_LENGTH = 300;

class Assignment
{
private:
  char **_strings;
  int _numStrings;
  int _maxSize;
  int _startVal;
public:
  Assignment(int maxSize)
    {
      _strings = new (char*)[maxSize];
      _numStrings = 0;
      _maxSize = maxSize;
	  _startVal = 0;
    }

  ~Assignment()
    {
      int i;
      for (i = 0; i < _numStrings; i++)
	delete [] _strings[i];
      delete [] _strings;
    }

  void SetStartVal(int strt)
	{
	  _startVal = strt; 
    }	

  int AssignString(char *str)
    {
      int i;
      for (i = 0; i < _numStrings; i++)
	{
	  if (strcmp(str, _strings[i]) == 0)
	    return i;
	}
      _strings[_numStrings] = new char[strlen(str)+1];
      //cout << strlen(str) << " ";
      strcpy(_strings[_numStrings], str);
      _numStrings++;
      assert(_numStrings < _maxSize);
      return (_numStrings-1);
    }
  
  friend ofstream& operator<<(ofstream& ofile, Assignment& assn);
  
};

ofstream& 
operator<<(ofstream& ofile, Assignment& assn)
{
  int i;
  for (i = 0; i < assn._numStrings; i++)
    {
      ofile << i+assn._startVal << "\t" << assn._strings[i] << endl;
    }
  cout << "Wrote the mapping out" << endl;

}



main(int argc, char **argv)
{
  if (argc < 2)
    {
      cout << "Usage: <infile> <numLines> <outfile>" << endl;
      exit(1);
    }

  //input parameters
  char inputFileName[MAX_NAME_LENGTH];
  char outputFileName[MAX_NAME_LENGTH];
  int maxCardinality;

  strcpy(inputFileName, argv[1]);
  maxCardinality = atoi(argv[2]);
  strcpy(outputFileName, argv[3]);

  //parameters for processing
  ifstream ifile; 
  ofstream ofile, tmpOfile;
  char firstAuthor[MAX_NAME_LENGTH];
  char secondAuthor[MAX_NAME_LENGTH];
  char conference[MAX_NAME_LENGTH];
  char year[MAX_NAME_LENGTH];

  int mapFA, mapSA, mapConf, mapYr;
  int numFA = 0, numSA = 0, numConf = 0, numYr = 0;
  int numTuples=0;

  Assignment *mapFirstAuthor = new Assignment(maxCardinality);
  Assignment *mapSecondAuthor = new Assignment(maxCardinality);
  Assignment *mapConference = new Assignment(maxCardinality);
  Assignment *mapYear = new Assignment(maxCardinality);

  ifile.open(inputFileName);

  do 
    {
      ifile >> firstAuthor; 
      ifile >> secondAuthor;
      ifile >> conference;
      ifile >> year; 

      //cout << firstAuthor << "\t" << secondAuthor << "\t"
      //   << conference << "\t" << year << endl;

      mapFA = mapFirstAuthor->AssignString(firstAuthor);
      if (mapFA > numFA)
	numFA = mapFA;
      mapSA = mapSecondAuthor->AssignString(secondAuthor);
      if (mapSA > numSA)
	numSA = mapSA;
      mapConf = mapConference->AssignString(conference);
      if (mapConf > numConf)
	numConf = mapConf;
      mapYr = mapYear->AssignString(year);
      if (mapYr > numYr)
	numYr = mapYr;
    } while (ifile != NULL);
  ifile.close();

  cout << numFA+1 << "\t" << numSA+1 << "\t" << numConf+1 << "\t" << numYr+1 << endl;

  // variables for writing the transformed file;
  int beginFA, beginSA, beginConf, beginYr;
  int tupSz=6; // changed. 
  int arr[tupSz];

  beginFA = 0; mapFirstAuthor->SetStartVal(beginFA);
  //beginSA = numFA+1; mapSecondAuthor->SetStartVal(beginSA);
  //beginConf = numFA+1+numSA+1; mapConference->SetStartVal(beginConf);
  //beginYr = numFA+1+numSA+1+numConf+1; mapYear->SetStartVal(beginYr);

  beginSA=0;
  beginConf = numFA+1; mapConference->SetStartVal(beginConf);
  beginYr = numFA+1+numConf+1; mapYear->SetStartVal(beginYr);

  // output the formatted file;
  ofile.open(outputFileName);
  ifile.open(inputFileName);
  

  do 
    {
      ifile >> firstAuthor; 
      ifile >> secondAuthor;
      ifile >> conference;
      ifile >> year; 

      mapFA = mapFirstAuthor->AssignString(firstAuthor);
      assert(mapFA <= numFA);

      mapSA = mapSecondAuthor->AssignString(secondAuthor);
      assert(mapSA <= numSA);

      mapConf = mapConference->AssignString(conference);
      assert(mapConf <= numConf);

      mapYr = mapYear->AssignString(year);
      assert(mapYr <= numYr);

      arr[0] = numTuples;
      /***
      arr[1] = 4;
      arr[2] = beginFA+mapFA;
      arr[3] = beginSA+mapSA;
      arr[4] = beginConf+mapConf;
      arr[5] = beginYr+mapYr;
      arr[6] = -1;
      ***/
      arr[1] = 3;
      arr[2] = beginFA+mapFA;
      arr[3] = beginConf + mapConf;
      arr[4] = beginYr + mapYr;
      arr[5] = -1;
      numTuples++;
      ofile.write((char*)arr, tupSz*sizeof(int));
      
      
      //cout << arr[0] << "\t" << arr[1] << "\t" << arr[2] << "\t" << arr[3] 
      //   << "\t" << arr[4] << "\t" << arr[5] << "\t" << arr[6] << endl;
    } while (ifile != NULL);

  ifile.close();
  ofile.close();

  cout << "Dimension Domain Sizes: " << numFA+1 << " " << numSA+1 << " ";
  cout << " " << numConf+1 << " " << numYr+1 << endl;
  cout << "NumTuples :" << numTuples << endl;

  tmpOfile.open("/tmp/mapping");
  cout << "Writing the FirstAuthor map" << endl;
  tmpOfile << *mapFirstAuthor << endl << endl;
  cout << "Writing the SecondAuthor map" << endl;
  //tmpOfile << *mapSecondAuthor << endl << endl;
  cout << "Writing the conference map" << endl;
  tmpOfile << *mapConference << endl << endl;
  cout << "Writing the year map " << endl;
  tmpOfile << *mapYear << endl << endl;
  tmpOfile.close();

  delete mapFirstAuthor;
  delete mapSecondAuthor;
  delete mapConference;
  delete mapYear;
}

