#include <iostream.h>
#include <assert.h>
#include <fstream.h>
#include <math.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include "tuple.h"
#include "globals.h"

const int MAX_NUMTUPLES = 100000;

int _index(int k, int numAttrs, int i)
{
  return k*numAttrs+i;
}

main(int argc, char **argv)
{
  if (argc < 5)
    {
      cout << "Usage: <outPrefix> <numTuples> <numAttrs> <domSzs[]> <numExtraTuples> <numClusters> <lowrestrns[], highrestrns[]>*" << endl;
      exit(1);
    }
  
  int numAttrs;
  int numTuples;
  int numClusters;
  int *extraLowRestrns;
  int *extraHighRestrns;
  int *domSizes;
  int numExtraTuples;
  char fileNamePrefix[MAX_NAME_LENGTH];
  char fileName[MAX_NAME_LENGTH];
  GenTuple *tupleGenerator;
  int fd, nbytes;
  int i,j,k;
  int numAStep=0;

  strcpy(fileNamePrefix, argv[1]);
  numTuples = atoi(argv[2]);
  numAttrs = atoi(argv[3]);
  
  domSizes = new int[numAttrs];

  //get the domain sizes on each attribute.
  numAStep=0;
  for (i = 0; i < numAttrs; i++)
    {
      numAStep++;
      domSizes[i] = atoi(argv[3+i+1]);
      if (domSizes[i] > 0)
	{
	  for (j = 1; j < numAttrs; j++)
	    domSizes[j] = domSizes[i];
	  break;
	}
      else
	domSizes[i] = -1 * domSizes[i];
    }
  
  numExtraTuples = atoi(argv[3+numAStep+1]);
  numClusters = atoi(argv[3+numAStep+2]);

  extraLowRestrns = new int[numAttrs*numClusters];
  extraHighRestrns = new int[numAttrs*numClusters];

  // obtain the cluster boundaries on each attribute.
  for (k = 0; k < numClusters; k++)
    {
      if (1 == numAStep)
	{
	  extraLowRestrns[_index(k, numAttrs, i)] = 
	    atoi(argv[3+numAStep+2+ 2*k*numAStep+1 + i]);
	  extraHighRestrns[_index(k, numAttrs, i)] = 
	    atoi(argv[3+numAStep+2+ 2*k*numAStep+numAStep+1 + i]);

	  for (j = 1; j < numAttrs; j++)
	    {
	      extraLowRestrns[_index(k, numAttrs, j)] = 
		extraLowRestrns[_index(k, numAttrs, i)];
	      extraHighRestrns[_index(k, numAttrs, j)] = 
		extraHighRestrns[_index(k, numAttrs, i)];
	    }
	}
      else 
	{
	  for (i = 0; i < numAttrs; i++)
	    extraLowRestrns[_index(k, numAttrs, i)] = atoi(argv[3+numAStep+2+ 2*k*numAStep+1 + i]);
	  for (i = 0; i < numAttrs; i++)
	    extraHighRestrns[_index(k, numAttrs, i)] = atoi(argv[3+numAStep+2+ 2*k*numAStep+numAStep+1 + i]);
	}
    }

  cout << fileNamePrefix << endl;
  for (i = 0; i < numClusters*numAttrs; i++)
    {
      cout << extraLowRestrns[i] << " " 
	   << extraHighRestrns[i] << endl;
    }


  //All the arguments have been taken in; Start the data generation.
 
  //form the output filename. 
  //ideally, the fileName should be an interpretable extension of the prefix.
  sprintf(fileName, "%s%dd", fileNamePrefix, numAttrs); 
  fd = open(fileName, O_WRONLY | O_CREAT | O_TRUNC, 511);
  assert(fd != -1);



  int numTupGenerated=0;
  int numToBeGenerated=0;
  int tupSz = 2+numAttrs+1;
  int *tuples = new int[(2+numAttrs+1)*(MAX_NUMTUPLES)];
  tupleGenerator = new GenTuple(numAttrs, domSizes);

  cout << endl;
  cout << sizeof(int)*(numTuples+numExtraTuples*numClusters)*tupSz << endl;

  while (numTupGenerated < numTuples)
    {
      numToBeGenerated=0;
      if (numTuples - numTupGenerated < MAX_NUMTUPLES)
	numToBeGenerated = numTuples-numTupGenerated;
      else
	numToBeGenerated = MAX_NUMTUPLES;

      //tuples = new int[(2+numAttrs+1)*(numTuples + numClusters*numExtraTuples)];
      
      for (i = 0; i < numToBeGenerated; i++)
	{
	  tuples[i*(tupSz)] = numTupGenerated+i; //tid; 
	  tuples[i*tupSz + 1] = numAttrs; //tid; 
	  tupleGenerator->GenNextTuple(tuples+2+i*tupSz, NULL);
	  tuples[i*tupSz +2+ numAttrs] = -1; //for the lits computation;
	}

      nbytes = write(fd, tuples, sizeof(int)*(numToBeGenerated)*tupSz);
      assert(nbytes == sizeof(int)*(numToBeGenerated)*tupSz);

      numTupGenerated += numToBeGenerated;
    }

  cout << numTupGenerated << "=?" << numTuples << endl;
  delete [] tuples; 
  tuples = new int[(2+numAttrs+1)*(numExtraTuples*numClusters)];

  for (k = 0; k < numClusters; k++)
    for (i = 0; i < numExtraTuples; i++)
      {
	tuples[(k*numExtraTuples + i)*tupSz] = numTuples + i; // tid;
	tuples[(k*numExtraTuples + i)*tupSz + 1] = numAttrs;
	tupleGenerator->GenNextTuple(tuples+(k*numExtraTuples + i)*tupSz+2, 
				     extraLowRestrns + k*numAttrs, 
				     extraHighRestrns + k*numAttrs);
	tuples[(k*numExtraTuples +i)*tupSz+2+numAttrs] = -1;
      }


  nbytes = write(fd, tuples, sizeof(int)*(numExtraTuples*numClusters)*tupSz);
  assert(nbytes == sizeof(int)*(numExtraTuples*numClusters)*tupSz);
  close(fd);
  
  delete [] domSizes;
  delete [] extraLowRestrns;
  delete [] extraHighRestrns;
  delete [] tuples;
  delete tupleGenerator;
}
