#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <math.h>
#include "node.h"
#include "Set.h"
#include "globals.h"
#include "globals_catClus.h"
#include "trans_utils.h"
#include "PrefixTree.h"
#include "timeutil.h"

extern void Cactus1(char *inputFile, int numTuples, int numAttrs, 
		    int *domSizes, Graph& graph1, Graph& graph2, int numIterns);
extern void Cactus2(char *inputFile, int numTuples, int numAttrs, int *domSizes,
		    int distSetSz, char *outputFile);

int
main(int argc, char **argv)
{
  if (argc < 11)
    {
      cout << "Usage: algo <algNum (0:STIRR, 1:Cactus1, 2:Cactus2)> <inputfile> <numTuples> <numAttrs> <domSzArr> <basinSz> <numIterns> <minSup> <distSetSz> <outputfile>" << endl;
      exit(1);
    }

  char inputFile[MAX_NAME_LENGTH];
  char outputFile[MAX_NAME_LENGTH];
  int numTuples;
  Algorithm algNum; // which algorithm to run;
  int numA;
  int *domSizes;
  int *basinSzs;
  int numIterns;
  int minSup;
  int distSetSz;
  int i,j;
  int temp;
  
  temp = atoi(argv[1]);
  if (0 == temp)
    algNum = STIRR;
  else if (1 == temp)
    algNum = CACTUS1;
  else 
    algNum = CACTUS2;

  strcpy(inputFile, argv[2]);
  numTuples = atoi(argv[3]);
  numA = atoi(argv[4]);
  domSizes = new int[numA];
  basinSzs = new int[numA];
  int numItems = 0;
  PrefixTree *pt = NULL;
  int numAStep=0;

  for (i = 0; i < numA; i++)
    {
      numAStep++;
      domSizes[i] = atoi(argv[4+i+1]);
      if (domSizes[i] > 0)
	{
	  numItems += domSizes[i];
	  for (j=1; j < numA; j++)
	    {
	      numItems += domSizes[i];
	      domSizes[j] = domSizes[i];
	    }
	  break;
	}
      else
	{
	  domSizes[i] = -1 * domSizes[i]; //changing sign
	  numItems += domSizes[i];
	}
    }

  //All attributes have the same basin sizes (can be diff.)
  basinSzs[0] = atoi(argv[4+numAStep+1]);
  for (i=1; i < numA; i++)
    basinSzs[i] = basinSzs[0];

  numIterns = atoi(argv[4+numAStep+2]);
  minSup = atoi(argv[4+numAStep+3]);
  distSetSz = atoi(argv[4+numAStep+4]);
  strcpy(outputFile, argv[4+numAStep+5]);

  /***
  if (1 == algNum)
    {
      pt = driverApriori(inputFile, numItems, minSup);
    } 
  ***/

  // create the graph;
  srand48(1001);
  Graph *graph1 = new Graph(numA, domSizes, basinSzs);
  Graph *graph2 = new Graph(numA, domSizes, basinSzs);

  //graph1->Initialize(RANDOM);
  graph1->Initialize(IDENTITY);
  graph2->Initialize(IDENTITY);
  
  if (CACTUS1 == algNum) 
    {
      cout << "Iterating..." << endl;
      Cactus1(inputFile, numTuples, numA, domSizes, *graph1, *graph2, numIterns);
    }
  if (CACTUS2 == algNum)
    Cactus2(inputFile, numTuples, numA, domSizes, distSetSz, outputFile);
  else 
    {
      Timer stirrTimer("Time taken by STIRR for clustering");
      stirrTimer.Start();
      for (i = 0; i < numIterns; i++)
	{
	  cout << " Iteration number " << i << endl;
#ifdef DEBUG
	  graph1->TestOrthoNormality();
#endif
	  graph1->IterateSTIRR(inputFile, graph2, pt);
	  
	  *graph1 = *graph2;
	}
      stirrTimer.Stop();
      ofstream ofile;
      ofile.open(outputFile, ios::app);
      ofile << stirrTimer << endl << endl;
      ofile.close();
      cout << *graph1 << endl;
    }
  
  delete [] domSizes;
  delete [] basinSzs;
  delete graph1;
  delete graph2;
  /***
  if (1 == algNum)
    delete pt;    
  ***/
}
