#include <iostream.h>
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "Set.h"
#include "globals.h"
#include "globals_catClus.h"
#include "trans_utils.h"
#include "node.h"
#include "cluster.h"
#include "2DArray.h"
#include "PrefixTree.h"


ThruAttrClusters::ThruAttrClusters(int numC)
{
  _numClusters = numC;
  _maxNumClusters = MAX_NUMCLUSTERS;


  _clusStrengths = new int[MAX_NUMCLUSTERS];
  _nodeIds = new (OneDArray*)[MAX_NUMCLUSTERS];
  memset((void*)_clusStrengths, 0, MAX_NUMCLUSTERS*sizeof(int));
  memset((void*)_nodeIds, 0, MAX_NUMCLUSTERS*sizeof(OneDArray*));
  //_shortestPaths = new (SymTwoDArray*)[_numClusters];
  //memset((void*)_nodeIds, 0, _numClusters*sizeof(SymTwoDArray*));

#ifdef PATH_CLUSTERING
  _pathLengths = new (SymTwoDArray*)[MAX_NUMCLUSTERS];
#endif

#if 0 // only needed for Cactus1;
  int i, j;
  int low, high;

  for (i = 0; i < _numClusters; i++)
    {
      if (0 == i)
	{
	  low = 0;
	  high = breaks[0];
	}
      else
	{
	  low = breaks[i-1];
	  high = breaks[i];
	}
      _nodeIds[i] = new OneDArray(high-low, 2*(high-low));
      for (j = low; j < high; j++)
	{
	  _nodeIds[i]->Set(j-low, sortedIndices[j]);
	}
    }
#endif
}

/****** Clusters attribute ``attr'' using the similarity information obtained
using ``thruAttr'';
**********/
void
ThruAttrClusters::ClusterAttrCactus2(SummarizedInfo& summary, int attr, int thruAttr,
				     int distSetSz)
{
  _CompleteThruAttrClusters(summary, attr, thruAttr, distSetSz);
}


//Finds all possible distinguishing sets of size less than or equal to distSetSz
//and then extends them to find all cluster projections.
void
ThruAttrClusters::_CompleteThruAttrClusters(SummarizedInfo& summary, int attr, 
					    int thruAttr, int distSetSz)
{
#ifdef DEBUG
  cout << "Inside ThruAttrClusters::_CompleteThruAttrClusters" << endl;
  cout << "*************ThruClusters on " << attr << " : " << thruAttr << endl;
#endif DEBUG

  int i,j;
  int Mark = 1; // to mark a set of vals as a cluster;
  int numNew = 0;
  int numAttrVals = summary.DomSize(attr);
  int strength=0;
  int size = 2;

#ifdef DEBUG
  int numStrongPairs=0;
#endif

  //summary->ComputeAttrInfo(attr, thruAttr);
  SymTwoDArray& attrInfo = *(summary.AttrInfo(attr));
  int arr[2];
  Set candSet(0);
  candSet.newArr(2, arr);

  PrefixTree **ptArr = new (PrefixTree*)[distSetSz+1];
  for (i = 0; i < distSetSz+1; i++)
    ptArr[i] = NULL;

  // a hack to allow arbitrary starting pts for attribute values.
  ptArr[2] = new PrefixTree(summary.StartAttrVal(attr) + numAttrVals, 
			    NULL, 0);
  ptArr[2]->setMinSupport(1); 
  
  
  for (i = 0; i < summary.StartAttrVal(attr); i++)
    ptArr[2]->SetItemCountStatus(i, 0, small);

  for (i=summary.StartAttrVal(attr); i < summary.StartAttrVal(attr)+numAttrVals; i++)
    {
      ptArr[2]->SetItemCountStatus(i, Mark, candidate);
      for (j = i+1; j < summary.StartAttrVal(attr)+numAttrVals; j++)
	{
	  strength = attrInfo.get(i, j, thruAttr);
	  if (strength > 0) // strongly connected pair;
	    {
	      //cout << i << "," << j << "=" << strength << " ; " << endl;
	      arr[0] = i;
	      arr[1] = j;
	      candSet.newArr(2, arr);
	      ptArr[2]->insertWithCount(candSet, Mark, negBorder);
#ifdef DEBUG	      
	      numStrongPairs++;
#endif
	    }
	}
    }
  candSet.Nullify();

#ifdef DEBUG
  cout << "CompleteThruClusters:: Number of strong pairs = " << numStrongPairs << endl;
#endif

  numNew = 1;
  while ((numNew > 0) && (size < distSetSz))
    {
      //keep generating new candidates through an index nested loops join;
      numNew=0;
      cout << " Joining...." << endl;
      ptArr[size+1] = ptArr[size]->JoinCC(numNew, summary, attr, thruAttr);
      cout << " ******** Number of New candidates " << numNew << endl;
      size++;
    } 

  //cout << " Extending Distinguishing Sets" << endl;
  //extend distSets and populate clusters;
  if (size == distSetSz)
    _FindClustersFromDistSets(summary, ptArr, distSetSz, attr, thruAttr);

  //cout << " Extended Distinguishing Sets" << endl;
 
  for (i = 0; i < distSetSz+1; i++)
    if (ptArr[i] != NULL)
      delete ptArr[i];
  delete [] ptArr;
}


//ptArr[0]=ptArr[1]==NULL;
void
ThruAttrClusters::_FindClustersFromDistSets(SummarizedInfo& summary, 
					    PrefixTree **ptArr, int distSetSz,
					    int attr, int thruAttr)
{
  int i=distSetSz;
  int j,k;
  int oldNumClusters=0;

  for (i = 2; i <= distSetSz; i++)
    {
      ptArr[i]->clearCounts();
      ptArr[i]->SetPassLevel(i);
    }

  i = distSetSz;
  while ( i >= 2)
    {
      oldNumClusters = NumClusters();
#ifdef DEBUG
      cout << "Inside _FindClustersFromDistSets " << i << endl; 
#endif
      ptArr[i]->ExtendDistSets(summary, attr, thruAttr, *this);

      //cout << oldNumClusters << " : " << NumClusters() << endl; 
      //marking subsets of new clusters in prefix tree consisting
      // of cliques smaller the distSetSz
      for (j = i-1; j >=2; j--)
	{
	  for (k = oldNumClusters; k < NumClusters(); k++)
	    {
	      //cout << ClusStrength(k) << endl;
	      Set xact(Size(k));
	      ReturnClusterSet(k, xact);
	      assert(Size(k) == xact.size());
	      ptArr[j]->incrementCounts(xact, ClusStrength(k), allLevels);
	    }
	}
      i--;
#ifdef DEBUG
      cout << "End of while loop: _FindClustersFromDistSets " << i << endl; 
#endif
    }
}


//returns the index of the new cluster;
//strength is the size on thruAttr dimension of the cluster.
int 
ThruAttrClusters::ExtendDistSet(SummarizedInfo& summary, int attr, int thruAttr, 
				Set& distSet, int& strength)
{
  int i;
  int beginAttr, endAttr;
  int *siblingSet = new int[MAX_CLUSTERSIZE];
  TwoDArray* attrPairInfo = NULL;

  SymTwoDArray& attrInfo = *(summary.AttrInfo(attr));
  attrPairInfo = summary.TwoDArrayInfo(attr, thruAttr);
  //if (attr < thruAttr)
  //else
  //attrPairInfo = summary.TwoDArrayInfo(thruAttr, attr);

  strength=0;
  strength = summary.FindSiblingSet(attr,thruAttr,distSet, siblingSet);

  if (strength > 0)
    {
      _clusStrengths[_numClusters] = strength;
      _nodeIds[_numClusters] = new OneDArray(0, MAX_CLUSTERSIZE);

      beginAttr = attrInfo.Begin();
      endAttr = attrInfo.End();

      for (i = 0; i < distSet.size(); i++)
	_nodeIds[_numClusters]->Append(distSet.elements(i));

      for (i = beginAttr; i <= endAttr; i++)
	{
	  if (FALSE == _nodeIds[_numClusters]->Member(i))
	    {
	      //if (AvgStrength(_numClusters, i, attrInfo, thruAttr) >= BETA*strength)
	      if (AvgStrength(i, *attrPairInfo, attr, thruAttr, 
			      siblingSet, strength) >= BETA*strength)
		_nodeIds[_numClusters]->Append(i);
	    }
	}
      _nodeIds[_numClusters]->Sort();
      _numClusters++;
      assert(_numClusters < MAX_NUMCLUSTERS);
#ifdef REALDATA
      cout << "The distinguishing set is ";
      for (i = 0; i < distSet.size(); i++)
	cout << distSet.elements(i) << " ";
      cout << endl;

      cout << "The sibling  set is ";
      for (i = 0; i < strength; i++)
	cout << siblingSet[i] << " ";
      cout << endl;

      cout << "The extended set is ";
      for (i = 0; i < _nodeIds[_numClusters-1]->Size(); i++)
	cout << _nodeIds[_numClusters-1]->Get(i) << " ";
      cout << endl;
#endif
      attrPairInfo = NULL;
      delete [] siblingSet;
      return (_numClusters-1);
    }
  attrPairInfo = NULL;
  delete [] siblingSet;
  return -1;
}


//augment the clusters detected uptil now with "attrVal";
void
ThruAttrClusters::AssimilateNewAttrVal(int attrVal, SymTwoDArray& attrInfo,
				       double threshold, int thruAttr)
{
  int i;
  bool flag=FALSE;
  double strength=0.0;

  //cout << "Assim.NewVal: Threshold " << attrVal << "\t" << threshold << endl;
  if (0 == _numClusters)
    {
      _nodeIds[_numClusters] = new OneDArray(0, MAX_CLUSTERSIZE);
      _nodeIds[_numClusters]->Append(attrVal);
      _numClusters++;
    }
  else 
    {
      for (i = 0; i < _numClusters; i++)
	{
	  assert(0);
	  exit(1);
	  //strength = AvgStrength(i, attrVal, attrInfo, thruAttr);
	  //cout << " Strength with cluster " << i << " is " << strength << endl;
	  if (strength >= threshold)
	    {
	      flag = TRUE;
	      _nodeIds[i]->Append(attrVal);
	    }
	}
      if (FALSE == flag)
	{
	  _nodeIds[_numClusters] = new OneDArray(0,MAX_CLUSTERSIZE);
	  _nodeIds[_numClusters]->Append(attrVal);
	  _numClusters++;
	}
    }
}

//SymTwoDArray& attrInfo, int thruAttr)
// strength += attrInfo.get(attrVal, attrVal2, thruAttr);

double 
ThruAttrClusters::AvgStrength(int attrVal, TwoDArray& attrPairInfo,
			      int attr, int thruAttr, int *siblingSet, int sibStr)
{
  int j;
  int attrVal2;
  double strength = 0.0;

  for (j = 0; j < sibStr; j++)
    if (attr < thruAttr)
      {
	if (attrPairInfo.get(attrVal, siblingSet[j]) >= attrPairInfo.Threshold())
	  strength += 1.0;
      }
    else
      {
	if (attrPairInfo.get(siblingSet[j], attrVal) >= attrPairInfo.Threshold())
	  strength += 1.0;
      }
  return (double)strength/((double)sibStr);
}

void
ThruAttrClusters::RemoveSmallClusters()
{
  int j, i=0, numDeleted=0;

  for (i = 0; i < _numClusters; i++)
    {
      if (_nodeIds[i]->Size() <= 1)
	{
	  delete _nodeIds[i]; 
	  _nodeIds[i] = NULL;
	  numDeleted++;
	}
    }

  if (0 == numDeleted) return;

  i = 0; j = _numClusters;

  while (i < j)
    {
      while (_nodeIds[i] != NULL) i++;
      while (_nodeIds[j] == NULL) j--;
      if (i < j)
	{
	  _nodeIds[i] = _nodeIds[j];
	  _nodeIds[j] = NULL;
	  i++;
	  j--;
	}
    }
  _numClusters -= numDeleted;
}

/****
bool 
AttrClusters::GrowCluster(int attrVal, int cluster, SymTwoDArray& attrInfo)
{
  double strength = AvgStrength(attrVal, cluster, attrInfo);

  if (strength >= _minStrength)
    {
      _nodeIds[cluster]->Append(attrVal);
      return TRUE;
    }
  return FALSE;
}


bool 
AttrClusters::CropCluster(int attrVal, int cluster, SymTwoDArray& attrInfo)
{
  double strength = AvgStrength(attrVal, cluster, attrInfo);
  if (strength < _minStrength)
    {
      _nodeIds[cluster]->Delete(attrVal);
      return TRUE;
    }
  return FALSE;
}


void 
AttrClusters::AdjustClusters(int attr, Graph &graph,
				  SymTwoDArray& attrInfo)
{
  int i, j;
  int start = graph.StartAttrVal(attr);

  for (i = start; i < start + graph.DomainSz(attr); i++)
    {
      for (j = 0; j < NumClusters(); j++)
	{
	  if (FALSE == Member(i, j))
	    {
	      GrowCluster(i, j, attrInfo);
	    }
	  else
	    CropCluster(i, j, attrInfo);
	}
    }
}
***/


ostream& 
operator<<(ostream& of, ThruAttrClusters& clus)
{
  of << "\tNumber of clusters = " << clus._numClusters << endl;
  
  int i, j;
  for (i = 0; i < clus._numClusters; i++)
    {
      //if (clus._nodeIds[i]->Size() <= 1)
      //continue;
      of << "\tCluster Number " << i << " of size " << clus._nodeIds[i]->Size() << endl;
      of << "\t";
      for (j = 0; j < clus._nodeIds[i]->Size(); j++)
	of << "\t" << clus._nodeIds[i]->Get(j);
      of << endl;
    }
  return of;
}


AttrClusters::AttrClusters(int attr, int numAttrs, int numC)
{
  _attr = attr;
  _numAttrs = numAttrs;
  _baseClusterNum = -1;
  _thruClusters = new (ThruAttrClusters*)[_numAttrs+1]; 
  //the extra one is for the final intersection

  int i;
  for (i = 0; i < _numAttrs; i++)
    if (i != _attr)
      _thruClusters[i] = new ThruAttrClusters(numC);
  _thruClusters[_numAttrs] = NULL;
  _thruClusters[_attr] = NULL;
}


void
AttrClusters::IntersectClusters()
{
  // currently not implemented. 
  // Should come back to this and implement it;
  // for now, just changing pointers;
  
  int i;
  assert(NULL == _thruClusters[_numAttrs]);
  assert(NULL != _thruClusters[(_attr+1)%_numAttrs]);

#ifdef SUBSPACE
  //union all clusters into  _thruClusters[_numAttrs];
  int numUnintersectedClus = 0;
  int j;
  for (i = 0; i < _numAttrs; i++)
    if (_attr != i)
      numUnintersectedClus += _thruClusters[i]->NumClusters();
  
  cout << "Tot num unintersected clusters on attr " << _attr
       << " is " << numUnintersectedClus << endl;

  _thruClusters[_numAttrs] = new ThruAttrClusters(0); 
  OneDArray* temp;
  for (i = 0; i < _numAttrs; i++)
    {
      if (i != _attr)
	for (j = 0; j < _thruClusters[i]->NumClusters(); j++)
	  {
	    temp = _thruClusters[i]->ReturnClusterSetNull(j);
	    assert(temp != NULL);
	    _thruClusters[_numAttrs]->AddNewCluster(temp);
	    temp = NULL;
	  }
    }
#else
  //Ideally, it should be a repeated intersection join. 
  //For now, a lazy second.
#ifndef INTERSECTION
  _thruClusters[_numAttrs] = _thruClusters[(_attr+1)%_numAttrs];
  _thruClusters[(_attr+1)%_numAttrs] = NULL;
  //_thruClusters[_numAttrs]->SortClusters();
#else
  Initialize _thruClusters[_numAttrs];
  for each attribute call a new function. 
#endif

#endif

  //we can even delete all the other _thruClusters;
  for (i = 0; i < _numAttrs; i++)
    {
      if ((i != _attr) && (NULL != _thruClusters[i]))
	delete _thruClusters[i];
      _thruClusters[i] = NULL;
    }
}

#ifdef INTERSECTION
void
ThruAttrClusters::IntersectClusters(ThruAttrClusters *result)
{


}
#endif

void 
AttrClusters::ClusterAttrCactus2(SummarizedInfo &summary, int distSetSz)
{
  int i;

  for (i = 0; i < _numAttrs; i++)
    {
      cout << "Clustering Attribute " << _attr << " w.r.t. " << i << endl;
      if (_attr != i)
	_thruClusters[i]->ClusterAttrCactus2(summary, _attr, i, distSetSz);
    }

#ifdef DEBUG
  assert(NULL != _thruClusters[(_attr+1)%_numAttrs]);
  cout << *_thruClusters[(_attr+1)%_numAttrs] << endl;
#endif
  
  cout << "ClusterAttrCactus2: Intersecting Clusters " << endl;
  IntersectClusters();
  cout << "ClusterAttrCactus2: Intersected Clusters " << endl;
}


ostream& 
operator<<(ostream& of, AttrClusters& clus)
{
  of << "Base cluster Number = " << clus.BaseClusterNum() << endl;
  of << *(clus._thruClusters[clus._numAttrs]);
  return of;
#if 0
  int i;
  for (i = 0; i < clus._numAttrs; i++)
    {
      if (i != clus._attr)
	{
	  //cout << " i " << i << " clus._attr " << clus._attr << endl;
	  assert(clus._thruClusters[i] != NULL);
	  of << *(clus._thruClusters[i]);
	  //break;
	}
    }  
  return of;
#endif
}



ClusterColln::ClusterColln(int numA)
{
  _numAttrs = numA;
  _attrClusters = new (AttrClusters*)[_numAttrs];
  memset((void*)_attrClusters, 0, _numAttrs*sizeof(AttrClusters*));
}

void ClusterColln::InitAttrClusters(int attr)
{
  assert(_attrClusters[attr] == NULL);
  _attrClusters[attr] = new AttrClusters(attr, _numAttrs, 0);
}


double
ClusterColln::ConnectionStrength(int attr1, int cluster1, int attr2, int cluster2,
				 TwoDArrayColln& attrPairsInfo)
{
  int strength = 0, linkStrength=0;
  int i, j;
  int sz1, sz2;
  int attrVal1=-1, attrVal2=-1;

  assert(cluster1 < _attrClusters[attr1]->NumClusters());
  assert(cluster2 < _attrClusters[attr2]->NumClusters());

  sz1 = _attrClusters[attr1]->Size(cluster1);
  sz2 = _attrClusters[attr2]->Size(cluster2);

#ifdef DEBUG
  cout << "Threshold for " << attr1 << "," << cluster1 << " " << attr2 << "," 
       << cluster2 << " is " 
       << attrPairsInfo.Threshold(attr1, attr2) << " :: strength = ";
#endif

  for (i = 0; i < sz1; i++)
    for (j = 0; j < sz2; j++)
      {
	attrVal1 = _attrClusters[attr1]->AttrVal(cluster1, i);
	attrVal2 = _attrClusters[attr2]->AttrVal(cluster2, j);
	linkStrength = attrPairsInfo.Count(attr1, attrVal1, attr2, attrVal2);
	if (linkStrength > attrPairsInfo.Threshold(attr1, attr2))
	  strength++;
      }
#ifdef DEBUG
  cout << (strength/(double)(sz1 * sz2)) << endl;
#endif
  return (strength/(double)(sz1 * sz2));
}


ostream&
operator<<(ostream& of, ClusterColln& clusterColln)
{
  int i;
  for (i = 0; i < clusterColln._numAttrs; i++)
    {
      of << " The set of clusters on Attribute " << i << endl;
      of << *(clusterColln._attrClusters[i]);
      of << endl;
    }
  return of;
}


SubspaceClusters::SubspaceClusters(int numA, ClusterColln& attrClusters)
{
  _numClusters=0;
  _numAttrGroups = 0;

  int i;
  int startVal=0;
  int numOneClusters = attrClusters.TotNum1DClusters(attrClusters.NumAttrs()-1);
  _numAttrs = attrClusters.NumAttrs();
  //_pt = new PrefixTree(numOneClusters, NULL, 0);
  //_pt->setMinSupport(0);

  if (_numAttrs > MAX_NUMATTRSPERGRP)
    {
      _numAttrGroups = (0 == _numAttrs % MAX_NUMATTRSPERGRP) ? 
	(int)(_numAttrs/MAX_NUMATTRSPERGRP) :	
	_numAttrGroups = (int)(_numAttrs/MAX_NUMATTRSPERGRP)+1;
    }
  else
    _numAttrGroups = 1;

  _beginVals = new int[_numAttrGroups+1];
  _pt = new (PrefixTree*)[_numAttrGroups];
  startVal=0;
  for (i = 0; i < _numAttrGroups; i++)
    {
      _beginVals[i] = startVal;
      startVal += MAX_NUMATTRSPERGRP;
      _pt[i] = new PrefixTree(numOneClusters, NULL, 0);
      _pt[i]->setMinSupport(0);
    }
  _beginVals[_numAttrGroups] = _numAttrs;
}


SubspaceClusters::~SubspaceClusters()
{
  int i;
  for (i = 0; i < _numAttrGroups; i++)
    if (NULL != _pt[i]) delete _pt[i];
  delete [] _pt;
  //delete _pt;
}

void 
SubspaceClusters::Driver(ClusterColln& attrClusters, TwoDArrayColln& attrPairsInfo,
			 char *inputFile)
{
  int i;

  attrClusters.SetBaseClusNums();
  
  /******
  int numCands = 0;

  Set set(0);
  int arr[2];
  //insert valid two itemsets (or clusters on two dimns);
  for (i = 0; i < attrClusters.NumAttrs(); i++)
    for (j = i+1; j < attrClusters.NumAttrs(); j++)
      for (k = attrClusters.Base(i); k < attrClusters.TotNum1DClusters(i); k++)
	for (l = attrClusters.Base(j); l < attrClusters.TotNum1DClusters(j); l++)
	  {
	    if (attrClusters.ConnectionStrength(i, k-attrClusters.Base(i), 
						j, l-attrClusters.Base(j), 
						attrPairsInfo) >= BETA)
	      {
		arr[0] = k;
		arr[1] = l;
		set.newArr(2,arr);
		_pt->insertWithCount(set, VALID, candidate);
		_numClusters++;
	      }
	  }
  cout << " Joining clusters on fewer subsets " << _numClusters << endl;
  do
    {
      _pt->join(numCands);
      _numClusters += numCands;
      //cout << " Num new Cands " << numCands << endl;
    } while (numCands > 0);
  cout << _numClusters << endl;
  //cout << *_pt << endl;
  set.Nullify();
  ********/

  int dummy=0;
  PrefixTree *ptNew;
  for (i = 0; i < _numAttrGroups; i++)
    {
      DriverForGroup(attrClusters, attrPairsInfo, i);
      //mark maximal clusters in _pt[i];
      ptNew = new PrefixTree(attrClusters.TotNum1DClusters(attrClusters.NumAttrs()-1),NULL,0);
      _pt[i]->MarkPossMaximalIts(dummy);
      _pt[i]->UnMarkNonMaximalIts(dummy, ptNew);
      delete _pt[i];
      _pt[i] = ptNew;
      ptNew = NULL;
    }

  //we need to generate larger groups here.
  ValidateClusters(inputFile, attrClusters);
}

/*
 *If the number of attributes is very high then the set of attributes is 
 *divided into _numAttrGroups groups; DriverForGroup(...,i) is the bottom-up
 *cluster synthesizer for group i;
 *groupNum: the id for the group of attributes;
 */

void
SubspaceClusters::DriverForGroup(ClusterColln& attrClusters, 
				 TwoDArrayColln& attrPairsInfo, int groupNum)
{
  int i, j, k, l;
  int numCands = 0;
  int level;

  Set set(0);
  int arr[2];

  assert(groupNum < _numAttrGroups);
  //insert valid two itemsets (or clusters on two dimns);
  for (i = _beginVals[groupNum]; i < _beginVals[groupNum+1]; i++)
    for (j = i+1; j < _beginVals[groupNum+1]; j++)
      for (k = attrClusters.Base(i); k < attrClusters.TotNum1DClusters(i); k++)
	for (l = attrClusters.Base(j); l < attrClusters.TotNum1DClusters(j); l++)
	  {
	    if (attrClusters.ConnectionStrength(i, k-attrClusters.Base(i), 
						j, l-attrClusters.Base(j), 
						attrPairsInfo) >= BETA)
	      {
		arr[0] = k;
		arr[1] = l;
		set.newArr(2,arr);
		_pt[groupNum]->insertWithCount(set, VALID, candidate);
		_numClusters++;
	      }
	  }
  cout << " Joining clusters on fewer subsets " << _numClusters << endl;
  //cout << *_pt[groupNum] << endl;
  level = 2;
  do
    {
      _pt[groupNum]->join(numCands);
      _numClusters += numCands;
      cout << " Num new Cands " << numCands << endl;
      level++;
    } while ((numCands > 0) && (level < 3));

  cout << _numClusters << endl;
  //cout << *_pt[groupNum] << endl;
  set.Nullify();
}


void
SubspaceClusters::ValidateClusters(char *inputFile, ClusterColln& attrClusters)
{

  cout << "Starting second scan " << endl;
  bool subspace=FALSE;
  int i;
  
  for (i = 0; i < _numAttrGroups; i++)
    cout << _beginVals[i] <<" ";
  cout << endl;
  
#ifdef ASCII
    // Potentially very wasteful but we only need one buffer and set at a time.
    int strSize = MAX_TRANS_BYTES(numItems);
    char buf[strSize];
    
    xfile.getline(buf, strSize);

    while (xfile) {
      Set transaction(buf);
      xfile.getline(buf, strSize);
#else
      int xNumber = 0;
      char buffer[TRANS_BLOCK_SIZE];
      Set transaction(0);

      //ofstream ofile("/tmp/debug");
      //cout << inputFile << endl;

      SetupXactScan(inputFile, buffer);
      while (getNextXact(transaction, xNumber, buffer) != -1) 
	{
#endif  
	  for (i = 0; i < _numAttrGroups; i++)
	    {
	      //transaction.suffix(_beginVals[i]);
	      _pt[i]->IncrementClusterCounts(transaction, attrClusters, _beginVals[i],
					     subspace);
	      //transaction.suffix(-_beginVals[i]);
	    }
	}
      transaction.Nullify();

      //_pt->WeedOutNonClusters();

      cout << "End of Second Scan " << _numAttrGroups << endl;
      for (i = 0; i < _numAttrGroups; i++)
	cout << *(_pt[i]) << endl;
}








  /**** ThruClusters::ClusterAttrCactus2
  int i;
  SymTwoDArray& tempInfo = *(summary.AttrInfo(attr));
  double threshold = summary.Threshold(attr, thruAttr);

  for (i = summary.StartAttrVal(attr); i < summary.StartAttrVal(attr) + summary.DomSize(attr); i++)
    AssimilateNewAttrVal(i, tempInfo, threshold, thruAttr);

  //Repeat the process so that order independence is achieved.
  for (i=summary.StartAttrVal(attr); i<summary.StartAttrVal(attr)+summary.DomSize(attr);i++)
    AssimilateNewAttrVal(i, tempInfo, threshold, thruAttr);

  RemoveSmallClusters();

  // put these clusters in the prefix tree for further processing
  // to achieve completeness.
  ****/


    /****
double 
ThruAttrClusters::AvgStrength(int cluster, int attrVal, TwoDArray& attrPairInfo,
			      int attr, int thruAttr, int *siblingSet, int sibStr)
{
  int i, j;
  int attrVal2;
  double strength = 0.0;

  //cout << " AvgStrength : ClusterNumber " << cluster << "\t" << attrVal << endl;
  //for (i = 0; i < _nodeIds[cluster]->Size(); i++)
  //{
      //attrVal2 = _nodeIds[cluster]->Get(i);
      //cout << "\t" << attrVal2 << endl;
      for (j = 0; j < sibStr; j++)
	if (attr < thruAttr)
	  {
	    if (attrPairInfo.get(attrVal, siblingSet[j]) >= attrPairInfo.Threshold())
	      strength += 1.0;
	  }
	else
	  {
	    if (attrPairInfo.get(siblingSet[j], attrVal) >= attrPairInfo.Threshold())
	      strength += 1.0;
	  }
      //}
      //cout << "AvgStrength is " << (double)strength/((double)_nodeIds[cluster]->Size());
      //return (double)strength/((double)_nodeIds[cluster]->Size());
      return (double)strength/((double)sibStr);
}
     ***/
