#ifndef __CLUSTER_H__
#define __CLUSTER_H__

class SymTwoDArray;
class TwoDArray;
class TwoDArrayColln;
class SymTwoDArrayColln;
class SummarizedInfo;

class OneDArray
{
private:
  int _sz;
  int _maxSz;
  int *_arr;

public:
  OneDArray(int size, int maxSize)
    {
      _sz = size;
      _maxSz = maxSize;
      _arr = new int[_maxSz];
    }

  ~OneDArray()
    {
      delete [] _arr;
    }

  int Size() { return _sz; }

  bool Member(int attrVal)
    {
      int i;
      for (i = 0; i < _sz; i++)
	if (attrVal == _arr[i])
	  return TRUE;
      return FALSE;
    }

  int Get(int indx)
    {
      assert(indx < _sz);
      return _arr[indx];
    }

  void Set(int indx, int val)
    {
      assert(indx < _sz);
      _arr[indx] = val;
    }

  void Append(int val)
    {
      //cout << "Append " << _sz << " " << _maxSz << endl;
      assert(_sz < _maxSz);
      /*****
      if (FALSE == Member(val))
	{
	  _arr[_sz] = val;
	  _sz++;
	}
      ***/

      int insertPos=0, indx=0;
      if (0 == _sz)
	{
	  _arr[_sz] = val;
	  _sz++;
	}
      else
	{
	  for (insertPos = 0; insertPos < _sz && _arr[insertPos] < val; insertPos++);
	  if (_arr[insertPos] != val)
	    {
	      for (indx = _sz; indx > insertPos; indx--)
		_arr[indx] = _arr[indx-1];
	      _arr[insertPos] = val;
	      _sz++;
	    }
	}
    }

  void Delete(int attrVal)
    {
      int i;
      int indx = -1;

      for (i = 0; i < _sz; i++)
	{
	  if (_arr[i] == attrVal)
	    {
	      indx = i;
	      break;
	    }
	}
      if (-1 == indx)
	assert(0);
      assert(indx < _sz);

      for (i = indx; i < _sz-1; i++)
	_arr[i] = _arr[i+1];
      _sz--;

      assert(_sz > 0);
    }

  bool Equal(OneDArray& temp)
    {
      int i;
      if (Size() != temp.Size())
	return FALSE;

      for (i = 0; i < Size(); i++)
	{
	  if (Get(i) != temp.Get(i))
	    return FALSE;
	}
      return TRUE;
    }

  void Sort()
    {
    }

};


/*
 *The data structures for maintaining clusters are as follows.
 *ThruAttrClusters: clusters on *attr* through *thruAttr*.
 *AttrClusters: clusters on a given *attr* through every other attribute.
 *ClusterColln: an array of AttrClusters on each attribute.
 *
 *SubspaceClusters: just a driver class for discovering clusters in subspaces
 */

class ThruAttrClusters
{
private:
  OneDArray **_nodeIds;
  int *_clusStrengths;

  //SymTwoDArray **_shortestPaths;
  //bool *_clusterTypes;
  int _numClusters;
  int _numAttrs;
  int _maxNumClusters;
  
  /*
   *summary consists of all inter-attribute and intra-attribute similarities.
   *distSetSz is the distinguishing set size
   *The following function computes the set of clusters on *attr* through 
   *the *thruAttr* (required information is in summary)
   */
  void _CompleteThruAttrClusters(SummarizedInfo& summary, int attr, 
				 int thruAttr, int distSetSz);

  /*
   *Given a set of distinguishing sets in **ptArr, each of them is extended 
   *to a cluster projection. In addition, all subsets of the cluster projection
   *are also marked and will not be extended further. 
   *the extended cluster projections are stored in **_nodeIds in the sorted order
   */
  void _FindClustersFromDistSets(SummarizedInfo& summary, PrefixTree **ptArr, 
				 int numPTs, int attr, int thruAttr);

public:
  ThruAttrClusters(int numC=0);
  ~ThruAttrClusters()
    {
      int i;
      assert(_numClusters <= _maxNumClusters);

      for (i = 0; i < _maxNumClusters; i++)
	if (NULL != _nodeIds[i]) delete _nodeIds[i];

      delete [] _nodeIds;
      delete [] _clusStrengths;
    }

  /*Some utilitarian functions*/
  int NumClusters() { return _numClusters;}
  int Size(int i) { return _nodeIds[i]->Size();}
  int ClusStrength(int i) { return _clusStrengths[i];}
  int AttrVal(int clus, int indx) { return _nodeIds[clus]->Get(indx);}

  /*
   *returns the cluster with "indx" in the equivalent set representation.
   */
  void ReturnClusterSet(int indx, Set& set)
    {
      int i;
      bool flag;
      for (i = 0; i < _nodeIds[indx]->Size(); i++)
	{
	  flag = set.insert(_nodeIds[indx]->Get(i));
	  assert(TRUE == flag);
	}
    }

  /*
   *Extend the distinguishing set *distSet* to its cluster projection;
   *summary consists of the inter and intra-attribute summaries.
   *strength is the sibling strength of the distSet
   */   
  int ExtendDistSet(SummarizedInfo& summary, int attr, int thruAttr, Set& distSet,
		    int& strength);
  /*
   *find cluster projections on *attr* through the *thruAttr*
   */
  void ClusterAttrCactus2(SummarizedInfo& attrInfo, int attr, int thruAttr,
			  int distSetSz);

  //double AvgStrength(int cluster, int attrVal, SymTwoDArray& attrInfo, int thruAttr);
  /*
   *average of the strength of *attrVal* with the *siblingSet*
   *information is available in attrPairInfo
   */
  double AvgStrength(int attrVal, TwoDArray& attrPairInfo,
		     int attr, int thruAttr, int *siblingSet, int sibStr);



  /*
   *result consists of the ``current'' set of clusters and then these are 
   *intersected with clusters in *this; Finally, result consists of the 
   *intersection join of *result and *this
   */
  void IntersectClusters(ThruAttrClusters *result);

  //Not sure whether we need the following two functions any more. CHECK;
  void AssimilateNewAttrVal(int attrVal, SymTwoDArray& attrInfo, double threshold, 
			    int thruAttr);
  //Post processing functions;
  void RemoveSmallClusters();
  // AttrClusters *attrClus;



  bool Member(int attrVal, int cluster)
    {
      return (_nodeIds[cluster]->Member(attrVal));
    }

  /*
   *Return the cluster and set the corresponding _nodeIds pointer to NULL;
   *useful for merging clusters;
   */
  OneDArray* ReturnClusterSetNull(int indx)
    {
      OneDArray* temp = _nodeIds[indx];
      _nodeIds[indx] = NULL;
      return temp;
    }

  bool AddNewCluster(OneDArray* newCluster)
    {
      int i;
      for (i = 0; i < _numClusters; i++)
	if (TRUE == _nodeIds[i]->Equal(*newCluster))
	  return FALSE;

      assert(NULL == _nodeIds[_numClusters]);
      _nodeIds[_numClusters] = newCluster;
      _numClusters++;
      assert(_numClusters < _maxNumClusters);
      return TRUE;
    }

  //void AdjustClusters(int attr, SummarizedInfo& info);
  //bool GrowCluster(int attrVal, int cluster, SymTwoDArray& attrInfo);
  //bool CropCluster(int attrVal, int cluster, SymTwoDArray& attrInfo);

  
  friend ofstream& operator<<(ofstream& of, ThruAttrClusters& cluster);
  friend ostream& operator<<(ostream& of, ThruAttrClusters& cluster);
};


class AttrClusters
{
private:
  int _numAttrs;
  int _attr;
  ThruAttrClusters **_thruClusters;

  int _baseClusterNum;
  int _primaryThruAttr;

public:
  AttrClusters(int attr, int numAttrs, int numC);
  ~AttrClusters()
    {
      int i;
      for (i = 0; i < _numAttrs+1; i++)
	if ((i != _attr) && (NULL != _thruClusters[i]))
	  delete _thruClusters[i];
      delete [] _thruClusters;
    }

  /*
   *why is this necessary? some funda about naming the cluster indices.
   *only for subspaces?
   */
  void SetBaseClusterNum(int startVal) 
    { 
      _baseClusterNum = startVal;
      _primaryThruAttr = _numAttrs;
    }
  int BaseClusterNum()
    {
      assert(_baseClusterNum != -1);
      return _baseClusterNum;
    }

  //The following function just calls the corr. _thruClusters[_numAttrs]
  //assuming that the intersectClusters has already been called.
  int NumClusters() 
    { 
#ifndef SUBSPACE
      assert(_thruClusters[_numAttrs] != NULL);
#endif //revisit and implement the SUBSPACE option too.
      return _thruClusters[_numAttrs]->NumClusters();
    }

  int Size(int clus)
    {
      assert(_thruClusters[_numAttrs] != NULL);
      return _thruClusters[_numAttrs]->Size(clus);
    }

  int AttrVal(int clus, int indx)
    {
      return (_thruClusters[_numAttrs]->AttrVal(clus,indx));
    }

  bool Member(int attrVal, int clusNum)
    {
      //ideally, should use the intersected set of clusters in _thruClusters[_numAttrs];
      //However, making do with something for now. 
      return (_thruClusters[_numAttrs]->Member(attrVal, clusNum));      
    }


  void ClusterAttrCactus2(SummarizedInfo& summary, int distSetSz);

  /*
   *Finds the intersections of clusters obtained thru all attributes;
   *For each cluster projection through any attribute, reduce it to the 
   *barest minimum and insert it into the set of cluster projections if
   *it is already not there.
   */
  void IntersectClusters();


  friend ofstream& operator<<(ofstream& of, AttrClusters& cluster);
  friend ostream& operator<<(ostream& of, AttrClusters& cluster);
};

class ClusterColln
{
private:
  int _numAttrs;
  AttrClusters **_attrClusters;

public:
  ClusterColln(int numA);
  ~ClusterColln()
    {
      int i;
      for (i = 0; i < _numAttrs; i++)
	if (NULL != _attrClusters[i])
	  delete _attrClusters[i];
      delete [] _attrClusters;
    }

  /*
   *All the cluster projections on *attr* through every other attribute are 
   *uniquely numbered for subspace clustering.
   *SetBaseClusNums() marks the clusters with unique ids.
   *cluster projections on *attr* through any attribute are consecutively numbered.
   */
  void SetBaseClusNums()
    {
      int i;
      int start = 0;
      
      cout << "Base Cluster Numbers " << endl;
      for (i = 0; i < _numAttrs; i++)
	{
	  cout << start << " ";
	  _attrClusters[i]->SetBaseClusterNum(start);
	  start += _attrClusters[i]->NumClusters();
	}
      cout << endl;
    }

  int Base(int attr) 
    {
      int result = _attrClusters[attr]->BaseClusterNum();
      assert(result != -1);
      return result;
    }

  int TotNum1DClusters(int attr) 
    {
      int i;
      int start=0;
      for (i = 0; i <= attr; i++)
	start += _attrClusters[i]->NumClusters();
      return start;
    }
  
  

  int NumAttrs() { return _numAttrs;}

  void InitAttrClusters(int attr);
  //void CreateClustersForAttr(int attr, int numC);
  AttrClusters* GetAttrClusters(int attr) {return _attrClusters[attr];}


  // PostProcessing Functions.
  /*
   *Number of connections between Cluster1 (number) on attr1 with Cluster2 on attr2;
   */
  double ConnectionStrength(int attr1, int cluster1, int attr2, int cluster2,
			    TwoDArrayColln& attrPairsInfo);

  /*
   *Is attrVal a member of the cluster indexed by clusterNum?
   */
  bool Member(int attr, int attrVal, int clusterNum)
    {
      if (clusterNum < Base(attr))
	return FALSE;
      if ((attr < _numAttrs-1) && (clusterNum >= Base(attr+1)))
	return FALSE;
      return _attrClusters[attr]->Member(attrVal, clusterNum-Base(attr));
    }

  /*
   *I don't think this function is being called from any function.
   */
  bool MemberExpensive(int attrVal, int clusterNum)
    {
      int i = 0; 
      for (i = 0; i < _numAttrs-1; i++)
	{
	  if (Base(i) > clusterNum)
	    break;
	}

      //cout << "MemExp " << attrVal << " " << clusterNum << endl;

      if (0 == i)
	return _attrClusters[i]->Member(attrVal, clusterNum);
      else if (i == _numAttrs-1)
	return _attrClusters[i]->Member(attrVal, clusterNum-Base(i));
      return _attrClusters[i-1]->Member(attrVal, clusterNum-Base(i-1));
    }

  /***

      start += _attrClusters[i]->NumClusters();
      if (start >= clusterNum)
	break;
      if (i < _numAttrs)
	if (0 == i)
	  return _attrClusters[i]->Member(attrVal, clusterNum);
	else 

      return FALSE;

  void AdjustClusters(int attr, SummarizedInfo& attrInfo)
    {
      assert(_attrClusters[attr] != NULL);
      _attrClusters[attr]->AdjustClusters(attr, attrInfo);
    }

  // to be done.
  //void FindSubSpaceClusters();
  ***/

  friend ostream& operator<<(ostream& of, ClusterColln& clusterCollns);
};

class SubspaceClusters
{
private:
  int _numAttrs;
  int _numAttrGroups;
  int *_beginVals;
  int _numClusters;
  //PrefixTree *_pt;
  PrefixTree **_pt;
public:
  SubspaceClusters(int numAtts, ClusterColln& attrClusters);
  ~SubspaceClusters();

  /*
   *Since I have not implemented the MDL or some other pruning to restrict
   *the possible number of candidates, I split the number of attributes into
   *groups such that each group consists of not more than a maximum number of groups.
   *"DriverForGroup" then finds clusters in subspaces in each group. To find clusters on 
   *all attributes, these clusters can be merged (in Driver).
   *However, a full-fledged subspace clustering requires the implementation of MDL
   *pruning.
   */
  void Driver(ClusterColln& attrClusters, TwoDArrayColln& attrPairsInfo,
	      char *inputFile);

  void DriverForGroup(ClusterColln& attrClusters, TwoDArrayColln& attrPairsInfo,
		      int groupNum);

  /*
   *Each candidate cluster has to be validated to check if it has enough support.
   */
  void ValidateClusters(char *inputFile, ClusterColln& attrClusters);
};

#endif
