#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include "node.h"
#include "2DArray.h"
#include "Set.h"
#include "trans_utils.h"
#include "cluster.h"


ostream& operator<<(ostream& of, SymTwoDArray& obj)
{
  of << "Dimension " << obj._dimn << " " << obj._begin << "  " << obj._end << endl;
  int j, i;

  for (i = obj._begin; i <= obj._end; i++)
    for (j = i; j <= obj._end; j++)
      of << i << " " << j << " " << obj.get(i,j) << endl;

  of << endl;
  return of;
}



TwoDArrayColln::TwoDArrayColln(int numTuples, int nA, int *domSizes)
{
  _numAttrs = nA;
  int i, j;
  int *startEndVals = new int[_numAttrs+1];
  int sum = 0;

  for (i = 0; i < _numAttrs; i++)
    {
      startEndVals[i] = sum;
      sum += domSizes[i];
    }
  startEndVals[_numAttrs] = sum;

  cout << " StartEndVals ::: ";
  for (i = 0; i < _numAttrs+1; i++)
    cout << startEndVals[i] << " " ;
  cout << endl << endl;

  _indexToAttrPairs = new (TwoDArray*)[_numAttrs * (_numAttrs-1)/2];
  for (i = 0; i < _numAttrs; i++)
    for (j = i+1; j < _numAttrs; j++)
      {
	_indexToAttrPairs[_index(i, j)] = new TwoDArray(domSizes[i], domSizes[j], startEndVals[i], startEndVals[i+1]-1, startEndVals[j], startEndVals[j+1]-1);
	_indexToAttrPairs[_index(i,j)]->SetThreshold((int)((double)(domSizes[i]*domSizes[j])/numTuples));
      }

  _attrValCounts = new int[sum];
  memset((void*)_attrValCounts, 0, sizeof(int)*sum);

  delete [] startEndVals;
}

// assuming here that the tuple has all k-fields;
// also that the entries in the first fields are
// smaller than the second and so on...
void TwoDArrayColln::IncrementCounts(Set& tuple)
{
  int sz = tuple.size();
  int i,j;

  for (i = 0; i < sz; i++)
    {
      _attrValCounts[tuple.elements(i)]++;
      for (j = i+1; j < sz; j++)
	_indexToAttrPairs[_index(i,j)]->increment(tuple.elements(i), tuple.elements(j));
    }
}

// Self-join on the reln (attr1, attr2)and then 
// group by each pair of values in the same column;
// 
void TwoDArrayColln::Join(int attr1,int attr2,bool First,SymTwoDArray& counts)
{
  int begin = counts.Begin();
  int end = counts.End();
  int i, j, k;

  assert(attr1 < attr2);
  TwoDArray& temp = *_indexToAttrPairs[_index(attr1, attr2)];
  
  // attr1 corresponds to the counts array;
  // or, similarity for values on attr1 using attr2;
  if (TRUE == First)
    {
      for (k = temp.Begin2(); k <= temp.End2(); k++)
	for (i = begin; i <= end; i++)
	  {
	    if (temp.get(i,k) > temp.Threshold())
	      for (j = i+1; j <= end; j++)
		{
		  counts.increment(i, j, temp.get(i,k)*temp.get(j,k));
		}
	    counts.set(i,i,_attrValCounts[i]*_attrValCounts[i]);
	    //counts.set(i,i,_attrValCounts[i]);
	  }
    }
  else
    {
      for (k = temp.Begin1(); k <= temp.End1(); k++)
	for (i = begin; i <= end; i++)
	  {
	    if (temp.get(k,i) > temp.Threshold())
	      for (j = i+1; j <= end; j++)
		{
		  counts.increment(i, j, temp.get(k,i)*temp.get(k,j));
		}
	    counts.set(i,i,_attrValCounts[i]*_attrValCounts[i]);
	  }
    }

  /***  
  ofstream ofile("/tmp/link-strengths");
  ofile << "After join of " << attr1 << " , " << attr2 << endl;
  ofile << counts; 
  ofile.close();
  ****/
}
