#ifndef __TWODARR_H__
#define __TWODARR_H__

#include <assert.h>
#include <string.h>

class TwoDArray 
{
private:
  int _dimn1;
  int _dimn2;
  int _begin1;
  int _begin2;
  int _end1;
  int _end2;
  int *_array;

  int _threshold;

  int _index(int i, int j)
    {
      return ((i-_begin1)*_dimn2 + (j-_begin2));
    }
public:
  TwoDArray(int d1, int d2, int b1, int e1, int b2, int e2) 
    {
      _dimn1 = d1;
      _dimn2 = d2;
      _begin1 = b1;
      _begin2 = b2;
      _end1 = e1;
      _end2 = e2;
      _array = new int[_dimn1 * _dimn2];
      cout << "TwoDArray::TwoDArray " << endl;
      cout << _dimn1 << " " << _dimn2 << " " << _begin1 << " " << _end1 << " " << _begin2 << " " << _begin2 << endl;
      memset((void*)_array, 0, _dimn1*_dimn2*sizeof(int));
      assert(_array != NULL);
      
      _threshold = 0;
    }

  ~TwoDArray()
    {
      delete [] _array;
    }


  int Begin1() { return _begin1;}
  int Begin2() { return _begin2;}
  int End1() { return _end1;}
  int End2() { return _end2;}

  int Threshold() { return _threshold;}
  void SetThreshold(int val) 
    {
      cout << "Setting Threshold : " << val << " for " << _begin1 << " " << _end1 << endl;
      _threshold = val;
    }

  int get(int i, int j)
    {
      //cout << i << " " << _begin1 << " " << _dimn1 << " ";
      assert((i-_begin1) < _dimn1);
      assert((j-_begin2) < _dimn2);
      return _array[_index(i,j)];
    }
  void set(int i, int j, int val)
    {
      _array[_index(i,j)] = val;
    }
  void increment(int i, int j)
    {
      //cout << i << " " << j << " " << _dimn1 << " " << _dimn2 << endl;
      assert((i-_begin1) < _dimn1);
      assert((j-_begin2) < _dimn2);
      _array[_index(i,j)] += 1;
    }
};

// an array for storing num links between
// all pairs (a1, a2) of the same attribute.
class SymTwoDArray
{
private:
  int _dimn;
  int _begin;
  int _end;
  int *_array;

  int _index(int i, int j) 
    {
      assert(i <= j);
      return ((i-_begin)*_dimn  - (i-_begin)*((i-_begin)-1)/2 + j-i);
    }
  int _size() { return _dimn * (_dimn+1)/2; }

public:
  SymTwoDArray(int d, int b, int e) 
    {
      _dimn = d;
      _begin = b;
      _end = e;
      _array = new int[_size()];
      memset((void*)_array, 0, _size()*sizeof(int));
      assert(_array != NULL);
    }

  ~SymTwoDArray()
    {
      delete [] _array;
    }

  int Begin() { return _begin;}
  int End() { return _end;}
  int Dimn() { return _dimn;}

  // i < j;
  inline int get_fast(int i, int j)
    {
      assert(i <= j);
      assert(j-_begin < _dimn);
      return _array[_index(i,j)];
    }

  // not necessary that i < j;
  inline int get(int i, int j)
    {
      //assert (i != j);
      return ((i <= j) ? get_fast(i, j) : get_fast(j,i));
    }

  void set(int i, int j, int val)
    {
      assert(i <= j);
      assert(j-_begin < _dimn);
      _array[_index(i,j)] = val;
    }

  void increment(int i, int j, int ct=1)
    {
      assert(i < j);
      _array[_index(i,j)] += ct;
    }
  
#ifdef DEBUG
friend ostream& operator<<(ostream& of, SymTwoDArray& obj);
#endif  

};


class TwoDArrayColln
{
private:
  int _numAttrs;
  TwoDArray  **_indexToAttrPairs;
  int *_attrValCounts;

  int _index(int i, int j)
    {
      assert(i < j);
      assert(j < _numAttrs);
      return (i*_numAttrs - i*(i+1)/2 + j+i-1);
    }

  int _min(int val1, int val2) {return (val1 < val2) ? val1 : val2; }

public:
  TwoDArrayColln(int nT, int nA, int *domSizes);
  ~TwoDArrayColln()
    {
      delete [] _indexToAttrPairs;
      delete [] _attrValCounts;
    }
  
  void IncrementCounts(Set& tuple);
  void Join(int attr1, int attr2, bool First, SymTwoDArray& counts);
};

#endif






