#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include "node.h"
#include "2DArray.h"
#include "Set.h"
#include "trans_utils.h"
#include "cluster.h"

void OutputSortedAttrVals(int *sortedAttrVals, int *breaks, Graph& graph, int attr, 
			  int basin);

int count1and2links(   char *inputFile,
		       TwoDArrayColln& countColln
#ifdef ASCII
		       ,int numItems, int minSupport
#endif
		       )
{
  int large2Itemsets = 0;
  
#ifdef SAMPLING
  for (int index = 0; index < sampleSize; index++) {
    Set transaction(*sample[index]);
#elif defined(ASCII)
    // Potentially very wasteful but we only need one buffer and set at a time.
    int strSize = MAX_TRANS_BYTES(numItems);
    char buf[strSize];
    
    xfile.getline(buf, strSize);

    while (xfile) {
      Set transaction(buf);
      xfile.getline(buf, strSize);
#else
      int xNumber = 0;
      char buffer[TRANS_BLOCK_SIZE];
      Set transaction(0);

      ofstream ofile("/tmp/debug");

      SetupXactScan(inputFile, buffer);
      while (getNextXact(transaction, xNumber, buffer) != -1) 
	{
#endif  // SAMPLING
	  ofile << transaction << endl;
	  countColln.IncrementCounts(transaction);
	}
      transaction.Nullify();
      ofile.close();
      return large2Itemsets;
}


TwoDArrayColln* PreProcess(char *inputFile, int numTuples, int numAttrs, 
			   int *domSizes)
{
  TwoDArrayColln* countColln = new TwoDArrayColln(numTuples, numAttrs, domSizes);
  count1and2links(inputFile, *countColln);
  return countColln;
}



void Multiply(SymTwoDArray& attrInfo, Graph& graph1, Graph& graph2, 
	      int attr)
{
  int dimn = attrInfo.Dimn();
  int basins = graph1.NumBasins(attr);
  assert(dimn == graph1.DomainSz(attr));

  int i, j, b;
  double val = 0.0;
  int begin = attrInfo.Begin();
  int end = attrInfo.End();

  for (b = 0; b < basins; b++)
    {
      //      cout << " Basin " << b << endl;
      for (i = begin; i <= end; i++)
	{			
	  val = 0.0;
	  for (j = begin; j < i; j++)
	  //for (j = begin; j <= i; j++)
	    {
	      val += ((double)(attrInfo.get_fast(j,i)))*graph1.GetWt(j,attr,b);
	    }
	  for (j = i+1; j <= end; j++)
	    {
	      val += ((double)(attrInfo.get_fast(i,j)))*graph1.GetWt(j,attr,b);
	    }
	  graph2.AddWtToNode(i, attr, b, val);
	}
    }
}

// cout << j << " " << i << " " << attrInfo.get_fast(j,i) << " " << graph1.GetWt(j,attr,b) << endl;
// cout << attrInfo.get_fast(i,j) << " " << graph1.GetWt(j,attr,b) << endl;
//cout << val << endl;

void ClusterAttr(SymTwoDArray& attrInfo, Graph& graph1, Graph& graph2, 
		 int attr, int numIterns)
{
  int i;
  for (i = 0; i < numIterns; i++)
    {
      Multiply(attrInfo, graph1, graph2, attr);
      graph2.OrthoNormalizeAttr(attr);
      graph1.AssignAttr(graph2, attr); 
      //graph1.TestOrthoNormality(attr);
    }
  //cout << graph1 << endl;
}

void PostProcess(TwoDArrayColln& countColln, SymTwoDArray& attrInfo,
		 Graph& graph, int attr)
{
  int *sortedAttrVals = new int[graph.DomainSz(attr)];
  int *breaks = new int[graph.DomainSz(attr)];
  int numPartns;

  graph.Sort(attr, NON_PRINC_BASIN, sortedAttrVals);
  numPartns = graph.Partition(attr, NON_PRINC_BASIN, sortedAttrVals, breaks);

#ifdef DEBUG
  OutputSortedAttrVals(sortedAttrVals, breaks, graph, attr, NON_PRINC_BASIN);
#endif

  ClusterColln *clusters = new ClusterColln(graph.NumAttrs());
  clusters->CreateClustersForAttr(attr, numPartns, sortedAttrVals, breaks);

  delete [] sortedAttrVals;
  delete [] breaks;

  clusters->AdjustClusters(attr, graph, attrInfo);
  
  cout << *(clusters->GetAttrClusters(attr));
}


void OutputSortedAttrVals(int *sortedAttrVals, int *breaks, Graph& graph, int attr, 
			  int basin)
{
  int i;
  int breakNum = 0;
  cout << "Partition number 0" << endl;
  for (i = 0; i < graph.DomainSz(attr); i++)
    {
      if (i == breaks[breakNum])
	{
	  cout << endl << "Partition number " << breakNum+1 << endl;
	  breakNum++;
	}
      cout << sortedAttrVals[i] << " : " << graph.GetWt(sortedAttrVals[i], attr, basin) << endl; 
    }
}


void Algorithm(char *inputFile, int numTuples, int numAttrs, int *domSizes,
	       Graph& graph1, Graph& graph2, int numIterns)
{
  int i, j;
  int sum=0;

  TwoDArrayColln *countColln;
  countColln = PreProcess(inputFile, numTuples, numAttrs, domSizes);

  for (i = 0; i < 1; i++)
    //for (i = 0; i < numAttrs; i++)
    {
      SymTwoDArray attrInfo(domSizes[i], sum, sum+domSizes[i]-1);

      cout << "Created Sym Two D array^^^^^^ " << sum << " " << sum+domSizes[i]-1 << endl;
      for (j = 0; j < i; j++)
	countColln->Join(j, i, FALSE, attrInfo);
      for (j = i+1; j < numAttrs; j++)
	countColln->Join(i, j, TRUE, attrInfo);

      cout << "Joined attributes&&&&&& " << endl;
      ClusterAttr(attrInfo, graph1, graph2, i, numIterns);
      
      PostProcess(*countColln, attrInfo, graph1, i);
      sum += domSizes[i];
    }

}


