#include <iostream.h>
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include "Set.h"
#include "globals.h"
#include "globals_catClus.h"
#include "trans_utils.h"
#include "node.h"
#include "cluster.h"
#include "2DArray.h"

AttrClusters::AttrClusters(int numC, int *sortedIndices, int *breaks, 
			   double minS=0.0)
{
  _numClusters = numC;
  _maxNumClusters = 2*_numClusters;
  _minStrength = minS;

  int i, j;
  int low, high;

  _nodeIds = new (OneDArray*)[_maxNumClusters];

  for (i = 0; i < _numClusters; i++)
    {
      if (0 == i)
	{
	  low = 0;
	  high = breaks[0];
	}
      else
	{
	  low = breaks[i-1];
	  high = breaks[i];
	}
      _nodeIds[i] = new OneDArray(high-low, 2*(high-low));
      for (j = low; j < high; j++)
	{
	  _nodeIds[i]->Set(j-low, sortedIndices[j]);
	}
    }
}

double AttrClusters::AvgStrength(int cluster, int attrVal, SymTwoDArray& attrInfo)
{
  int i;
  int attrVal2;
  int strength = 0;

  for (i = 0; i < _nodeIds[cluster]->Size(); i++)
    {
      attrVal2 = _nodeIds[cluster]->Get(i);
      strength += attrInfo.get(attrVal, attrVal2);
    }
  return (double)strength/((double)_nodeIds[cluster]->Size());
}

bool AttrClusters::GrowCluster(int attrVal, int cluster, SymTwoDArray& attrInfo)
{
  double strength = AvgStrength(attrVal, cluster, attrInfo);

  if (strength >= _minStrength)
    {
      _nodeIds[cluster]->Append(attrVal);
      return TRUE;
    }
  return FALSE;
}


bool AttrClusters::CropCluster(int attrVal, int cluster, SymTwoDArray& attrInfo)
{
  double strength = AvgStrength(attrVal, cluster, attrInfo);
  if (strength < _minStrength)
    {
      _nodeIds[cluster]->Delete(attrVal);
      return TRUE;
    }
  return FALSE;
}


void AttrClusters::AdjustClusters(int attr, Graph &graph,
				  SymTwoDArray& attrInfo)
{
  int i, j;
  int start = graph.StartAttrVal(attr);

  for (i = start; i < start + graph.DomainSz(attr); i++)
    {
      for (j = 0; j < NumClusters(); j++)
	{
	  if (FALSE == Member(i, j))
	    {
	      GrowCluster(i, j, attrInfo);
	    }
	  else
	    CropCluster(i, j, attrInfo);
	}
    }
}



ostream& operator<<(ostream& of, AttrClusters& clus)
{
  of << " Number of clusters = " << clus._numClusters << endl;
  
  int i, j;
  for (i = 0; i < clus._numClusters; i++)
    {
      of << "Cluster Number " << i << endl; 
      for (j = 0; j < clus._nodeIds[i]->Size(); j++)
	of << clus._nodeIds[i]->Get(j) << endl;
    }
  return of;
}




ClusterColln::ClusterColln(int numA)
{
  _numAttrs = numA;
  _attrClusters = new (AttrClusters*)[_numAttrs];
  int i;
  for (i = 0; i < _numAttrs; i++)
    _attrClusters[i] = NULL;
}

void ClusterColln::CreateClustersForAttr(int attr, int numPartns, int *sortedAttrVals,
					 int *breaks)
{
  assert(_attrClusters[attr] == NULL);
  _attrClusters[attr] = new AttrClusters(numPartns, sortedAttrVals, breaks);
}
