#ifndef __CLUSTER_H__
#define __CLUSTER_H__

class SymTwoDArray;

class OneDArray
{
private:
  int _sz;
  int _maxSz;
  int *_arr;

public:
  OneDArray(int size, int maxSize)
    {
      _sz = size;
      _maxSz = maxSize;
      _arr = new int[_maxSz];
    }
  ~OneDArray()
    {
      delete [] _arr;
    }

  int Size() { return _sz; }

  bool Member(int attrVal)
    {
      int i;
      for (i = 0; i < _sz; i++)
	if (attrVal == _arr[i])
	  return TRUE;
      return FALSE;
    }

  int Get(int indx)
    {
      assert(indx < _sz);
      return _arr[indx];
    }
  void Set(int indx, int val)
    {
      assert(indx < _sz);
      _arr[indx] = val;
    }

  void Append(int val)
    {
      assert(_sz < _maxSz);
      _arr[_sz] = val;
      _sz++;
    }

  void Delete(int attrVal)
    {
      int i;
      int indx = -1;

      for (i = 0; i < _sz; i++)
	{
	  if (_arr[i] == attrVal)
	    {
	      indx = i;
	      break;
	    }
	}
      if (-1 == indx)
	assert(0);
      assert(indx < _sz);

      for (i = indx; i < _sz-1; i++)
	_arr[i] = _arr[i+1];
      _sz--;

      assert(_sz > 0);
    }

};

class AttrClusters
{
private:
  OneDArray **_nodeIds;
  //bool *_clusterTypes;
  int _numClusters;
  int _maxNumClusters;
  double _minStrength;

public:

  AttrClusters(int numC, int *sortedIndices, int *breaks, double minS=0.0);
  ~AttrClusters()
    {
      int i;
      for (i = 0; i < _numClusters; i++)
	delete _nodeIds[i];
      delete [] _nodeIds;
      //delete [] _clusterTypes;
    }

  int NumClusters() { return _numClusters;}
  
  void SetMinStrength(double val) { _minStrength = val;}
  
  bool Member(int attrVal, int cluster)
    {
      return (_nodeIds[cluster]->Member(attrVal));
    }

  void AdjustClusters(int attr, Graph& graph, SymTwoDArray& attrInfo);

  bool GrowCluster(int attrVal, int cluster, SymTwoDArray& attrInfo);
  bool CropCluster(int attrVal, int cluster, SymTwoDArray& attrInfo);

  double AvgStrength(int cluster, int attrVal, SymTwoDArray& attrInfo);

  friend ofstream& operator<<(ofstream& of, AttrClusters& cluster);
  friend ostream& operator<<(ostream& of, AttrClusters& cluster);
};

class ClusterColln
{
private:
  int _numAttrs;
  AttrClusters **_attrClusters;

public:
  ClusterColln(int numA);
  ~ClusterColln()
    {
      int i;
      for (i = 0; i < _numAttrs; i++)
	if (NULL != _attrClusters[i])
	  delete _attrClusters[i];
      delete [] _attrClusters;
    }

  void CreateClustersForAttr(int attr, int numPartns, int *sortedAttrVals, 
			      int *breaks);
  AttrClusters* GetAttrClusters(int attr) { return _attrClusters[attr];}
  
  void AdjustClusters(int attr, Graph& graph, SymTwoDArray& attrInfo)
    {
      assert(_attrClusters[attr] != NULL);
      _attrClusters[attr]->AdjustClusters(attr, graph, attrInfo);
    }

  void FindSubSpaceClusters();
};
#endif
