#include <iostream.h>
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include "Set.h"
#include "globals.h"
#include "globals_catClus.h"
#include "trans_utils.h"
#include "cat_utils.h"
#include "node.h"
#include "PrefixTree.h"



Graph::Graph(int nA, int *domSz, int *basinSz, CombinationOp op=ADD)
{
  int i;
  _totNumAttrVals = 0;
  _totNumBasins = 0;
  
  _numAttrs = nA;
  _combineOp = op;
  _domSizes = new int[_numAttrs];
  _numBasins = new int[_numAttrs];
  _startAttrVals = new int[_numAttrs];
  _startNodeIndx = new int[_numAttrs];

  for (i = 0; i < _numAttrs; i++)
    {
      _domSizes[i] = domSz[i];
      _startAttrVals[i] = _totNumAttrVals;
      _totNumAttrVals += _domSizes[i];
      _numBasins[i] = basinSz[i];
      _totNumBasins += _numBasins[i];
    }

  cout << "StartNodeIndx Array : ";
  for (i = 0; i < _numAttrs; i++)
    {
      if (0 == i)
	_startNodeIndx[i] = 0;
      else
	_startNodeIndx[i] = _startNodeIndx[i-1] + _domSizes[i-1]*_numBasins[i-1];

      cout << _startNodeIndx[i] << "  ";
    }
  cout << endl;

  _numNodes = _startNodeIndx[_numAttrs-1] + _domSizes[_numAttrs-1] * _numBasins[_numAttrs-1];
  cout << "Creating " << _numNodes << " nodes" << endl;
  _nodes = new Node[_numNodes];
  for (i = 0; i < _numNodes; i++)
    _nodes[i].Init(i,0.0);
}


Graph::~Graph()
{
  delete [] _domSizes;
  delete [] _numBasins;
  delete [] _startAttrVals;
  delete [] _startNodeIndx;
  delete [] _nodes;
}


void Graph::Initialize(InitialState option)
{
  int i;

  switch(option)
    {
    case RANDOM:
      for (i = 0; i < _numNodes; i++)
	_nodes[i].SetWt(drand48());
      break;
    case UNIFORM:
      {
	for (i = 0; i < _numNodes; i++)
	  _nodes[i].SetWt(1.0);
	break;
      }
    case IDENTITY:
      for (i = 0; i < _numAttrs; i++)
	_InitializeIDbasins(i);
      break;
    default:
      assert(0);
      break;
    }
}

void Graph::_InitializeIDbasins(int attr)
{
  int basin=0;
  int i;
  
  for (basin = 0; basin < _numBasins[attr]; basin++)
    {
      for (i=StartNodeIndx2(attr, basin); i<StartNodeIndx2(attr,basin)+_domSizes[attr]; i++)
	_nodes[i].SetWt(0.0);
      
      _nodes[StartNodeIndx2(attr,basin)+basin].SetWt(1.0);
    }
}

void Graph::AssignAttr(Graph& graph, int attr)
{
  int i;
  int start = StartNodeIndx(attr);
  int numN = NumBasins(attr)*DomainSz(attr);

  for (i = start; i < start+numN; i++)
    _nodes[i].SetWt(graph.NodeWt(i));    
}

void Graph::operator=(Graph &graph)
{
  assert(_numAttrs == graph.NumAttrs());
  assert(_numNodes == graph.NumNodes());

  int i;
  for (i = 0; i < _numNodes; i++)
  {
    _nodes[i].SetWt(graph.NodeWt(i));    
  }
  // startVals etc;
}

void Graph::ClearWts()
{
  int i;
  for (i = 0; i < _numNodes; i++)
    _nodes[i].SetWt(0.0);
}

void Graph::IterateSTIRR(char *inputFile, Graph *graph2, PrefixTree *pt=NULL)
{
  char buffer[TRANS_BLOCK_SIZE];
  int tid = 0;

  graph2->ClearWts();

  if (NULL == pt)
    {
      tid = 0;
      Set xactIter(0);
      SetupXactScan(inputFile, buffer);
      while (getNextXact(xactIter, tid, buffer) != -1)
	{
	  UpdateSTIRR(xactIter, graph2, 1);
	}
      xactIter.Nullify();
    }
  else
    {
      //pt->UpdateGraph(this, graph2);
    }
  //cout << *graph2 << endl;
  graph2->OrthoNormalizeAllAttrs();
}

void Graph::UpdateSTIRR(Set& tuple, Graph *graph2, int count=1)
{
  int i, basin;
  double addlWt;
  int numbasins; // just to make sure all of them are same.

  assert(1 == count);

  numbasins = _numBasins[0];
  for (i = 0; i < tuple.size(); i++)
    {
      for (basin = 0; basin < numbasins; basin++)
	{
	  addlWt = _Combine(tuple, i, basin);
	  graph2->AddWtToNode(tuple.elements(i), i, basin, addlWt);
	}
    }
}


double Graph::_Norm(int attr, int basin)
{
  double norm = 0.0;
  int i;
  
  for (i = StartNodeIndx2(attr, basin); i < StartNodeIndx2(attr,basin) + _domSizes[attr]; i++)
    {
      norm += _nodes[i].Wt() * _nodes[i].Wt();
    }
  return sqrt(norm);
}



void Graph::_NormalizeAttr(int attr, int basin)
{
  int i;
  double norm = _Norm(attr, basin);
  //cout << *this << endl;
  //cout << "_NormalizeAttr : attr " << attr << " basin " << basin << " " << norm << endl;
  //assert(norm > 0);
  if (0 == norm)
    return;

  for (i = StartNodeIndx2(attr, basin); i < StartNodeIndx2(attr,basin) + _domSizes[attr]; i++)
    _nodes[i].SetWt(_nodes[i].Wt()/norm);
}





double Graph::_InnerProduct(int attr1, int basin1, int attr2, int basin2)
{
  int i;
  double prod = 0.0;
  assert(_domSizes[attr1] == _domSizes[attr2]);
  for (i = 0; i < _domSizes[attr1]; i++)
    prod += _nodes[StartNodeIndx2(attr1, basin1)+i].Wt()*_nodes[StartNodeIndx2(attr2,basin2)+i].Wt();
  return prod;
}

void Graph::OrthoNormalizeAttr(int attr)
{
  int i;
  _NormalizeAttr(attr, 0); // normalize the first basin;

  for (i = 1; i < _numBasins[attr]; i++)
    _NormalizeVec(attr, i);
}

// vec_attr -= (vec_0,vec_attr)vec_0 ... + (vec_(attr-1),vec_attr)vec_(attr-1)
void Graph::_NormalizeVec(int attr, int basin)
{
  int i;
  double dotProd;
  double norm1;
  //double norm2 = _Norm(attr, basin);
  
  //cout << "norm" << attr << ":" << basin << "=" << norm2 << endl;
  for (i = 0; i < basin; i++)
    {
      norm1 = _Norm(attr, basin);
      //cout << "norm" << attr << ":" << i << "=" << norm1 << endl;
      dotProd = _InnerProduct(attr, i, attr, basin);
      _ReduceFromVec(attr, basin, i, dotProd);
    }
  _NormalizeAttr(attr, basin);
}

//vec_basin1 -= factor*vec_basin2;
void Graph::_ReduceFromVec(int attr, int basin1, int basin2, double factor)
{
  int i;
  double val;

  for (i = 0; i < _domSizes[attr]; i++)
    {
      val = _nodes[StartNodeIndx2(attr, basin1)+i].Wt();
      //cout << val << ":";
      val -= factor*_nodes[StartNodeIndx2(attr, basin2)+i].Wt();
      //cout << val << "    ";
      _nodes[StartNodeIndx2(attr, basin1)+i].SetWt(val);
    }
  //  cout << endl;
}

void Graph::OrthoNormalizeAllAttrs()
{
  int i;
  for (i = 0; i < _numAttrs; i++)
    OrthoNormalizeAttr(i);
}

#ifdef DEBUG
void Graph::TestOrthoNormality(int attr)
{
  int i, j;

  cout << " Testing orthonormality for attr " << attr << endl;
  for (i = 0; i < _numBasins[attr]; i++)
    {
      cout << "norm for basin " << i << " " << _Norm(attr, i) << endl;
      for (j = 0; j < i; j++)
	cout << i << " (basin1.basin2)" << j << " " << _InnerProduct(attr, i, attr, j) << endl;
    }
}

void Graph::TestOrthoNormality()
{
  int i;
  for (i = 0; i < _numAttrs; i++)
    TestOrthoNormality(i);
}
#endif


ofstream& operator<<(ofstream& of, Graph& grph)
{
  int i, j, k, start;

  of << grph._numAttrs << " " << grph._numNodes << endl;
  for (i = 0; i < grph._numAttrs; i++)
    {
      of << grph._domSizes[i] << "\t";
    }
  of << endl;

  for (i = 0; i < grph._numAttrs; i++)
    {
      of << "Attribute " << i << endl;
      start = grph.StartNodeIndx2(i,0);
      for (j = 0; j < grph._domSizes[i]; j++)
	{
	  //start = grph.StartNodeIndx2(i,j);
	  of << start + j << " ";
	  for (k = 0; k < grph._numBasins[i]; k++)
	    of << grph._nodes[start + k*grph._domSizes[i] + j].Wt() << "\t";
	  of << endl;
	}
      of << endl;
    }
  return of;
}

ostream& operator<<(ostream& of, Graph& grph)
{
  int i, j, k, start;

  of << grph._numAttrs << " " << grph._numNodes << endl;
  for (i = 0; i < grph._numAttrs; i++)
    {
      of << grph._domSizes[i] << "\t";
    }
  of << endl;


  for (i = 0; i < grph._numAttrs; i++)
    {
      of << "Attribute " << i << endl;
      start = grph.StartNodeIndx2(i,0);
      for (j = 0; j < grph._domSizes[i]; j++)
	{
	  of << start + j << " ";
	  for (k = 0; k < grph._numBasins[i]; k++)
	    of << grph._nodes[start + k*grph._domSizes[i] + j].Wt() << "\t";
	  of << endl;
	}
      of << endl;
    }

  return of;
}

/***** Post-processing functions *********/


void Graph::Sort(int attr, int basinNum, int *sortedAttrVals)
{
  int i;
  int start = StartNodeIndx(attr);

  for (i = 0; i < DomainSz(attr); i++)
    sortedAttrVals[i] = start+i;

  SetupComparisonFn(attr, basinNum, this);
  qsort((void*)(sortedAttrVals), DomainSz(attr), sizeof(int), compareNodeWts);
}


// assume that space for breaks has already been allocated;
int Graph::Partition(int attr, int basinNum, int *sortedAttrVals, int *breaks)
{
  int numBreaks = 0;
  int i;
  double avgSeparation = 2.0/double(DomainSz(attr));
  double wt1, wt2;

  wt1 = GetWt(sortedAttrVals[0], attr, basinNum);
  for (i = 1; i < DomainSz(attr); i++)
    {
      wt2 = GetWt(sortedAttrVals[i], attr, basinNum);
      assert(wt2 >= wt1);
      if ((wt2-wt1) > ALPHA*avgSeparation)
	breaks[numBreaks++] = i;
      wt1 = wt2;
    }
  breaks[numBreaks] = StartNodeIndx(attr)+DomainSz(attr);
  
  cout << "Num partitions " << numBreaks+1 << endl;
  return numBreaks+1;
}
