/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef HASH_NODE_INLINE
#define HASH_NODE_INLINE

//#include "globals.h"

#ifdef DIC_ALG
bool HashNode::squareCount(Set &set, int minSupport)
{
  int bucket;

  bucket = HASH(a, b, h, set.elements(0));

  if (!children[bucket])
    return false;
  else
    return children[bucket]->squareCount(set, minSupport);
}
#endif


#ifndef DIC_ALG
int HashNode::pruneByCount(
#ifdef SAMPLING
			   PruneAction action,
#endif
			   int support, bool &allPruned)
{
	bool localAllPruned, globalAllPruned = true;
	int numPruned = 0;

	for (int i=0; i<h; i++)
	  if (children[i])
	  {
#ifdef SAMPLING
	    numPruned += children[i]->pruneByCount(action, support, localAllPruned);
#else
	    numPruned += children[i]->pruneByCount(support, localAllPruned);
#endif
	    globalAllPruned = globalAllPruned && localAllPruned;
		if (localAllPruned)
		{
			delete children[i];
			children[i] = 0;
		}
	  }

	allPruned = globalAllPruned;
	return numPruned;
}
#endif

int HashNode::incrementCounts(Set &transaction, int level
#ifndef DIC_ALG
					, int pass
#endif
#ifdef SAMPLING
			      ,int cnt, IncCountsAction action, int *arr, 
			      bool BlkCounted, TIDListColln *newBlk, int tid
#endif
				)
{

  int link, incCount = 0;

  int temp = transaction.size();
  for (int index = 0; index < temp; index++) {

    // (a * transaction.elements[index] + b) % h;
    link = HASH(a, b, h, transaction.elements(index));
    if (children[link]) {
      // The subtree is not empty

      transaction.suffix(index);
#ifdef SAMPLING
      incCount += children[link]->incrementCounts(transaction, level, pass, cnt, action,
						  arr, BlkCounted, newBlk, tid);
#elif !defined(DIC_ALG)
      incCount += children[link]->incrementCounts(transaction, level, pass);
#else
      incCount += children[link]->incrementCounts(transaction, level);
#endif
      transaction.suffix(-index);
    }
  }
  return incCount;
}

bool HashNode::insert(Set &set, int count 
#ifdef DIC_ALG
			, int startXaction,
			bool dashed,
			bool circle
#elif defined(SAMPLING)
		      , PrefixNodeStatus pnStatus = candidate
#endif
		      )
{
	assert(set.size() > 0);

	int bucket;

	bucket = HASH(a, b, h, set.elements(0));

	if (!children[bucket])
	  children[bucket] = new PrefixNodeArray(DEF_SETS_PER_LEAF);
#ifdef SAMPLING
	  children[bucket]->insert(set, count, pnStatus);
#elif defined(DIC_ALG)
	  children[bucket]->insert(set, count, startXaction, dashed, circle);
#else
	  children[bucket]->insert(set, count);
#endif

	return true;
}

bool HashNode::member(Set &toBeTested, int minSup)
{
	int bucket = HASH(a, b, h, toBeTested.elements(0));
	if (children[bucket] == 0)
		return false;
	else
		return children[bucket]->member(toBeTested, minSup);
}

#endif
