/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef PREFIX_NODE_ARRAY_INLINE
#define PREFIX_NODE_ARRAY_INLINE

#ifdef DIC_ALG
bool PrefixNodeArray::squareCount(Set &set, int minSupport)
{
  for (int i=0; i<numNodes; i++)
  {
    if (prefixNodes[i].item == set.elements(0))
    {
      if (set.size() == 1)
        return (prefixNodes[i].count >= minSupport && !prefixNodes[i].circle);
      else
      {
	if (prefixNodes[i].child)
	{
	  set.suffix(1);
          bool result=prefixNodes[i].child->squareCount(set, minSupport);
	  set.suffix(-1);
          return result;
	}
	else
	  return false;
      }
    }
  }
  return false;
}
#endif

#ifndef DIC_ALG
int PrefixNodeArray::pruneByCount(
#ifdef SAMPLING
			   PruneAction action,
#endif
			   int support, bool &allPruned)
{
	// Don't decrease maxSize 

	int numPruned = 0;
	for (int i = 0; i < numNodes; i++)
	{
		if ((prefixNodes[i].count < support))
		{
			numPruned++;
#ifdef SAMPLING
			// numPruned is incremented first so that the join loop can be
			// stopped earlier in apriori.  So numPruned reflects the number
			// of itemsets that were either pruned or marked as negBorder.
			switch (action) {
			case markNegBorder :
			  prefixNodes[i].status = negBorder;
			  continue;
			case markSmall :
			  prefixNodes[i].status = small;

			  // Not impossible for sampling.
			  if (prefixNodes[i].child)
			    delete prefixNodes[i].child;
			  prefixNodes[i].child = NULL;
			  continue;
			default :  // prune
			  // Not impossible for sampling.
			  if (prefixNodes[i].child)
			    delete prefixNodes[i].child;
			  prefixNodes[i].child = NULL;
			  // Now let this node be deleted below.
			  break;
			}
#endif

			if (prefixNodes[i].child)
			{
				assert(0); // shouldn't have child

				// destroy subtree rooted here entirely
				// delete prefixNodes[i].child; 
			}
			
			if (i < numNodes - 1)
			{
				// move the last node here, compact
				prefixNodes[i]=prefixNodes[numNodes-1];

				// check this newly moved here node
				i--;
			}

			numNodes--;
		}
		else
		{
		  prefixNodes[i].status = candidate; // VENKY

		  if (!prefixNodes[i].child)
		    continue;
		  bool childrenAllPruned;
                  numPruned +=
#ifdef SAMPLING
		    prefixNodes[i].child->pruneByCount(action, support, childrenAllPruned);
#else
		    prefixNodes[i].child->pruneByCount(support, childrenAllPruned);
#endif
                  if (childrenAllPruned)
                  {
                       delete prefixNodes[i].child;
                       prefixNodes[i].child = 0;
                   }
                }

	}
	if (numNodes == 0)
		allPruned = true;
	else
		allPruned = false;

	return numPruned;
}
#endif


/****
				     int *arr=NULL, bool SampleBlk=FALSE, 
				     TIDListColln *newBlk = NULL, int tid=-1
****/

// if BlkCounted then do not increment counts;
// if tid == -1 then do not write TID Lists;
int PrefixNodeArray::incrementCounts(Set &transactionRemains, int level
#ifndef DIC_ALG
					, int pass
#endif
#ifdef SAMPLING
				     ,int cnt, IncCountsAction action = passLevelOnly, 
				     int *arr=NULL, bool BlkCounted, 
				     TIDListColln *newBlk, int tid
#endif
				    )
{
  int incCount = 0;

  assert(level < MAX_IT_LENGTH);

  for (int index = 0; index < numNodes; index++)
    if (prefixNodes[index].item == transactionRemains.elements(0)) 
      {
#ifndef DIC_ALG
#ifdef SAMPLING
	if ((FALSE == BlkCounted) && (action == allLevels || pass == level))
#else
	if (pass == level)
#endif SAMPLING
	  prefixNodes[index].count += cnt;
#ifdef FANCYTID
	if (tid != -1)
	  {
	    assert(level > 1);
	    arr[level-1] = prefixNodes[index].item; // VENKY: for fancy tid's

	    //if ((FALSE == SampleBlk) && (prefixNodes[index].status == maximal))
	    if ((MATERIALIZED_LEVEL == level) &&
		((maximal == prefixNodes[index].status) || (candidate == prefixNodes[index].status)))
	      {
		if (NULL == prefixNodes[index].tidListPtr)
		  {
		    //cout << " Creating a new fancy tid list " << endl;
		    prefixNodes[index].tidListPtr = newBlk->CreateTIDListToWrite(arr,level);
		    assert(NULL != prefixNodes[index].tidListPtr);
		  }
		prefixNodes[index].tidListPtr->WriteTID(tid);
	      }
	  }
#endif	   
#else
      if (prefixNodes[index].dashed)
	prefixNodes[index].count += cnt;
#endif DIC_ALG

      if (prefixNodes[index].child) {
	// The subtree is not empty

	transactionRemains.suffix(1);
#ifdef SAMPLING
	incCount += prefixNodes[index].child->incrementCounts(transactionRemains, level+1, pass, cnt, action, arr, BlkCounted, newBlk, tid);
#elif !defined(DIC_ALG)
	incCount += prefixNodes[index].child->incrementCounts(transactionRemains, level+1, pass );
#else
	incCount += prefixNodes[index].child->incrementCounts(transactionRemains, level+1 );
#endif SAMPLING
	transactionRemains.suffix(-1);
      }
   }
  return incCount; 
}

bool PrefixNodeArray::insert (Set &set, int count
#ifdef DIC_ALG
		, int startTransaction,
		  bool dashed,
		  bool circle
#elif defined(SAMPLING)
			      , PrefixNodeStatus pnStatus = candidate
#endif
	    )
{
  int j;
  for (j=0; j<numNodes; j++)
    {
      if (prefixNodes[j].item == set.elements(0))
	{
	  if (set.size() == 1)
	    return 0;
	    //assert(0); //No reinsert! (Revisit: removed for CatClus)
	  else
	    {
	      set.suffix(1);
	      bool inserted;
	      if (prefixNodes[j].child == 0)
		prefixNodes[j].child = new HashNode;
	      
#ifdef SAMPLING
	      if (maximal == pnStatus)
		prefixNodes[j].status = candidate;
	      inserted = prefixNodes[j].child->insert(set, count, pnStatus);
#elif defined(DIC_ALG)
	      inserted = prefixNodes[j].child->insert(set, count, startTransaction, dashed, circle);
#else
	      inserted = prefixNodes[j].child->insert(set, count);
#endif
	      set.suffix(-1);
	      return inserted;
	    }
	}
    }
  
  // at this point we know the first item in set is not in the
  // array yet
  
  // a set can not be inserted unless all its prefix sets has been
  // inserted
  //#ifndef INSTRUMENT // we may insert candidates without parents;
  //#ifndef USUAL_DYNAMIC // when will this happen?
  //#ifndef DEVIATION  
#if !defined(INSTRUMENT) && !defined(USUAL_DYNAMIC) && !defined(DEVIATION) && !defined(CATCLUS)
    assert(set.size() == 1);//removed this assertion for DEVIATIONS;and for CATCLUS
#endif  
  
  int p; // position of the new element in the array
  
  if (numNodes == maxSize)
    {
      // need to grow the array

      //cout << " PNA:insert::Growing the prefix node array " << numNodes << endl;
      //cout << set << endl;
      PrefixNode *newNodes = new PrefixNode[maxSize * 2];
      if (!newNodes)
	{
	  cout << "Failed when new PrefixNode[maxSize * 2]\n";
	  exit(1);
	}
      for (int i=0; i<numNodes; i++)
	{
	  // shift by 1, reserve [0] for the newly added
	  newNodes[i+1] = prefixNodes[i];
	}
      delete[] prefixNodes;
      prefixNodes = newNodes;
      
      // insert as first since it'll probably be counted
      // more times from now on
      p = 0;
      maxSize = 2 * maxSize; // VENKY
    }
  else
    p = numNodes;
  
  numNodes++;
  
  prefixNodes[p].item = set.elements(0);
  prefixNodes[p].count = count;
  prefixNodes[p].child = 0;
  
#ifdef DIC_ALG
  prefixNodes[p].dashed = dashed;
  prefixNodes[p].startTransaction = startTransaction;
  prefixNodes[p].circle = circle;
#elif defined(SAMPLING)
  if (set.size() == 1)
    prefixNodes[p].status = pnStatus;
#ifdef CATCLUS
  else 
    prefixNodes[p].status = candidate;
#else
  else if (maximal == pnStatus)
    prefixNodes[p].status = candidate;
  else 
    prefixNodes[p].status = pnStatus;
#endif
#endif DIC_ALG

#if defined(DEVIATION) || defined(INSTRUMENT) || defined(CATCLUS)
  // we will insert candidates without any parents;
  if (set.size() > 1) // insert the set again
    {
      set.suffix(1);
      bool inserted;
      if (prefixNodes[p].child == 0)
	prefixNodes[p].child = new HashNode;
	      
#ifdef SAMPLING
      inserted = prefixNodes[p].child->insert(set, count, pnStatus);
#elif defined(DIC_ALG)
      inserted = prefixNodes[p].child->insert(set, count, startTransaction, dashed, circle);
#else
      inserted = prefixNodes[p].child->insert(set, count);
#endif
      set.suffix(-1);
      return inserted;
    }
#endif DEVIATION
  
  return true;
}


int PrefixNodeArray::DeletePrefixNode(int indx)
{
  
  assert((indx >= 0) && (indx < numNodes));

  // Should we write the counts of children here?

  if (prefixNodes[indx].child)
    delete prefixNodes[indx].child;
  prefixNodes[indx].status = negBorder;
  
  /*
  for (i = indx; i < numNodes-1; i++)
    {
       prefixNodes[i] = prefixNodes[i+1];
       //prefixNodes[i].item = prefixNodes[i+1].item;
       //prefixNodes[i].count = prefixNodes[i+1].count;
       //prefixNodes[i].child = prefixNodes[i+1].child;
       //prefixNodes[i].status = prefixNodes[i+1].status;
    }
  prefixNodes[numNodes].child = NULL;
  numNodes--;
  */
  prefixNodes[indx].child = NULL;
  return numNodes;
}
	
bool PrefixNodeArray::member(Set &toBeTested, int minSup)
{
	for (int i=0; i<numNodes; i++)
	  {
	    if (prefixNodes[i].item == toBeTested.elements(0))
	      {
		if ((minSup != EXISTENCE_ONLY) && (prefixNodes[i].count < minSup))
		  return false;
		
		if (toBeTested.size() == 1)
		  return true;

		if (NULL == prefixNodes[i].child) // not part of the tree.
		  return false;

		toBeTested.suffix(1);
		bool mem = prefixNodes[i].child->member(toBeTested, minSup);
		toBeTested.suffix(-1);
		return mem;
	      }
	  }
	return false;
}

#endif

