/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include "PrefixTree.h"

#include "globals.h"
#include "trans_utils.h"
#include "2DArray.h"
#include "Set.h"

#ifdef CATCLUS
#include "cluster.h"
#endif



/*************** For categorical clustering only *************/

#ifdef CATCLUS

#include "node.h"

void 
PrefixTree::UpdateGraphCC(Graph *oldGraph, Graph *newGraph)
{
  Item setArr[MAX_IT_LENGTH];
  int i;
  
  for (i = 0; i < numItems; i++)
    {
      setArr[0] = root[i].item;
      if ((negBorder == root[i].item) || (small == root[i].item))
	continue;
      if (NULL != root[i].child)
	{
	  root[i].child->UpdateGraphCC(oldGraph, newGraph, setArr, 2);
	}
      // else do nothing;
    }
}

void 
HashNode::UpdateGraphCC(Graph *oldGraph, Graph *newGraph, int *itemsetArr,
			int level)
{
  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->UpdateGraphCC(oldGraph, newGraph, itemsetArr, level);
    }
}

void 
PrefixNodeArray::UpdateGraphCC(Graph *oldGraph, Graph *newGraph, 
			       int *itemsetArr, int level)
{
  int i;
  Set tempSet(0);
  for (i = 0; i < numNodes; i++)
    {
      itemsetArr[level-1] = prefixNodes[i].item;
      if ((negBorder == prefixNodes[i].status)
	  || (small == prefixNodes[i].status))
	continue;

      if ((candidate == prefixNodes[i].status) || 
	  (maximal == prefixNodes[i].status))
	{
	  tempSet.newArr(level, itemsetArr);
	  oldGraph->UpdateSTIRR(tempSet, newGraph, prefixNodes[i].count);
	}

      if (maximal == prefixNodes[i].status)
	continue;
      if (NULL != prefixNodes[i].child)
	{
	  prefixNodes[i].child->UpdateGraphCC(oldGraph, newGraph, itemsetArr,
					      level+1);
	}
    }
  tempSet.Nullify();
}



PrefixTree*
PrefixTree::JoinCC(int& numNewCands, SummarizedInfo& summary, int attr, int thruAttr)
{
  int i;
  int prefix[MAX_IT_LENGTH];

  int numAttrVals = summary.DomSize(attr);
  PrefixTree *ptNew = new PrefixTree(summary.StartAttrVal(attr) + numAttrVals, 
				     NULL, 0);
  ptNew->setMinSupport(1); 
  //setting the counts for single items;
  for (i=summary.StartAttrVal(attr); i < summary.StartAttrVal(attr)+numAttrVals; i++)
    ptNew->SetItemCountStatus(i, 1, candidate);

  //cout << " Inside PT::JoinCC " << endl;

  for (i=summary.StartAttrVal(attr); i < summary.StartAttrVal(attr)+numAttrVals; i++)
    if ((root[i].count >= returnMinSupport()) && (NULL != root[i].child))
      {
	prefix[0] = root[i].item;
	root[i].child->JoinCC(numNewCands, prefix, 2, this, ptNew, summary,
			      attr, thruAttr); //level=2 for children;
      }
  return ptNew;
}

void
HashNode::JoinCC(int& numNewCands, int *prefix, int level, PrefixTree *ptOld,
		 PrefixTree *ptNew, SummarizedInfo& summary, int attr, int thruAttr)
{
  int i;
  
  //cout << "Inside HN::JoinCC" << endl;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->JoinCC(numNewCands, prefix, level, ptOld, ptNew, summary,
			      attr, thruAttr);
    }
}

void 
PrefixNodeArray::JoinCC(int& numNewCands, int *prefix, int level, PrefixTree *ptOld,
			PrefixTree *ptNew, SummarizedInfo& summary, int attr, 
			int thruAttr)
{
  int i,j;
  int numCands=0;
  int **cands = new (int*)[MAX_NUMCANDS];
  Set candSet(0);

  //cout << " Inside PNA::JoinCC" << endl;
  for (i = 0; i < numNodes; i++)
    {
      prefix[level-1] = prefixNodes[i].item;
      if (prefixNodes[i].status == candidate)
	{
	  if (prefixNodes[i].child != NULL)
	    prefixNodes[i].child->JoinCC(numNewCands, prefix, level+1, ptOld, ptNew, 
					 summary, attr, thruAttr);
	}
      else if (prefixNodes[i].status == negBorder)
	{
	  numCands = 0;
	  assert(NULL == prefixNodes[i].child);
	  //check if this itemset is already in pt.	  
	  //Otherwise, search for joinable itemsets
	  candSet.newArr(level, prefix);
	  if (!(ptNew->AlreadyInTree(candSet)))
	    {
	      ptOld->GenCandClustersCC(prefix, level, cands, numCands, 
				       summary, attr, thruAttr);
	      
	      assert(numCands <= MAX_NUMCANDS);
	      //cout << "Inserting new candidates " << endl;
	      for (j = 0; j < numCands; j++)
		{
		  candSet.newArr(level+1, cands[j]); // size(candSet) = level+1
		  //cout << "Inserting in new pt " << candSet << endl;
		  ptNew->insertWithCount(candSet, 1, negBorder);
		  candSet.Nullify();
		  delete [] cands[j];
		  cands[j] = NULL;
		}
	      numNewCands += numCands;
	    }
	  candSet.Nullify();
	  prefixNodes[i].status = candidate; // Revisit and check if correct;
	}
    }
  delete [] cands;
}


// returns candidate clusters only; the validity check is done
// during the search.
bool
//SubsetJoinCC(Set& set1, Set& set2)
SubsetJoinCC(int *set1, int sizeOfSet1, int *set2, int sizeOfSet2,
	     int& element1, int& element2)
{
  int i;
  bool yanked=FALSE, result=FALSE;
  int tmp, element;
  
  element1 = -1;
  element2 = -1;

  if (sizeOfSet1 != sizeOfSet2)
    return FALSE;

  Set candSet1(0);
  candSet1.newArr(sizeOfSet1, set1);
  Set candSet2(0);
  candSet2.newArr(sizeOfSet2, set2);

  if (candSet1 == candSet2)
    {
      candSet1.Nullify();
      candSet2.Nullify();
      return FALSE;
    }

  for (i = 0; i < sizeOfSet1; i++)
    {
      yanked = candSet1.yankOut(tmp, i);
      assert(yanked);
      element = candSet1.ImmediateSubset(candSet2);
      //cout << candSet1 << endl;
      //cout << candSet2 << endl;
      //cout << tmp << " " << element << endl;
      if (-1 != element)
	{
	  element1 = tmp;
	  element2 = element;
	  result = TRUE;
	  candSet1.putBack(tmp);
	  break;
	}
      candSet1.putBack(tmp);
    }

  //cout << element1 << " " << element2 << endl;
  candSet1.Nullify();
  candSet2.Nullify();
  return result;
}

void 
PrefixTree::GenCandClustersCC(int *set1, int sizeOfSet1, int **cands,
			      int& numCands,
			      SummarizedInfo& summary, int attr, int thruAttr)
{
  int set2[MAX_IT_LENGTH];
  int i;

  numCands = 0;
  //for (i = 0; i < numItems; i++)
  for (i = 0; i < set1[1]+1; i++) // a hack to speed it up; (1,5,10) can only 
    // join with something starting between 1 and 5 (inclusive);
    {
      set2[0] = i;
      if ((root[i].count >= returnMinSupport()) && (NULL != root[i].child))
	root[i].child->GenCandClustersCC(set1, sizeOfSet1, cands, numCands,
					 set2, 2, summary, attr, thruAttr);
    }
}


void
HashNode::GenCandClustersCC(int *set1, int sizeOfSet1, int **cands,
			    int& numCands,
			    int *set2, int level, SummarizedInfo& summary,
			    int attr, int thruAttr)
{
  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->GenCandClustersCC(set1, sizeOfSet1, cands, numCands,
					 set2, level, summary, attr, thruAttr);
    }
}


void
PrefixNodeArray::GenCandClustersCC(int *set1, int sizeOfSet1, int **cands, 
				   int& numCands,int* set2, int level, 
				   SummarizedInfo& summary, int attr, int thruAttr)
{
  int i, j;
  int iPos=0; //marks the position to insert element2 in set1;
  int element1=0, element2=0;
  bool joinResult=FALSE; //validateResult=FALSE;
  int validateResult=FALSE;

  //cout << "Entering PNA:: GenCandClustersCC " << numNodes << endl;
  for (i = 0; i < numNodes; i++)
    {
      set2[level-1] = prefixNodes[i].item;
      if (prefixNodes[i].status == candidate)
	{
	  if (prefixNodes[i].child != NULL)
	    prefixNodes[i].child->GenCandClustersCC(set1,sizeOfSet1,cands,
						    numCands,set2, level+1, summary, 
						    attr, thruAttr);
	}
      else if (prefixNodes[i].status == negBorder)
	{
	  assert(NULL == prefixNodes[i].child);
	  joinResult = SubsetJoinCC(set1, sizeOfSet1, set2, level, 
				    element1, element2);
	  if (TRUE == joinResult)
	    {
	      validateResult = summary.ValidateCluster(set1,sizeOfSet1, set2, level, 
						       element1, element2, attr, 
						       thruAttr);
	      if (TRUE == validateResult)
		{
		  cands[numCands] = new int[sizeOfSet1+1];
		  for (iPos=0; (iPos < sizeOfSet1)&&(element2 > set1[iPos]); iPos++)
		    {
		      (cands[numCands])[iPos] = set1[iPos];
		    }
		  (cands[numCands])[iPos] = element2;		  

		  for (j = iPos; j < sizeOfSet1; j++)
		    {
		      (cands[numCands])[j+1] = set1[j];
		    }

		  numCands++;
		  assert(numCands < MAX_NUMCANDS);
		  //cout << "GenCandClustersCC : " << numCands << endl;
		}
	    }
	}
    }
}

void
PrefixTree::ExtendDistSets(SummarizedInfo& summary, int attr, int thruAttr, 
			   ThruAttrClusters& clusterSet)
{
  int prefix[MAX_IT_LENGTH];
  //distSetSz should have been enough instead of MAX_IT_LENGTH;
  int i;
  
  for (i = 0; i < numItems; i++)
    if (NULL != root[i].child)
      {
	prefix[0] = i;
	root[i].child->ExtendDistSets(summary, attr, thruAttr, clusterSet,
				      prefix, 2, this);
      }
}

void
HashNode::ExtendDistSets(SummarizedInfo& summary, int attr, int thruAttr, 
			 ThruAttrClusters& clusterSet, int *prefix, int level,
			 PrefixTree *pt)
{
  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->ExtendDistSets(summary, attr, thruAttr, clusterSet,
				      prefix, level, pt);
    }
}


void 
PrefixNodeArray::ExtendDistSets(SummarizedInfo& summary, int attr, int thruAttr, 
				ThruAttrClusters& clusterSet, 
				int *prefix, int level,	PrefixTree *pt)
{
  int i;
  int strength=0;
  Set distSet(0);
  int newClustIndx;

  for (i = 0; i < numNodes; i++)
    {
      prefix[level-1] = prefixNodes[i].item;
      if (prefixNodes[i].child != NULL)
	prefixNodes[i].child->ExtendDistSets(summary, attr, thruAttr, clusterSet, 
					     prefix, level+1, pt);
      else if (prefixNodes[i].count == 0)
	{
	  distSet.newArr(level, prefix);
	  newClustIndx = clusterSet.ExtendDistSet(summary, attr, thruAttr, distSet,
						  strength);
	  distSet.Nullify();

	  if (newClustIndx >= 0)
	    {
	      Set newSet(clusterSet.Size(newClustIndx));
	      clusterSet.ReturnClusterSet(newClustIndx, newSet);
#ifdef DEBUG	      
	      cout << "Extended Set is " << newSet << endl << endl;
#endif
	      //mark distinguishing sets that are subsets of newSet;
	      pt->incrementCounts(newSet, strength, allLevels);
	    }
	}

    }
}


void
PrefixTree::IncrementClusterCounts(Set& xact, ClusterColln& attrClusters, 
				   int baseAttrVal, bool subspace)
{
  //int prefix[MAX_IT_LENGTH];
  int i,j;
  int level=baseAttrVal+1;

  //assert(subspace == FALSE);

  for (i = 0; i < numItems; i++)
    {
      if (NULL != root[i].child)
	if (FALSE == subspace)
	  {
	    if (attrClusters.Member(level-1, xact.elements(level-1), root[i].item))
	      {
		root[i].count++;
		root[i].child->IncrementClusterCounts(xact, attrClusters, level+1, 
						      subspace);
	      }
	  }
	else
	  { 
	    assert(1 == level);
	    for (j = 0; j < xact.size()-level+1; j++)
	      {
		//xact.suffix(j);
		if (attrClusters.Member(level-1+j, xact.elements(level-1+j), root[i].item))
		  {
		    root[i].count++;
		    if ((xact.size() > level+j) && (root[i].child != NULL))
		      root[i].child->IncrementClusterCounts(xact, attrClusters, level+1+j,
							    subspace);
		  }
		//xact.suffix(-j);
	      }
	  }
    }
}

void
HashNode::IncrementClusterCounts(Set& xact, ClusterColln& attrClusters, int level,
				 bool subspace)
{
  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->IncrementClusterCounts(xact, attrClusters, level, subspace);
    }
}


void 
PrefixNodeArray::IncrementClusterCounts(Set& xact, ClusterColln& attrClusters, 
					int level, bool subspace)
{
  int i, j;
  
  //assert(subspace == FALSE);
  for (i = 0; i < numNodes; i++)
    {
      if (FALSE == subspace)
	{
	  if (attrClusters.Member(level-1, xact.elements(level-1), prefixNodes[i].item))
	    {
	      prefixNodes[i].count++;
	      if (prefixNodes[i].child != NULL)
		prefixNodes[i].child->IncrementClusterCounts(xact, attrClusters, 
							     level+1, subspace);
	      //else if (prefixNodes[i].status == maximal)
	      //{
	      //  prefixNodes[i].count++;
	      //}
	    }
	}
      else
	{
	  for (j = 0; j < xact.size()-level+1; j++)
	    {
	      //xact.suffix(j);
	      if (attrClusters.Member(level-1+j, xact.elements(level-1+j), prefixNodes[i].item))
		{
		  prefixNodes[i].count++;
		  if ((xact.size() > level+j) && (prefixNodes[i].child != NULL))
		    prefixNodes[i].child->IncrementClusterCounts(xact, attrClusters, 
								 level+j+1, subspace);
		  //else if (prefixNodes[i].status == maximal)
		  //{
		  //  prefixNodes[i].count++;
		  //}
		}
	      //xact.suffix(-j);
	    }
	}
    }
}
#endif
