#include <iostream.h>
#include "PrefixTree.h"

//JOHBEGIN
static int overallMaxItemCount = 0;
static int overallLargeItemCount = 0;
static int numMaxItemsets = 0;
//JOHEND


/***************** Member functions of HashNode *****************/

// I am thinking that maybe the default H here may not be the same as the
// the default H in a hash tree interior node.  If so, we may need to use 
// something other than DEF_HASH_H here. 
HashNode::HashNode(int inputH=DEF_HASH_H, int inputA=DEF_HASH_A, int inputB=DEF_HASH_B)
{
	a = inputA;
	b = inputB;
	h = inputH;

	children = new (PrefixNodeArray*)[h];
	assert(children != NULL);
	for (int i = 0; i < h; i++)
	  children[i] = 0;
	dead = 0;
}

HashNode::~HashNode()
{
  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  {
	    delete children[i];
	    children[i] = NULL;
	  }
    }
  delete[] children;
  children = NULL;
  if (NULL != dead) delete[] dead;
}

#ifndef DIC_ALG
#ifdef SAMPLING
int HashNode::join(int depth, PrefixTree *whereAmI, const Set &pre, JoinAction action = candidateOnly)
#else
int HashNode::join(int depth, PrefixTree *whereAmI, const Set &pre)
#endif SAMPLING
{
  int numNew=0;
  int i;

#ifdef SAMPLING
  // If action is candidateOnly join only the currently processed level.
  if (action == candidateOnly)
#endif

  if (depth < whereAmI->returnPass())
  {
    for (i=0; i<h; i++)
      if (children[i])
#ifdef SAMPLING
 	numNew += children[i]->join(depth, whereAmI, pre, action);
#else
 	numNew += children[i]->join(depth, whereAmI, pre);
#endif
  }
  else
    {
      for (i=0; i<h; i++)
	if (children[i])
	  {
	    int temp = children[i]->numNodes;
	    for (int j=0; j < temp; j++)
	      // For action of candidateOnly, the check below implies that node j
	      // is not in the negative border.
	      if (children[i]->prefixNodes[j].count >= whereAmI->returnMinSupport())
#ifdef SAMPLING
		numNew += whereAmI->join(this, &pre, children[i], j, true, action);
#else
	        numNew += whereAmI->join(this, &pre, children[i], j, true);
#endif
	  }
    }

#ifdef SAMPLING
  // If action is nonSmallNegBorder, then we should join all levels.
  else {
    for (i=0; i<h; i++)
      if (children[i]) {
	// First join lower levels.
 	numNew += children[i]->join(depth, whereAmI, pre, action);

	// Now join the current level.
	int temp = children[i]->numNodes;
	for (int j=0; j< temp; j++)
	  // Instead of checking for minSupport, we check that node j is not small.
	  if (children[i]->prefixNodes[j].status != small) {
	    // Let the join go through all of the nodes (not just the ones whose
	    // item's values are larger).  This is reflected in passing false to
	    // monotonicJoin.  Then we can mark prefix node j as a candidate right
	    // away.  This marking will eliminate unnecessary attempts to join
	    // together two itemsets whose join was already done before.
	    numNew += whereAmI->join(this, &pre, children[i], j, false, action);
	    // children[i]->prefixNodes[j].status = large;
	    children[i]->prefixNodes[j].status = candidate;
	    assert(0); // VENKY
	  }
      }
  }
#endif

  return numNew;
}
#endif

// Sets counts of all itemsets in the tree to 0.
void HashNode::clearCounts() {
  for (int i = 0; i < h; i++)

    if (children[i])

      children[i]->clearCounts();
}


#ifdef SAMPLING
// Returns true iff there exists a large itemset marked negBorder.
bool HashNode::containsLargeNegBorder(int support) {
  for (int i = 0; i < h; i++)

    if (children[i])

      if (children[i]->containsLargeNegBorder(support))
	return true;

  return false;
}
#endif

#ifdef DIC_ALG
int HashNode::M_transaction_process(PrefixTree *whereAmI, const Set &pre, int lastXaction)
{
  int i;
  int numNewCandidates=0;
  int startXaction = (lastXaction + 1) % NUM_TRANSACTIONS;

    for (i=0; i<h; i++)
      if (children[i])
      {
	int temp = children[i]->numNodes;
	for (int j=0; j< temp; j++)
	{
	  if (children[i]->prefixNodes[j].dashed) 
	  {
	    if (children[i]->prefixNodes[j].startTransaction == 
			(lastXaction + 1) % NUM_TRANSACTIONS)
	    {
	      children[i]->prefixNodes[j].dashed = false;
	      whereAmI->decNumDash(1); // one less dashed node in the tree
	      if (children[i]->prefixNodes[j].count < whereAmI->returnMinSupport())
		{
		  delete children[i]->prefixNodes[j].child;
		  children[i]->prefixNodes[j].child = NULL;
		}
	      else
		if (children[i]->prefixNodes[j].circle)
		{
		  children[i]->prefixNodes[j].circle = false;
	      	  numNewCandidates += whereAmI->
			generateCandidates(this, &pre, children[i], j, startXaction);
		}
	    }
	    else
	    {
	      if (children[i]->prefixNodes[j].count >= whereAmI->returnMinSupport() 
		  && children[i]->prefixNodes[j].circle)
	      {
		children[i]->prefixNodes[j].circle = false;
	      	numNewCandidates += whereAmI->
		   generateCandidates(this, &pre, children[i], j, startXaction);
	      }
      	    }
	  }
	  else // !children[i]->prefixNodes[j].dashed
	  {
	    if ((!children[i]->prefixNodes[j].circle) &&
		children[i]->prefixNodes[j].child)
	    {
 	      Set *newPre = pre.addElement(children[i]->prefixNodes[j].item);
 	      numNewCandidates += children[i]->prefixNodes[j].child->
			M_transaction_process(whereAmI, *newPre, lastXaction);
	      delete newPre;
	    }
	  }
	}
      }

  return numNewCandidates;
}
#endif

void HashNode::output(const Set *set, ostream& os)
{
  for (int i = 0; i<h; i++)
    {
      if (children[i])
	{
	  children[i]->output(set, os);
	}
    }
}

/***************** Member functions of PrefixNodeArray ****************/

PrefixNodeArray::PrefixNodeArray(int max = DEF_SETS_PER_LEAF)
{
	maxSize = max;
	numNodes = 0;
	prefixNodes = new PrefixNode[maxSize];
	assert(prefixNodes != NULL);
}

PrefixNodeArray::~PrefixNodeArray()
{
  /***
	for (int i=0; i<numNodes; i++)
	  if (prefixNodes[i].child)
	    {
	      delete prefixNodes[i].child;
	      prefixNodes[i].child = NULL;
	    }
  ***/
	delete[] prefixNodes;
	prefixNodes = NULL;
}

#ifndef DIC_ALG 
#ifdef SAMPLING
int PrefixNodeArray::join(int depth, PrefixTree *whereAmI, const Set &pre, JoinAction action = candidateOnly)
#else
int PrefixNodeArray::join(int depth, PrefixTree *whereAmI, const Set &pre)
#endif SAMPLING
{
  int numNew=0;
  for (int i=0; i<numNodes; i++)
    if (prefixNodes[i].child)
    {
      Set *newPre = pre.addElement(prefixNodes[i].item);
#ifdef SAMPLING
      //if (action == candidateOnly && (prefixNodes[i].status == large ||
      //			      prefixNodes[i].status == maximal ||
      //			      prefixNodes[i].status == candidate) ||
      if (action == candidateOnly && prefixNodes[i].status == candidate || 
        action == nonSmallNegBorder && prefixNodes[i].status != small)
	numNew += prefixNodes[i].child->join(depth+1, whereAmI, *newPre, action);
#else
      numNew += prefixNodes[i].child->join(depth+1, whereAmI, *newPre);
#endif
      delete newPre;
    }
  return numNew;
}
#endif

// Sets counts of all itemsets in the tree to 0.
void PrefixNodeArray::clearCounts() {
  for (int i = 0; i < numNodes; i++) {

    prefixNodes[i].count = 0;

    if (prefixNodes[i].child)
      prefixNodes[i].child->clearCounts();
  }
}


#ifdef SAMPLING
// Returns true iff there exists a large itemset marked negBorder.
bool PrefixNodeArray::containsLargeNegBorder(int support) {
  for (int i = 0; i < numNodes; i++) {

    if (prefixNodes[i].status == negBorder && prefixNodes[i].count >= support)
      return true;

    if (prefixNodes[i].child)
      if (prefixNodes[i].child->containsLargeNegBorder(support))
	return true;
  }

  return false;
}
#endif


void PrefixNodeArray::output(const Set *set, ostream& os)
{
  for (int i=0; i<numNodes; i++)
  {
    // VENKY
    if ((negBorder != prefixNodes[i].status) && (small != prefixNodes[i].status))
      {
   	int temp = set->size();

	if (prefixNodes[i].count > 1)
	  {
	    for (int j=0; j<temp; j++)
	      os << set->elements(j) << ",";
	    os << prefixNodes[i].item << "\t\t\t\t\t" << prefixNodes[i].count;
	    
#ifdef DIC_ALG
	    if (prefixNodes[i].dashed)
	      os << "\tdashed";
	    else
	      os << "\tsolid";
	    
	    if (prefixNodes[i].circle)
	      os << "\tcircle";
	    else
	      os << "\tsquare";
	    
#elif defined(SAMPLING) && !defined(CATCLUS)

	    if (PRESENT_MAXIT == prefixNodes[i].status)
	      os << "\tPRESENT_MAXIT";
	    if (PRESENT_LARGEIT == prefixNodes[i].status)
	      os << "\tPRESENT_LARGEIT";
	    if (maximal == prefixNodes[i].status)
	      os << "\tmaximal";
	    if (candidate == prefixNodes[i].status)
	      os << "\tlarge";
	    if (prefixNodes[i].status == negBorder)
	      os << "\tnegBorder";
	    else if (prefixNodes[i].status == small)
	      os << "\tsmall";
#endif
	    
	    os << "\n";
	    //JOHBEGIN
	    overallLargeItemCount+= prefixNodes[i].count;
	    //JOHEND
	  }
	if (prefixNodes[i].child)
	  {
	    if (maximal == prefixNodes[i].status)
	      {
		overallMaxItemCount+= prefixNodes[i].count;
		numMaxItemsets++; 
	      }
	    Set *newSet;
	    newSet = set->addElement(prefixNodes[i].item);
	    prefixNodes[i].child->output(newSet, os);
	    delete newSet;
	  }
      }
  }
}

/********************************************************************/

PrefixTree::PrefixTree(int numOfItems, int *count1, int *count2, 
		       double relSup, int numTrans, int H=DEF_HASH_H, int A=DEF_HASH_A, 
		       int B=DEF_HASH_B, bool lazy, char *fname)
{
	h = H;
	a = A;
	b = B;
	_LAZY = lazy;
	if (fname != NULL)
	  {
	    delItsFileName = new char[MAX_NAME_LENGTH];
	    strcpy(delItsFileName, fname);
	  }
	else
	  delItsFileName = NULL;

	//	minSupport = support;
	relSupport = relSup;
	numXacts = numTrans;
	numItems = numOfItems;
	_numLargeIts = 0;
	_numMaximalIts = 0;
	_numNegBorderIts = 0;

	minSupport = (int)(relSup * numTrans);

#ifdef DIC_ALG
	numDash = 0;
#endif

	root = new PrefixNode[numItems];
	for (int i=0; i< numItems; i++)
	  {
	    root[i].item = i;
	    root[i].count = count1[i];
#ifdef DIC_ALG
		root[i].dashed = false;
		if (root[i].count >= minSupport)
		  root[i].circle = false;
		else
		  root[i].circle = true;
		// startTransaction doesn't matter since not dashed
#endif
		if (count1[i] < minSupport)
		  {
#ifdef SAMPLING // #ifdef DYNAMIC
		    root[i].status = negBorder;
#endif
		    root[i].child = 0;
		    continue;  // no large 2-itemset can hang off here
		  }
#ifdef SAMPLING // #ifdef DYNAMIC
		//else root[i].status = large;
		else root[i].status = candidate;
#endif

		// The following segment of code insert all *large* 
		// 2-itemsets with the first item i
		// itemsets are inserted only if they are large
		// It is possible that 0 itemset is inserted, in which
		// case we set root[i].child to NULL

		root[i].child = new HashNode(h, a, b);
		int numLargeTwoItemsetInserted = 0;
		//int numNegBorderTwoItemsetsInserted = 0;
		
		Item a[1];
		a[0]=-1;
		Set set(0);
		set.newArr(1,a);
		for (int j=i+1; j<numItems; j++)
		{
		  //cout << " Inserting a 2-itemset " << count2[i * numItems + j] << endl;
		  if (count2[i*(numItems)+j] >= returnMinSupport())
		  {
	  	    numLargeTwoItemsetInserted++;
		    _numLargeIts++;
      		    a[0]=j;

#ifndef DIC_ALG // automatically inserted as a lit for SAMPLING
		    if (!root[i].child->insert(set, count2[i*(numItems)+j]))
#else
		      // pretend that we are at a stop and all the 2-itemsets
		      // here have just become large, which means they are
		      // dashed circle. But don't worry, they should all be
		      // changed to solid square by the next M_transaction_pr.
		      // set the startXaction to FIRST_TRANSACTION is important!
		      // Otherwise it won't work
		      numDash++;
		    if (!root[i].child->insert(set, count2[i*(numItems)+j], FIRST_TRANSACTION, true, true))
#endif
		      {
			cout << "something wrong at HashNode::insert\n";
			exit (1);
		      }
		    count2[i*numItems+j] = NOT_COUNTED;
		  }
#ifdef SAMPLING  // or #ifdef DYNAMIC
		  /*** // VENKY: maintaining these in an array.
		  else if (count1[j] >= returnMinSupport()) 
		    {
		      numNegBorderTwoItemsetsInserted++;
		      _numNegBorderIts++;
		      _numMaximalIts += 2; // has two children
		      
		      //Item j1[0] = {j};
		      a[0] = j; // set.newArr(1,a); // equivalent stmts.
		      //Set set(1, j1);
		      
		      if (!root[i].child->insert(set, count2[i*(numItems)+j], negBorder)) {
			cout << "something wrong at HashNode::insert\n";
			exit (1);
		      }
		    }
		  ***/
#endif SAMPLING		      
		}
		set.Nullify();
		//delete [] a;
		if (numLargeTwoItemsetInserted == 0)
		  {
		    delete root[i].child;
		    root[i].child = 0;
		  }
	}
	
#ifndef DIC_ALG
	pass = 2;
#endif
}

PrefixTree::~PrefixTree()
{
  /***
  for (int i=0; i<numItems; i++)
    if (NULL != root[i].child)
    {
      delete root[i].child;
      root[i].child = NULL;
      }
  ***/
  delete[] root;
}

#ifndef DIC_ALG
#ifdef SAMPLING
int PrefixTree::join(HashNode *parent, const Set *pre, 
		     PrefixNodeArray *siblings, int offset, bool monotonicJoin, JoinAction action = candidateOnly)
#else
int PrefixTree::join(HashNode *parent, const Set *pre, 
		     PrefixNodeArray *siblings, int offset, bool monotonicJoin)
#endif SAMPLING
{
  PrefixNode toBeJoined;
  toBeJoined = siblings->prefixNodes[offset];
  Item item = toBeJoined.item;

#ifdef SAMPLING
  if (action == candidateOnly)
    // Otherwise count >= minSupport is not necessary.
#endif
  assert (toBeJoined.count >= minSupport);

  //cout << "PT:join addElement " << endl;

  Set *newSet = pre->addElement(item);
  int newCandidateGenerated = 0;

  int temp = parent->h;
  int i, j;
  for (i=0; i<temp; i++)
    {
      //cout << " PT:join :: i" << i << endl;
      if (PrefixNodeArray *cur = parent->children[i])
	{
	  int temp2 = cur->numNodes;
	  for (j = 0; j<temp2; j++)
	    {
	      //cout << " PT:join :: j" << j << endl;
#ifdef SAMPLING
	      if (action == nonSmallNegBorder) 
		{
		  if (!(cur->prefixNodes[j].status != small &&
			(toBeJoined.status == negBorder || cur->prefixNodes[j].status == negBorder) &&
			(!monotonicJoin && cur->prefixNodes[j].item != item ||
			 monotonicJoin && cur->prefixNodes[j].item > item)))
		    continue;
		} 
	      else  // action == candidateOnly
	  // For sampling this is equivalent to not joining a negBorder set.
		if (!(cur->prefixNodes[j].count >= minSupport && 
		      (!monotonicJoin && cur->prefixNodes[j].item != item ||
		       monotonicJoin && cur->prefixNodes[j].item > item)))
		  continue;
#else
	      if (cur->prefixNodes[j].count >= minSupport && 
		  (((!monotonicJoin)&&(cur->prefixNodes[j].item != item)) ||
		   (monotonicJoin && (cur->prefixNodes[j].item > item))))
#endif	  
		{
		  
		  //cout << " PT: join : before operator= and .insert " << endl;

		  Set candidateSet;
		  candidateSet = *newSet;
		  candidateSet.insert(cur->prefixNodes[j].item);
	  
		  // the following checks if every subset derived by yanking out
		  // one element from candidateSet is large

		  bool large = true;
		  int temp3 = candidateSet.size() - 2;

#ifdef SAMPLING
		  if (action == candidateOnly)
		    // Otherwise, don't do pruneBySubset (coded below).
#endif
		    // -2 because yanking out any of the last two items will always
		    // result in a large itemset
		    
		    // Note that the above test relies on the for loop being
		    // just one compound statement.
		    for (int k = 0; k<temp3; k++)
		      {
			Item tmp;
			candidateSet.yankOut(tmp, k);
			if (count(candidateSet) < minSupport)
			  large = false;
			candidateSet.putBack(tmp);
		      }

		  //delete &candidateSet; //VENKY

		  if (large)
		    {
		      newCandidateGenerated++;
		      // insert the new candidateSet into the proper place
		      HashNode *newHashNode;
		      if (item < cur->prefixNodes[j].item)
			{
			  newHashNode = siblings->prefixNodes[offset].child;
			  if (!newHashNode)
			    newHashNode = siblings->prefixNodes[offset].child
			      = new HashNode;
			  
			  Set *lastItemSet; // a one item set containing the last item
			  // of the new candidateSet set 
			  //VENKY: changed to remove the call to new;
			  Item oneItemArray[1];
			  oneItemArray[0] = cur->prefixNodes[j].item;
			  
			  //cout << "HashNode::join (1) : before creating a new Set" << endl;
			  
			  lastItemSet = new Set(1, oneItemArray);
#ifdef SAMPLING
			  if (action == candidateOnly)
			    newHashNode->insert(*lastItemSet, 0, candidate);
			  else {
			    assert(!newHashNode->member(*lastItemSet, 0));
			    // Mark the itemset as a newly generated one.
			    newHashNode->insert(*lastItemSet, 0, negBorder);
			  }
#else
			  newHashNode->insert(*lastItemSet, 0);
#endif
			  //cout << "HashNode::join (1) : before deleting the new Set" << endl;
			  
			  delete lastItemSet;
			}
		      else if (item > cur->prefixNodes[j].item)
			{
			  newHashNode = cur->prefixNodes[j].child;
			  if (!newHashNode)
			    newHashNode = cur->prefixNodes[j].child
			      = new HashNode;
			  
			  Set *lastItemSet; // a one item set containing the last item
			  // of the new candidateSet set
			  Item oneItemArray[1];
			  oneItemArray[0] = siblings->prefixNodes[offset].item;
			  
			  //cout << "HashNode::join (2) : before creating a new Set" << endl;
			  
			  lastItemSet = new Set(1, oneItemArray);
#ifdef SAMPLING
			  if (action == candidateOnly)
			    newHashNode->insert(*lastItemSet, 0, candidate);
			  else 
			    {
			      assert(!newHashNode->member(*lastItemSet, 0));
			      // Mark the itemset as a newly generated one.
			      newHashNode->insert(*lastItemSet, 0, negBorder);
			    }
#else
			  newHashNode->insert(*lastItemSet, 0);
#endif
			  //cout << "HashNode::join (2) : before deleting the new Set" << endl;
			  delete lastItemSet;
			}
		    }
		}
	    }
	}
    }
  
  delete newSet;

  return newCandidateGenerated;
}
#endif

#ifdef DIC_ALG
int PrefixTree::generateCandidates(HashNode *parent, const Set *pre, 
				 PrefixNodeArray *siblings,
		         	 int offset, int startTransaction)
{
  assert(pre->size()>=1); // one and two itemsets are counted using arrays!

  Item item = siblings->prefixNodes[offset].item;
  Set *seedSet = pre->addElement(item);
  int newCandidateGenerated = 0;

  for (int i=0; i<numItems; i++)
  {
    if (root[i].count >= minSupport && (!seedSet->member(root[i].item)))
    {
      // puting root[i].item into the set considered

      // local is true if we can insert the new candidate locally, that is,
      // in the subtree rooted at parent. Otherwise, we need to insert the
      // new candidate from the roots
      bool local = (root[i].item > seedSet->elements(seedSet->size()-2));

      Set newCandidate;
      newCandidate = *seedSet;
      newCandidate.insert(root[i].item);

      // the following checks if every subset derived by yanking out
      // one element from candidate is large, i.e., prune by subset

      bool large = true;
      int temp = newCandidate.size();
      for (int k = 0; k<temp; k++)
      {
        Item tmp;
        bool yanked = newCandidate.yankOut(tmp, k);
        assert(yanked);
        if (!squareCount(newCandidate, minSupport))
	{
          large = false;
	  newCandidate.putBack(tmp);
	  break;
	}
        newCandidate.putBack(tmp);
      }

      if (!large)
      {
	continue;
      }
      // end if prune by subset

      newCandidateGenerated++;

      if (local)
      {
	newCandidate.suffix(newCandidate.size() - 2);
	parent->insert(newCandidate, 0, startTransaction, true, true);
	newCandidate.suffix(2 - newCandidate.size());
      }
      else
	insert(newCandidate, startTransaction);

    }
  }
  delete seedSet;

  numDash += newCandidateGenerated;

  return newCandidateGenerated;
}
#endif

#ifndef DIC_ALG
#ifdef SAMPLING
PrefixTree* PrefixTree::join(int &numCandidates, JoinAction action = candidateOnly)
#else
PrefixTree* PrefixTree::join(int &numCandidates)
#endif SAMPLING
{
  assert(pass>=2);

  numCandidates = 0;
  Set set(0);

  for (int i=0; i<numItems; i++)
  {
#ifdef SAMPLING
    if (root[i].status == small)
      continue;

    // Join even singleton itemsets if action is nonSmallNegBorder.
    // One of the joining itemsets must be in the negative border.
    if (action == nonSmallNegBorder) {

      for (int j = i + 1; j < numItems; j++) {
	if (root[j].status == small)
	  continue;

	// LA: Revisit: make sure of this code's correctness.
	// At least one of the itemsets is in the negative border.  If this
	// is not the first iteration of the S := S U NB(S), then that set was
	// generated in the last iteration.  The resulting join should be new.
	if (root[i].status == negBorder || root[j].status == negBorder) {
	  Item aj[] = {j};
	  Set set(1, aj);
	  if (!root[i].child)
	    root[i].child = new HashNode(h, a, b);

	  assert(!root[i].child->member(set, 0));
	  root[i].child->insert(set, negBorder);
	}  
      }
	
      // Now mark the processed singleton {i} as a candidate, so
      // that it is not used in joins of the next iteration, since
      // it would generate the same sets again.
      root[i].status = candidate;
    }
#endif

    // During the apriori phase of sampling, action is candidateOnly, and
    // if root[i].status == negBorder, then root[i].child == NULL.
    // In the later phase, action is nonSmallNegBorder, and at this node
    // root[i].status != small, so we need to join.
    // LA: Revisit: The old test still works for now.  Check later.
    if (root[i].child)
    {
      // construct a Set with only one element: root[i].item
      Item a[1];
      a[0]=root[i].item;
      set.newArr(1,a);

#ifdef SAMPLING
      //if (action == nonSmallNegBorder || root[i].status == candidate) // VENKY
      if (action == nonSmallNegBorder || root[i].status == candidate)
	numCandidates += root[i].child->join(2, this, set, action);
#else
      numCandidates += root[i].child->join(2, this, set);
#endif
    }
  }

  set.Nullify();
  pass++;

  return this;
}
#endif

// Sets counts of all itemsets in the tree to 0.
void PrefixTree::clearCounts() {
  for (int i=0; i<numItems; i++)
  {
    root[i].count = 0;

    if (root[i].child)
      root[i].child->clearCounts();
  }
}

#ifdef SAMPLING
// Returns true iff there exists a large itemset marked negBorder.
bool PrefixTree::containsLargeNegBorder() {
  for (int i=0; i<numItems; i++)
  {
    if (root[i].status == negBorder && root[i].count >= minSupport)
      return true;

    if (root[i].child)
      if (root[i].child->containsLargeNegBorder(minSupport))
	return true;
  }

  return false;
}
#endif


#ifdef DIC_ALG
int PrefixTree::M_transaction_process(int lastTransaction)
{
  int numNewCandidates=0;

  for (int i=0; i<numItems; i++)
  {
    if (root[i].child)
    {
      // construct a Set with only one element: root[i].item
      Item a[1];
      a[0]=root[i].item;
      Set set(1, a);

      numNewCandidates += root[i].child->
			M_transaction_process(this, set, lastTransaction);
    }
  }

  return numNewCandidates;
}
#endif


ostream& operator<<(ostream& os, const PrefixTree& pt)
{
  os << "\nItemSet\t\t\t\t\tCount\n***********************************\n\n";
  os << pt.minSupport << " : " << pt.numXacts << endl;
  for (int i=0; i<pt.numItems; i++)
  {
    if (pt.root[i].count >= pt.minSupport)
      {
      os << pt.root[i].item << "\t\t\t\t\t" << pt.root[i].count;
#ifdef DIC_ALG
	  if (pt.root[i].dashed)
	    os << "\tdashed";
	  else 
	    os << "\tsolid";

	  if (pt.root[i].circle)
 	    os << "\tcircle";
	  else
	    os << "\tsquare";
#endif
#ifdef SAMPLING
	if (PRESENT_MAXIT == pt.root[i].status)
	  os << "\tPRESENT_MAXIT";
	if (PRESENT_LARGEIT == pt.root[i].status)
	  os << "\tPRESENT_LARGEIT";
	if (maximal == pt.root[i].status)
	  os << "\tmaximal";
	if (candidate == pt.root[i].status)
	  os << "\tlarge";
	if (pt.root[i].status == negBorder)
	  os << "\tnegBorder";
	else if (pt.root[i].status == small)
	  os << "\tsmall";
#endif
      os << "\n";

      //JOHBEGIN
      if (maximal == pt.root[i].status) {
	overallMaxItemCount+= pt.root[i].count;
      }
      overallLargeItemCount+= pt.root[i].count;
      //JOHEND

    }

    if (pt.root[i].child)
    {
      // construct a Set with only one element: root[i].item
      Item a[1];
      a[0]=pt.root[i].item;
      Set set(1, a);
      pt.root[i].child->output(&set, os);
    }
  }
  
  //JOHBEGIN
  //os << "Total number of maximal large itemsets = " << numMaxItemsets << endl;
  //os << "Total support of maximal large itemsets = " << overallMaxItemCount << endl;
  //os << "Total support of large itemsets = " << overallLargeItemCount << endl;
  //JOHEND


  return os;
  
}




ofstream& operator<<(ofstream& os, const PrefixTree& pt)
{
  os << "\nItemSet\t\t\t\t\tCount\n***********************************\n\n";
  os << pt.minSupport << " : " << pt.numXacts << endl;
  for (int i=0; i<pt.numItems; i++)
  {
    if (pt.root[i].count >= pt.minSupport)
    {
      os << pt.root[i].item << "\t\t\t\t\t" << pt.root[i].count;
#ifdef DIC_ALG
	  if (pt.root[i].dashed)
	    os << "\tdashed";
	  else 
	    os << "\tsolid";

	  if (pt.root[i].circle)
 	    os << "\tcircle";
	  else
	    os << "\tsquare";
#endif
#ifdef SAMPLING
	if (PRESENT_MAXIT == pt.root[i].status)
	  os << "\tPRESENT_MAXIT";
	if (PRESENT_LARGEIT == pt.root[i].status)
	  os << "\tPRESENT_LARGEIT";
	if (maximal == pt.root[i].status)
	  os << "\tmaximal";
	if (candidate == pt.root[i].status)
	  os << "\tlarge";
	if (pt.root[i].status == negBorder)
	  os << "\tnegBorder";
	else if (pt.root[i].status == small)
	  os << "\tsmall";
#endif
      os << "\n";

      //JOHBEGIN

      if (maximal == pt.root[i].status) {
	overallMaxItemCount+= pt.root[i].count;
      }
      overallLargeItemCount+= pt.root[i].count;
      //JOHEND

    }

    if (pt.root[i].child)
    {
      // construct a Set with only one element: root[i].item
      Item a[1];
      a[0]=pt.root[i].item;
      Set set(1, a);
      pt.root[i].child->output(&set, os);
    }
  }
  
  //JOHBEGIN
  //os << "Total number of maximal large itemsets = " << numMaxItemsets << endl;
  //os << "Total support of maximal large itemsets = " << overallMaxItemCount << endl;
  //os << "Total support of large itemsets = " << overallLargeItemCount << endl;
  //JOHEND


  return os;
  
}

