/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "PrefixTree.h"
#include "globals.h"
#include "trans_utils.h"
#include "2DArray.h"
#include "Set.h"

#ifdef CATCLUS
#include "cluster.h"
#endif

PrefixTree::PrefixTree(int nitems, Set **itemsets, int numItemsets, 
		       int H=DEF_HASH_H, int A=DEF_HASH_A, int B=DEF_HASH_B)
{
  h = H;
  a = A;
  b = B;

  numItems = nitems;
  numXacts = 0;
  relSupport = 0.0;
  minSupport = 0;
  _numLargeIts = 0;
  _numMaximalIts = 0;
  _numNegBorderIts = 0;
  _LAZY = FALSE;

  root = new PrefixNode[numItems];

  int i;
  for (i = 0; i < numItems; i++)
    {
      root[i].item = i;
      root[i].count = 0;
      root[i].status = candidate;
      root[i].child = NULL;
    }

  for (i = 0; i < numItemsets; i++)
    insert(*(itemsets[i]), 0, candidate);
#ifndef DIC_ALG
  pass = 2;
#endif

}

#ifndef CATCLUS
void PrefixTree::CountItemsets(char *datafile)
{
  char buffer[TRANS_BLOCK_SIZE];
  Set xactIter(0);
  int tid = 0;
  int numTrans = 0;

  memset((void*)buffer, 0, TRANS_BLOCK_SIZE);
  SetupXactScan(datafile, buffer);
  while (getNextXact(xactIter, tid, buffer) != -1)
    {
      numTrans++;
      incrementCounts(xactIter, 1, allLevels, FALSE, NULL, -1);
    }
#ifdef DEBUG
  cout << "Number of transactions read = " << numTrans << endl;
#endif
  xactIter.Nullify();
}
#endif CATCLUS


void 
PrefixTree::OpenDelItsFile()
{
  assert(delItsFileName != NULL);
  delItsFile.open(delItsFileName);
  assert(delItsFile != NULL);
}

void PrefixTree::WriteOldIts(Set *itemset, int count)
{
  // OPTIMIZE: should avoid too many calls to write. Store all 
  // these in an array and make one call to write!
  //cout << " Writing a deleted itemset: (Currently, nothing done)" << endl;
  if (delItsFileName != NULL)
    {
      delItsFile.open(delItsFileName);
      itemset->WriteToFile(delItsFile, count);
      delItsFile.close();
    }
}

void PrefixTree::CloseDelItsFile()
{
  assert(delItsFile != NULL);
  delItsFile.close();
  //  delItsFile = NULL;
}


int 
PrefixTree::MarkPossMaximalIts(int& totMaxItSup)
{
  int i;
  int numMaxIts = 0;
  PrefixNodeStatus status;
  int minSup = returnMinSupport();

  for (i = 0; i < numItems; i++)
    {
      if (NULL == root[i].child)
	{
	  if (root[i].count >= minSup)
	    {
	      root[i].status = maximal;
	      totMaxItSup += root[i].count;
	      numMaxIts++;
	      continue;
	    }
	}
      else
	{
	  status = root[i].child->MarkPossMaximalIts(numMaxIts, totMaxItSup, minSup);
	  if ((negBorder == status) || (small == status))
	    {
	      root[i].status = maximal;
	      totMaxItSup += root[i].count;
	      numMaxIts++;
	    }
	  else
	    root[i].status = candidate;
	}
    }
  cout << "Total number of Maximal itemsets : " << numMaxIts << endl;
  return numMaxIts;
}


PrefixNodeStatus 
HashNode::MarkPossMaximalIts(int& numMaxIts, int& totMaxItSup, int minSup)
{
  int i;
  PrefixNodeStatus st1, status = negBorder;

  if (NULL == children)
    return status;

  for (i = 0; i < h; i++)
    if (NULL != children[i])
      {
	st1 = children[i]->MarkPossMaximalIts(numMaxIts, totMaxItSup, minSup);
	if ((st1 == maximal) || (st1 == candidate))
	  {
	    status = candidate;
	  }
      }
  return status;
}

PrefixNodeStatus 
PrefixNodeArray::MarkPossMaximalIts(int &numMaxIts, int& totMaxItSup, 
				    int minSup)
{
  int i;
  PrefixNodeStatus st1, status = negBorder;

  for (i = 0; i < numNodes; i++)
    {
      if (prefixNodes[i].count >= minSup)
	{
	  prefixNodes[i].status = candidate;
	  if (prefixNodes[i].child != NULL)
	    {
	      st1 = prefixNodes[i].child->MarkPossMaximalIts(numMaxIts, totMaxItSup, minSup);
	      if (st1 == negBorder)
		{
		  prefixNodes[i].status = maximal; 
		  numMaxIts++;
		  totMaxItSup += prefixNodes[i].count;
		  status = candidate;
		}
	    }
	  else
	    {
	      prefixNodes[i].status = maximal;
	      numMaxIts++;
	      totMaxItSup += prefixNodes[i].count;
	      status = candidate;
	    }
	}
    }
  return status;  
}


int PrefixTree::UnMarkNonMaximalIts(int& totUnMarkedSup, PrefixTree *ptNew)
{
  int numUnMarked = 0;
  int i;
  Item arr[1];
  Set set(0);
  set.newArr(1, arr); // setting up a temp set variable.

  for (i = 0; i < numItems; i++)
    {
      if ((maximal == root[i].status) || (negBorder == root[i].status))
	continue;
      if (root[i].child != NULL)
	{
	  arr[0] = i;
	  totUnMarkedSup += root[i].child->UnMarkNonMaximalIts(numUnMarked, set, 
							       this, ptNew);
	}
    }
  set.Nullify();
  return numUnMarked;
}

int HashNode::UnMarkNonMaximalIts(int& numUnMarked, Set& prefix, PrefixTree *pt,
				  PrefixTree *ptNew)
{
  int i;
  int totUnMarkedSup = 0;

  if (NULL != children)
    for (i = 0; i < h; i++)
      if (NULL != children[i])
	totUnMarkedSup += children[i]->UnMarkNonMaximalIts(numUnMarked, prefix, 
							   pt, ptNew);
  return totUnMarkedSup;
}

int PrefixNodeArray::UnMarkNonMaximalIts(int &numUnMarked, Set& prefix, PrefixTree *pt,
					 PrefixTree *ptNew)
{
  int totUnMarkedSup = 0;
  int i,j;
  int tempSz;
  Item tmp;
  bool yanked;
  Set set(0);

  for (i = 0; i < numNodes; i++)
    {
      Set *newSet = prefix.addElement(prefixNodes[i].item);

      if (maximal == prefixNodes[i].status)
	{
	  tempSz = newSet->size();
	  for (j = 0; j < tempSz; j++)
	    {
	      yanked = newSet->yankOut(tmp,j);
	      assert(yanked);
	      totUnMarkedSup += pt->UnMarkMaxItsContainedIn(numUnMarked, newSet);
	      newSet->putBack(tmp);
	    }
	  if (NULL != ptNew)
	    {
	      ptNew->insertWithCount(*newSet, 1, maximal);
	    }
	}
      else if (prefixNodes[i].child != NULL)
	totUnMarkedSup += prefixNodes[i].child->UnMarkNonMaximalIts(numUnMarked, *newSet, pt, ptNew);

      delete newSet;
    }
  return totUnMarkedSup;
}

int PrefixTree::UnMarkMaxItsContainedIn(int& numUnMarked, Set *maxIt)
{
  int i, indx;
  int totSupContained = 0;
  int sz = maxIt->size();

  for (i = 0; i < sz; i++)
    {
      indx = maxIt->elements(i);
      if (maximal == root[indx].status)
	{
	  root[indx].status = candidate;
	  totSupContained += root[indx].count;
	  numUnMarked++;
	}
      else if ((root[indx].child != NULL) && (maxIt->size() > 1))
	{
	  maxIt->suffix(i+1);
	  totSupContained += root[indx].child->UnMarkMaxItsContainedIn(numUnMarked, maxIt);
	  maxIt->suffix(-i-1);
	}
    }
  return totSupContained;
}

int HashNode::UnMarkMaxItsContainedIn(int& numUnMarked, Set *maxIt)
{
  int i;
  int totSupContained = 0;
  int link;
  if (NULL != children)
    {
      for (i = 0; i < maxIt->size(); i++)
	{
	  link = HASH(a, b, h, maxIt->elements(i));
	  if (children[link])
	    {
	      maxIt->suffix(i);
	      totSupContained += children[link]->UnMarkMaxItsContainedIn(numUnMarked,
									 maxIt);
	      maxIt->suffix(-i);
	    }
	}
    }
  /***
    for (i = 0; i < h; i++)
      if (NULL != children[i])
      totSupContained += children[i]->UnMarkMaxItsContainedIn(numUnMarked, maxIt);  
  ****/

  return totSupContained;
}

int PrefixNodeArray::UnMarkMaxItsContainedIn(int& numUnMarked, Set *maxIt)
{
  int i;
  int totSupContained = 0;

  for (i = 0; i < numNodes; i++)
    {
      if (prefixNodes[i].item == maxIt->elements(0))
	{
	  if (maximal == prefixNodes[i].status)
	    {
	      //assert(maxIt->size() > 1);
	      prefixNodes[i].status = candidate;
	      totSupContained += prefixNodes[i].count;
	      numUnMarked++;
	      continue;
	    }
	  if ((prefixNodes[i].child != NULL) && (maxIt->size() > 1) 
	      && (prefixNodes[i].status != negBorder))
	    {
	      maxIt->suffix(1);
	      totSupContained += prefixNodes[i].child->UnMarkMaxItsContainedIn(numUnMarked, maxIt);
	      maxIt->suffix(-1);
	    }
	  //break;
	}
    }
  return totSupContained;
}

/*****
// should be called before DiffwithPT2
void PrefixTree::AdjustCountForDevn(Set& itemset)
{
  int i;
  int decrCnt = 0;

  for (i = 0; i < numItems; i++)
    {
      if ((root[i].item == candidate) && (!itemset.member(root[i].item)))
	{
	  Set newSet;
	  newSet = itemset;
	  newSet.insert(root[i].item);
	  
	  decrCnt += count(newSet);
	}
    }
  bool flag = DecrCount(itemset, decrCnt);
  assert(TRUE == flag);
}
***/

/**** Did not test these routines.
int PrefixTree::AdjustITCounts()
{
  int i;
  int arr[MAX_IT_LENGTH];
  int numAdjusted=0;

  cout << "************************" << endl;
  cout << " Inside AdjustITCounts " << endl;

  for (i = 0; i < numItems; i++)
    {
      arr[0] = root[i].item;
      switch(root[i].status)
	{
	case candidate:
	  root[i].child->AdjustITCounts(this, arr, 2, numAdjusted);
	  break;
	case TO_BE_ADJUSTED:
	  root[i].status = ADJUSTED_LARGEIT;
	  break;
	case maximal:
	  //root[i].status = ADJUSTED_MAXIT;
	case negBorder:
	  break;
	default:
	  assert(0);
	}
    }
  cout << "************************" << endl;
  cout << " After AdjustITCounts " << endl;
  return numAdjusted;
}


void HashNode::AdjustITCounts(PrefixTree *pt, int *arr, int level, 
			      int& numAdjusted)
{
  int i;
  
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->AdjustITCounts(pt, arr, level, numAdjusted);
    }
}

void PrefixNodeArray::AdjustITCounts(PrefixTree *pt, int *arr, int level,
				     int& numAdjusted)
{
  int i;
  Set set(0);

  for (i = 0; i < numNodes; i++)
    {
      arr[level-1] = prefixNodes[i].item;
      set.newArr(level, arr);

      switch(prefixNodes[i].status)
	{
	case maximal:
	  cout << set << " " << prefixNodes[i].count << endl;
	  pt->DecrCountsForDevn(set, prefixNodes[i].count);
	  prefixNodes[i].status = ADJUSTED_MAXIT;
	  numAdjusted++;
	  break;
	case TO_BE_ADJUSTED:
	  cout << set << " " << prefixNodes[i].count << endl;
	  pt->DecrCountsForDevn(set, prefixNodes[i].count);
	  numAdjusted++;
	  prefixNodes[i].status = ADJUSTED_LARGEIT;
	  break;
	case candidate:
	case UNADJUSTABLE_AS_YET:
	  if (prefixNodes[i].child != NULL)
	    prefixNodes[i].child->AdjustITCounts(pt, arr, level+1, numAdjusted);
	  break;
	default:
	  assert(0);
	}
    }
  set.Nullify();
}
*******/

/***
void PrefixTree::GetNonPresentMaxIts(Set **itemsets, int& numIts)
{
  int i;
  int arr[MAX_IT_LENGTH];
  numIts = 0;

  for (i = 0; i < numItems; i++)
    {
      arr[0] = root[i].item;
      if (maximal == root[i].status)
	itemsets[numIts++] = new Set(1, arr);
      else if (candidate == root[i].status)
	if (NULL != root[i].child)
	  root[i].child->GetNonPresentMaxIts(itemsets, numIts, 2, arr);
    }
  assert(numIts < MAX_SETCNT);
}

void HashNode::GetNonPresentMaxIts(Set **itemsets, int& numIts, int level, int *arr)
{

  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->GetNonPresentMaxIts(itemsets, numIts, level, arr);
    }
  assert(numIts < MAX_SETCNT);
}

void PrefixNodeArray::GetNonPresentMaxIts(Set **itemsets, int& numIts, int level, 
					  int *arr)
{
  int i;
  
  for (i = 0; i < numNodes; i++)
    {
      arr[level-1] = prefixNodes[i].item;
      if (maximal == prefixNodes[i].status)
	itemsets[numIts++] = new Set(level, arr);
      else if (candidate == prefixNodes[i].status)
	if (NULL != prefixNodes[i].child)
	  prefixNodes[i].child->GetNonPresentMaxIts(itemsets, numIts, level+1, arr);
    }
  assert(numIts < MAX_SETCNT);
}
****/

void PrefixTree::ChangeMarking(PrefixNodeStatus from, PrefixNodeStatus to)
{
  int i;
  for (i = 0; i < numItems; i++)
    if (from == root[i].status)
      root[i].status = to;

  if (root[i].child != NULL)
    root[i].child->ChangeMarking(from,to);
}

void HashNode::ChangeMarking(PrefixNodeStatus from, PrefixNodeStatus to)
{
  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->ChangeMarking(from,to);
    }
}

void PrefixNodeArray::ChangeMarking(PrefixNodeStatus from, PrefixNodeStatus to)
{
  int i;
  for (i = 0; i < numNodes; i++)
    if (from == prefixNodes[i].status)
      prefixNodes[i].status = to;

  if (NULL != prefixNodes[i].child)
    prefixNodes[i].child->ChangeMarking(from, to);
}

void PrefixTree::UnPRESENTIts()
{
  int i;
  
  for (i = 0; i < numItems; i++)
    {
      switch(root[i].status)
	{
	case PRESENT_MAXIT:
	  root[i].status = maximal;
	  break;
	case PRESENT_LARGEIT:
	  root[i].status = candidate;
	  break;
	case ADJUSTED_LARGEIT:
	  root[i].status = candidate;
	  break;
	case ADJUSTED_MAXIT:
	  root[i].status = maximal;
	  break;
	case UNADJUSTABLE_AS_YET:
	case maximal:
	case small:
	case negBorder:
	case candidate:
	  break;
	case TO_BE_ADJUSTED:
	  assert(0);
	default:
	  assert(0);
	}

	if (NULL != root[i].child)
	  root[i].child->UnPRESENTIts();
    }
}

void HashNode::UnPRESENTIts()
{

  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->UnPRESENTIts();
    }
}

void PrefixNodeArray::UnPRESENTIts()
{
  int i;
  
  for (i = 0; i < numNodes; i++)
    {
      switch(prefixNodes[i].status)
	{
	case PRESENT_MAXIT:
	  prefixNodes[i].status = maximal;
	  break;
	case PRESENT_LARGEIT:
	  prefixNodes[i].status = candidate;
	  break;
	case ADJUSTED_LARGEIT:
	  prefixNodes[i].status = candidate;
	  break;
	case ADJUSTED_MAXIT:
	  prefixNodes[i].status = maximal;
	  break;

	case UNADJUSTABLE_AS_YET:
	case maximal:
	case small:
	case negBorder:
	case candidate:
	  break;
	case TO_BE_ADJUSTED:
	  assert(0);
	default:
	  assert(0);
	}

      if (candidate == prefixNodes[i].status)
	if (NULL != prefixNodes[i].child)
	  prefixNodes[i].child->UnPRESENTIts();
    }
}


#ifdef INSTRUMENT
// all instrumentation code for experiments here.

// assume that the negative border is marked before calling
// this function;
int PrefixTree::GetRandNegBItemsets(Set **itemsetArr, int numItemsets)
{
  srandom(101);
  int randNum = 0;
  int numFilled = 0;
  int arr[MAX_IT_LENGTH];
  bool flag;
  int level = 1;

  flag = FALSE;
  srandom(1001);

  //for (k = 0; k < numItemsets; k++)
    {
      flag = FALSE;
      //randNum = random() % NumItems();
      //cout << randNum << endl;
      //  while (FALSE == flag)
      for (randNum = 0; randNum < numItems; randNum++)
	{
	  if (numFilled == numItemsets)
	    break;
	  arr[level-1] = root[randNum].item;
	  if (negBorder == root[randNum].status)
	    {
	      continue;
	      //itemsetArr[numFilled++] = new Set(1, arr);
	      //flag = TRUE;
	    }
	  else // go down the tree; 
	    {
	      if (NULL != root[randNum].child)
		flag = root[randNum].child->GetRandNegBItemset(itemsetArr, numFilled,
							    numItemsets, level+1, arr);
	    }
	}
    }
  return numFilled;
}


bool HashNode::GetRandNegBItemset(Set **itemsetArr, int& numFilled, 
				  int numItemsets, int level, int *arr)
{
  int randNum;
  srandom(10101);
  bool flag = FALSE;

  if (NULL == children)
    return FALSE;

  randNum = random() % h;
  for (randNum = 0; randNum < h; randNum++)
    {
      if (NULL != children[randNum])
	flag = children[randNum]->GetRandNegBItemset(itemsetArr, numFilled, 
						   numItemsets, level, arr);
      if (numFilled == numItemsets)
	return flag;
    }
  return flag;
}

bool PrefixNodeArray::GetRandNegBItemset(Set **itemsetArr, int& numFilled, 
					 int numItemsets, int level, int *arr)
{
  int randNum;
  bool flag = FALSE;

  if (0 == numNodes)
    {
      assert(0);
      return FALSE;
    }

  //srandom(30301);
  //while (FALSE == flag)
  // {
  //randNum = random() % numNodes;
  randNum = 0;
  for (randNum = 0; randNum < numNodes; randNum++)
    {
      if (numItemsets == numFilled)
	return FALSE;
      if (negBorder == prefixNodes[randNum].status)
	{
	  arr[level-1] = prefixNodes[randNum].item;
	  itemsetArr[numFilled++] = new Set(level, arr);
	  if (numFilled == numItemsets)
	    return TRUE;
	  flag = TRUE;
	}
      else
	if (NULL != prefixNodes[randNum].child)
	  {
	    arr[level-1] = prefixNodes[randNum].item;
	    flag = prefixNodes[randNum].child->GetRandNegBItemset(itemsetArr, numFilled, 
								  numItemsets, level+1, arr);
	  }
    }
  return flag;
}
#endif






