/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include "PrefixTree.h"

#include "globals.h"
#include "trans_utils.h"
#include "2DArray.h"
#include "Set.h"


// assumes that the maximality of an itemset is marked correctly;
// scans all maximal itemsets in *this, and for each one checks if it
// is also present in pt2. If so, find the deviation due to that;
// else, append it to itemsets array, and increment numIts;
void PrefixTree::DiffwithPT2(PrefixTree *pt2, Set **itemsets, int& numIts, double& diff)
{
  int i;
  Item arr[MAX_IT_LENGTH];
  
  Set set(0);
  //double diff = 0.0;
  numIts = 0; // initialize;

  int ntrans1 = getNumXacts();
  int ntrans2 = pt2->getNumXacts();
  int count2;
  double val1, val2;


  for (i = 0; i < numItems; i++)
    {
      arr[0] = root[i].item;
      if ((maximal == root[i].status) || (candidate == root[i].status))
	{
	  set.newArr(1, arr);
	  count2 = pt2->countForDevn(set);
	  if (count2 >= 0)
	    {
	      val2 = ((double)count2)/(double)ntrans2;
	      val1 = ((double)root[i].count)/(double)ntrans1;
	      AggregateFn(DeviationFn(val1, val2), diff);
	    }
	  else if (SUPERSET == count2)
	    {
	      // 1-item counts are always maintained in PT;
	      assert(0);
	      itemsets[numIts++] = new Set(1, arr);
	    }
	  // else SUBSET ; do nothing
	}
      
      if ((candidate == root[i].status) || (PRESENT_LARGEIT == root[i].status))
	{
	  // go down to its children;
	  if (NULL != root[i].child)
	    root[i].child->DiffwithPT2(pt2, itemsets, numIts, 2, arr, ntrans1, diff);
	}
    }
  set.Nullify();

  //return diff;
}

void HashNode::DiffwithPT2(PrefixTree *pt2, Set **itemsets, int& numIts, 
			   int level, int *arr, int ntrans1, double& diff)
{
  int i;
  if (NULL != children)
    {
      for (i = 0; i < h; i++)
	if (NULL != children[i])
	  children[i]->DiffwithPT2(pt2, itemsets, numIts, level, arr, ntrans1, diff);
    }
  //return diff;
}

void PrefixNodeArray::DiffwithPT2(PrefixTree *pt2, Set **itemsets, int& numIts, 
				    int level, int *arr, int ntrans1, double& diff)
{
  int i;
  //double diff = 0.0;
  Set set(0);

  int count2;
  double val1, val2;
  int ntrans2 = pt2->getNumXacts();
  
  for (i = 0; i < numNodes; i++)
    {
      arr[level-1] = prefixNodes[i].item;
      if ((maximal == prefixNodes[i].status) || (candidate == prefixNodes[i].status))
	{
	  set.newArr(level, arr);
	  count2 = pt2->countForDevn(set);
	  if (count2 > 0)
	    {
	      val2 = ((double)count2)/ntrans2;
	      val1 = ((double)prefixNodes[i].count)/ntrans1;
	      AggregateFn(DeviationFn(val1, val2), diff);
	    }
	  else if (count2 == SUPERSET)
	    itemsets[numIts++] = new Set(level, arr);
	}

      if ((candidate == prefixNodes[i].status) || (PRESENT_LARGEIT == prefixNodes[i].status))
	if (NULL != prefixNodes[i].child)
	  prefixNodes[i].child->DiffwithPT2(pt2, itemsets, numIts, level+1, arr, ntrans1, diff);
    }
  set.Nullify();
  //return diff;
}
