/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include "PrefixTree.h"
#include "blockSequence.h"
#include "globals.h"
#include "trans_utils.h"
#include "2DArray.h"
#include "Set.h"

#ifndef CATCLUS
int 
PrefixTree::UpdateAfterNewBlk(BlkSequence *blkSeq)
{
  int numNewCands = 0;
  int i;
  Item a[1];
  Set **allNewIts = new (Set*)[MAX_NUMITS];
  int *allCounts = new int[MAX_NUMITS];
  
  a[0] = -1;
  Set prefix(0); // just a dummy set variable to be used in the loop.

  for (i = 0; i < numItems; i++)
    {
      if ((negBorder == root[i].status) && (root[i].count > returnMinSupport()))
	{
	  //root[i].status = maximal;
	  root[i].status = maximal;
	  IncrNumMaximalIts();
	  IncrNumLargeIts();
	  DecrNumNegBorderIts();
	  
	  // blkSeq->CheckForLargeTwoItemsets(); 
	  // such newly large 2-itemsets would have already been inserted.
	}
      //else
      if (root[i].count <= returnMinSupport())
	{ // kick all children of this itemset out
	  
	  if (maximal == root[i].status)
	    DecrNumMaximalIts();
	  DecrNumLargeIts();
	  root[i].status = negBorder;
	  IncrNumNegBorderIts();
	  
	  if (root[i].child)
	    {
	      delete root[i].child;
	      root[i].child = NULL;
	    }
	}
      
      if (root[i].child)
	  {
	    a[0] = root[i].item;
	    prefix.newArr(1,a);
	    
	    root[i].child->UpdateAfterNewBlk(this, prefix, allNewIts, numNewCands, blkSeq);
	  }
    }
  prefix.Nullify();
  
  cout << "Total number of new candidates = " << numNewCands << endl;

  if (TRUE == Lazy()) // candidates haven't been counted.
    {
      int numLargeCands = 0;

#ifdef DEBUG
      cout << "Before blkSeq->Count() while updating the prefix tree : newCands ";
      cout << numNewCands << endl;
#endif
      blkSeq->Count(allNewIts, allCounts, numNewCands);
      for (i = 0; i < numNewCands; i++)
	{
	  if (allCounts[i] >= returnMinSupport())
	    {
	      numLargeCands++;
	      insertWithCount(*(allNewIts[i]), allCounts[i], negBorder);
	    }
	  else
	    insertWithCount(*(allNewIts[i]), allCounts[i], maximal);
	  delete allNewIts[i];
	  //insertWithCount(*(allNewIts[i]), allCounts[i], negBorder);
	}
      if (numLargeCands > 0) // just recurse and repeat the whole process;
	{
	  cout << " Recursing in UpdateAfterNewBlk " << endl;
	  UpdateAfterNewBlk(blkSeq); 
	}
    }
  delete [] allNewIts;
  delete [] allCounts;
  return numNewCands;
}

PrefixNodeStatus HashNode::UpdateAfterNewBlk(PrefixTree *pt, Set &prefix, Set **allNewIts, 
				int& numNewCands, BlkSequence *blkSeq)
{
  int i,j,k;
  int tempNewCands = 0, totNumNewCands = 0;
  Set **newIts = new Set*[MAX_NUMITS];
  int *newCounts;

  PrefixNodeStatus childStatus, returnStatus = negBorder;

  // cout << " Inside HN:UpdateAfterNewBlk " << endl;
  if (FALSE == pt->Lazy())
    newCounts = new int[MAX_NUMITS];
  else
    newCounts = NULL;

  for (i = 0; i < h; i++)
    {
      if ((NULL != children) && (children[i]))
	{
	  // while ((children[i]->prefixNodes[j].child != NULL) && (j < children[i]->numNodes))
	  j = 0;
	  while (j < children[i]->numNodes)
	    {	
	      if (negBorder == children[i]->prefixNodes[j].status)
		{
		  if (children[i]->prefixNodes[j].count >= pt->returnMinSupport())
		    {  // new large itemset! generate new cands and count.
		      children[i]->prefixNodes[j].status = candidate;
		      returnStatus = candidate;
		      pt->IncrNumMaximalIts();
		      pt->IncrNumLargeIts();
		      pt->DecrNumNegBorderIts();

		      tempNewCands=pt->generateCandidatesDyn(&prefix, children[i], 
							     j, newIts);
		      if (TRUE == pt->Lazy())
			{
			  AppendNewCandIts(allNewIts, newIts, numNewCands, 
					   tempNewCands);
			  //for (k = 0; k < tempNewCands; k++)
			  // {
			  //  allNewIts[numNewCands++] = newIts[k];
			  //}
			}
		      else
			{
			  totNumNewCands += tempNewCands;
#ifdef DEBUG
			  cout << " PT.venky: Before calling Count : tempNewCands " << tempNewCands << " " << totNumNewCands << " " << numNewCands << endl;
#endif
			  if (tempNewCands > 0)
			    {
			      blkSeq->Count(newIts, newCounts, tempNewCands);
			      for (k = 0; k < tempNewCands; k++)
				{
				  if (newCounts[k] > pt->returnMinSupport())
				    {
				      pt->insertWithCount(*(newIts[k]), newCounts[k], candidate);
				      children[i]->prefixNodes[j].status = maximal;

				      //recurse on UpdateAfterNewBlk at this HashNode
				      if (children[i]->prefixNodes[j].child)
					{
					  assert(FALSE == pt->Lazy());
					  Set *newPrefix = prefix.addElement(children[i]->prefixNodes[j].item);				      
					  childStatus = UpdateAfterNewBlk(pt, *newPrefix, allNewIts, numNewCands, blkSeq);
					  children[i]->prefixNodes[j].status = childStatus;
					  delete newPrefix;
					}
				      //pt->insertWithCount(*(newIts[k]), newCounts[k], candidate);
				      returnStatus = candidate;

				      // something more here: loop once more;
				      // for now, let us assume it does not happen.
				    }
				  else
				    {
				      pt->insertWithCount(*(newIts[k]), newCounts[k], negBorder);
				      pt->IncrNumNegBorderIts();
				  }
				  delete newIts[k];
				  newIts[k] = NULL;
				  newCounts[k] = 0;
				}
			    }
			}
		      if (TRUE == pt->Lazy())
			assert(numNewCands < MAX_NUMITS);
		      else
			assert(tempNewCands < MAX_NUMITS);
		    }
		  //else, do nothing.
		}
	      else if (small != children[i]->prefixNodes[j].status)
		{
		  Set *newPrefix = prefix.addElement(children[i]->prefixNodes[j].item);
		  if (children[i]->prefixNodes[j].count <= pt->returnMinSupport())
		    { // kick all children of this itemset out

		      if (maximal == children[i]->prefixNodes[j].status)
			pt->DecrNumMaximalIts();
		      pt->DecrNumLargeIts();
		      children[i]->prefixNodes[j].status = negBorder;
		      pt->IncrNumNegBorderIts();

		      //write the itemsets and their counts to a file.
		      //pt->WriteOldIts(newPrefix,children[i]->prefixNodes[j].count);
		      children[i]->DeletePrefixNode(j);
		    }
		  else
		    {
		      if (children[i]->prefixNodes[j].child)
			{
			  childStatus = children[i]->prefixNodes[j].child->UpdateAfterNewBlk(pt,*newPrefix, 
											     allNewIts, numNewCands, blkSeq);
			  children[i]->prefixNodes[j].status = childStatus;
			  returnStatus = candidate;
			  // update numMaximalIts etc.
			}
		    }
		  delete newPrefix;
		}
	      j++;
	    }
	}
    }

  if (FALSE == pt->Lazy())
    {
      //for (i = 0; i < tempNewCands; i++)
      //delete newIts[i];
      numNewCands += totNumNewCands;
      delete [] newCounts;
    }
  delete [] newIts;

  if (negBorder == returnStatus)
    returnStatus = maximal;
  return returnStatus;
}
#endif CATCLUS





int 
PrefixTree::generateCandidatesDyn(const Set *pre,
				  PrefixNodeArray *siblings,
				  int offset, Set **newIts)
{
  int i, k;
  bool local, large;

  assert(pre->size() >= 1); // one and two itemsets are counted using arrays!

  Item item = siblings->prefixNodes[offset].item;
  Set *seedSet = pre->addElement(item);
  int newCandidateGenerated = 0;

  for (i=0; i<numItems; i++)
  {
    if (root[i].count >= minSupport && (!seedSet->member(root[i].item)))
    {
      // puting root[i].item into the set considered

      // local is true if we can insert the new candidate locally, that is,
      // in the subtree rooted at parent. Otherwise, we need to insert the
      // new candidate from the roots
      local = (root[i].item > seedSet->elements(seedSet->size()-2));

      Set newCandidate;
      newCandidate = *seedSet;
      newCandidate.insert(root[i].item);

      // the following checks if every subset derived by yanking out
      // one element from candidate is large, i.e., prune by subset

      large = true;
      int temp = newCandidate.size();

      for (k = 0; k<temp; k++)
	{
	  Item tmp;
	  newCandidate.yankOut(tmp, k);
	  if (count(newCandidate) <= returnMinSupport())
	    {
	      large = false;
	      newCandidate.putBack(tmp);
	      break;
	    }
	  newCandidate.putBack(tmp);
	}

      if (!large)
      {
	continue;
      }
      // end if prune by subset
      
      // check if this itemset is already a member of the tree: VENKY
      if (AlreadyInTree(newCandidate))
	continue;

      // copy the generated itemsets to send them out.
      newIts[newCandidateGenerated] = new Set(newCandidate);
      newCandidateGenerated++;
      
    }
  }
  delete seedSet;

  //numDash += newCandidateGenerated;
  _numNegBorderIts += newCandidateGenerated;

  return newCandidateGenerated;
}




#ifdef FANCYTID
void PrefixTree::CloseAllTIDLists(TIDListColln *tidlistColln)
{
  int i;
  int arr[MAX_IT_LENGTH];

  for (i = 0; i < numItems; i++)
    if ((root[i].count >= returnMinSupport())&&(NULL != root[i].child))
      {
	arr[0] = root[i].item;
	root[i].child->CloseAllTIDLists(2, arr, tidlistColln);
      }
}

void HashNode::CloseAllTIDLists(int level, int *arr, TIDListColln *tidlistColln)
{
  int i;
  if (NULL != children)
    for (i = 0; i < h; i++)
      {
	if (NULL != children[i])
	  children[i]->CloseAllTIDLists(level, arr, tidlistColln);
      }
}

void PrefixNodeArray::CloseAllTIDLists(int level, int *arr, TIDListColln *tidlistColln)
{
  int i;
  
  for (i = 0; i < numNodes; i++)
    {
      arr[level-1] = prefixNodes[i].item;
      if ((level == MATERIALIZED_LEVEL) && 
	  ((maximal == prefixNodes[i].status)||(candidate == prefixNodes[i].status)))
	{
	  if (NULL != prefixNodes[i].tidListPtr) 
	    {
	      //prefixNodes[i].tidListPtr->Close();
	      //delete prefixNodes[i].tidListPtr;
	      tidlistColln->StorePtrToTIDList(arr, level, prefixNodes[i].tidListPtr);
	      prefixNodes[i].tidListPtr = NULL;
	    }
	}
      else if ((level < MATERIALIZED_LEVEL) && (NULL != prefixNodes[i].child))
	prefixNodes[i].child->CloseAllTIDLists(level+1, arr, tidlistColln);
    }
}
#endif


void PrefixTree::AllocateBufSizes(int *bufSzArr, int maxBufSz)
{
  int i;
  long totSup = TotSupForOneIts();

  for (i = 0; i < numItems; i++)
    {
      //cout << root[i].count << " " << totSup << endl;
      bufSzArr[i] = (int)((maxBufSz-numItems) * ((double)root[i].count/totSup)) + 1;
    }
}





