/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef PREFIX_TREE_H
#define PREFIX_TREE_H

#include <assert.h>

#include "macros.h"
#include "Set.h"
#include "PrefixTreeTypes.h"
#include "globals.h"

#include "blockSequence.h"
#include "tid-list.h"
#include "tidlist-colln.h"

#ifdef CATCLUS
class Graph;
class ThruAttrClusters;
class ClusterColln;
#endif

class PrefixNodeArray;
class PrefixTree;
class SummarizedInfo;

//static int _st_NumPrefixNodes = 0;

class HashNode {
  int a, b, h;
  PrefixNodeArray ** children; // An array of pointers to 
			       // child arrays of PrefixNodes
  PrefixNodeArray * dead; // pointing to the array of dead branches

public:
  HashNode(int h=DEF_HASH_H, int a=DEF_HASH_A, int b=DEF_HASH_B);
  ~HashNode();

  /*
   *return true if inserted successfully, false otherwise
   */
#ifdef SAMPLING
  /*
   *VENKY: changed the default value from candidate to large
   *inline bool insert(Set &set, int count, PrefixNodeStatus pnStatus = large);
   */  
  inline bool insert(Set &set, int count, PrefixNodeStatus pnStatus = candidate);
#elif defined(DIC_ALG)
  inline bool insert(Set &set, int count, int startXaction, bool dashed, bool circle);
#else
  inline bool insert(Set &set, int count);
#endif

#ifdef FANCYTID
  void CloseAllTIDLists(int level, int *arr, TIDListColln *tidlistColln);
#endif

  /*
   *numMaxIts is the number of maximal itemsets;
   *totMaxItSup: sum of their supports
   */
  PrefixNodeStatus HashNode::MarkPossMaximalIts(int& numMaxIts, int& totMaxItSup, 
						int minSup);
  int HashNode::UnMarkNonMaximalIts(int& numUnMarked, Set& prefix, PrefixTree *pt,
				    PrefixTree *ptNew=NULL);
  int HashNode::UnMarkMaxItsContainedIn(int& numUnMarked, Set *maxIt);

  /*
   *returns the total number of itemsets that are incremented in this subtree
   */
#ifdef SAMPLING
  inline int incrementCounts(Set &transactionRemains, int level, int pass, int cnt=1,
			     IncCountsAction action = passLevelOnly, int *arr=NULL,
			     bool FirstBlk=FALSE, TIDListColln *newBlk = NULL,
			     int tid=-1);
#elif !defined(DIC_ALG)
  inline int incrementCounts(Set &transactionRemains, int level, int pass);
#else
  inline int incrementCounts(Set &transactionRemains, int level);
#endif

  //Recursively test if the set is in the subtree rooted here
  inline bool member(Set &toBeTested, int minSup = EXISTENCE_ONLY);

  // VENKY:BEGIN
  PrefixNodeStatus UpdateAfterNewBlk(PrefixTree *pt, Set& prefix, Set **allNewIts, 
				     int& numNewCands, BlkSequence *blkSeq);
  // VENKY:END

#ifndef DIC_ALG
  // allPurned is true if everything was pruned
  // return number of itemsets pruned
  inline int pruneByCount(
#ifdef SAMPLING
  // When pruneNegBorder is false, the return value reflects the number
  // of itemsets that were either pruned or marked as negBorder.
			   PruneAction action,
#endif SAMPLING
			   int support, bool &allPruned);
#endif DIC_ALG

  // return the count of set if it can be found in this subtree, 0 otherwise
  //inline int count(Set &set);
  int count(Set &set);
  int countForDevn(Set &set); // VENKY: added this one for deviations;

#ifdef DIC_ALG
  inline bool squareCount(Set &set, int minSupport);
#endif

#ifndef DIC_ALG 
  // depth tells the length of the itemsets at this level
  // e.g., those hanging directly off roots will be passed 2
  // whereAmI points to the PrefixTree in which this HashNode is
  // pre is the prefix path from root to here
  // return # of new candidates generated
#ifdef SAMPLING
  int join(int depth, PrefixTree *whereAmI, const Set &pre, JoinAction action = candidateOnly);
#else
  int join(int depth, PrefixTree *whereAmI, const Set &pre);
#endif SAMPLING
#endif DIC_ALG

  // Sets counts of all itemsets in the tree to 0.
  void clearCounts();

#ifdef SAMPLING
  // Returns true iff there exists a large itemset marked negBorder.
  bool containsLargeNegBorder(int support);
#endif

#ifdef DIC_ALG
  int M_transaction_process(PrefixTree *whereAmI, const Set &pre, int lastXaction);
#endif

  //void HashNode::AdjustITCounts(PrefixTree *pt, int *arr, int level, 
  //int& numAdjusted);
  //void HashNode::DecrCountsForDevn(Set& set, int count, int level);

  bool IncrCountsForDevn(Set& xact, int level,int *arr,Set **supportedItSets,
			 int& numItSupported);
  void HashNode::DiffwithPT2(PrefixTree *pt2, Set **itemsets, int& numIts, 
			       int level, int *arr, int ntrans1, double& diff);
  void GetNonPresentMaxIts(Set **itemsets, int& numIts, int level, int *arr);
  void UnPRESENTIts();
  void ChangeMarking(PrefixNodeStatus from, PrefixNodeStatus to);

#ifdef CATCLUS  // VENKY
  void UpdateGraphCC(Graph *oldGraph, Graph *newGraph, int *itemsetArr, 
		     int level);
  void JoinCC(int& numNew, int *prefix, int level, PrefixTree *ptO, PrefixTree *ptN, 
	      SummarizedInfo& summary, int attr, int thruAttr);
  void GenCandClustersCC(int *set1, int sizeOfSet1, int **cands, 
			 int& numCands,
			 int *set2, int level, SummarizedInfo& summary,
			 int attr, int thruAttr);
  void ExtendDistSets(SummarizedInfo& summary, int attr, int thruAttr, 
		      ThruAttrClusters& clusterSet, 
		      int *prefix, int level, PrefixTree *pt);
  void IncrementClusterCounts(Set& xact, ClusterColln& attrClusters, int level,
			      bool subspace);
#endif  

  friend class PrefixTree;

  void output(const Set *set, ostream& os);
#ifdef INSTRUMENT
  bool GetRandNegBItemset(Set **itemsetArr, int& numFilled, 
			  int numItemsets, int level, int *arr);
#endif
};

struct PrefixNode {
  Item item;
  int count;
  HashNode * child;

#ifdef FANCYTID
  TIDList *tidListPtr; // to write tid lists when incrementing counts;
#endif

  // The members below are needed only for the DIC algorithm.
#ifdef DIC_ALG
  // Dashed or solid circle/square
  // Note count>minsup => square, count<minsup => circle.
  bool dashed;
  // When we started counting the current itemset
  int startTransaction;
  bool circle; // circle or square
#endif DIC_ALG
  //#elif defined(SAMPLING)
  // Is the itemset a candidate to large, in the negative border (which is not
  // used in joining or counting during the apriori run on the sample), or
  // definitely small?
  PrefixNodeStatus status;

  PrefixNode() 
    {
      child = NULL;
      count = 0;
#ifdef FANCYTID
      tidListPtr = NULL;
#endif
    }

  ~PrefixNode() { if (NULL != child) delete child;}

};

class PrefixNodeArray {
  // The maximum number of prefix nodes that can fit without growing the array.
  int maxSize;
  // The number of prefix nodes currently in the array.
  // Every time numNodes exceeds maxSize we double maxSize and reallocate
  // prefixNodes[] of size equal to the new maxSize.
  int numNodes;

  PrefixNode * prefixNodes;

public:

  PrefixNodeArray(int maxSize=DEF_SETS_PER_LEAF);
  ~PrefixNodeArray();

#ifdef SAMPLING
  // return true if inserted successfully, false otherwise
  // VENKY: changed default value to large from candidate
  inline bool insert(Set &set, int count, PrefixNodeStatus pnStatus = candidate);
  //inline bool insert(Set &set, int count, PrefixNodeStatus pnStatus = large);
#elif defined(DIC_ALG)
  // return true if inserted successfully, false otherwise
  inline bool insert(Set &set, int count, int startXction, bool dashed, bool circle);
#else
  // return true if inserted successfully, false otherwise
  inline bool insert(Set &set, int count);
#endif

  // VENKY: to delete an itemset represented by a particular prefix node
  // and all its descendants;
  // Useful for the dynamic version and DIC;
  inline int DeletePrefixNode(int indx);

#ifdef FANCYTID  
  void CloseAllTIDLists(int level, int *arr, TIDListColln *tidlistColln);
#endif

  PrefixNodeStatus PrefixNodeArray::MarkPossMaximalIts(int &numMaxIts, int& totMaxItSup, 
						       int minSup);
  int PrefixNodeArray::UnMarkNonMaximalIts(int &numUnMarked, Set& prefix, 
					   PrefixTree *pt, PrefixTree *ptNew=NULL);
  int PrefixNodeArray::UnMarkMaxItsContainedIn(int &numUnMarked, Set *maxIt);

#ifdef SAMPLING
  inline int incrementCounts(Set &transactionRemains, int level, int pass, int cnt=1,
			     IncCountsAction action = passLevelOnly, int *arr=NULL,
			     bool FirstBlk=FALSE, TIDListColln *newBlk = NULL,
			     int tid = -1);
#elif !defined(DIC_ALG)
  inline int incrementCounts(Set &transactionRemains, int level, int pass);
#else
  inline int incrementCounts(Set &transactionRemains, int level);
#endif

  int NumNodes() { return numNodes; } // VENKY

#ifndef DIC_ALG
  // LA: swap deleted element with element #(--numNodes) (last valid prefix node).
  // destroy the entire subtree rooted at the prefix node if call pruneByCount
  // 	to subtree returns allPruned==true
  // caller needs to delete the whole array if allPruned
  // return # of itmeset pruned
  inline int pruneByCount(
#ifdef SAMPLING
  // When pruneNegBorder is false, the return value reflects the number
  // of itemsets that were either pruned or marked as negBorder.
			   PruneAction action,
#endif SAMPLING
			   int support, bool &allPruned);
#endif DIC_ALG

  inline bool member(Set &toBeTested, int minSup = EXISTENCE_ONLY);

  // return the count of set if it is fount in this subtree, 0 otherwise
  // inline int count(Set &set);
  int count(Set &set);
  int countForDevn(Set &set); // VENKY: added for deviations

#ifdef DIC_ALG
  inline bool squareCount(Set &set, int minSupport);
#endif

#ifndef DIC_ALG 
  // depth tells the length of the itemsets at this level
  // whereAmI points to the PrefixTree in which this PrefixNodeArray is
  // pre is the prefix path from root to here
  // return # of new candidates generated
#ifdef SAMPLING
  int join(int depth, PrefixTree *whereAmI, const Set &pre, JoinAction action = candidateOnly);
#else
  int join(int depth, PrefixTree *whereAmI, const Set &pre);
#endif SAMPLING
#endif DIC_ALG

  // Sets counts of all itemsets in the tree to 0.
  void clearCounts();

#ifdef SAMPLING
  // Returns true iff there exists a large itemset marked negBorder.
  bool containsLargeNegBorder(int support);
#endif

  //void PrefixNodeArray::AdjustITCounts(PrefixTree *pt, int *arr, int level,
  //			       int& numAdjusted);
  //void PrefixNodeArray::DecrCountsForDevn(Set& set, int count, int level);
  // the above two routines have not been tested;

  int PrefixNodeArray::IncrCountsForDevn(Set& xact, int level,int *arr, 
					 Set **supportedItSets,int& numItSupported);
  void PrefixNodeArray::DiffwithPT2(PrefixTree *pt2, Set **itemsets, int& numIts, 
				      int level, int *arr, int ntrans1,double& diff);
  void GetNonPresentMaxIts(Set **itemsets, int& numIts, int level, int *arr);
  void UnPRESENTIts();
  void ChangeMarking(PrefixNodeStatus from, PrefixNodeStatus to);

#ifdef CATCLUS  // VENKY
  void UpdateGraphCC(Graph *oldGraph, Graph *newGraph, int *itemsetArr, 
		     int level);
  void JoinCC(int& numNewCands, int *prefix, int level, PrefixTree *ptO, 
	      PrefixTree *ptN, 
	      SummarizedInfo& summary, int attr, int thruAttr);
  void GenCandClustersCC(int *set1, int sizeOfSet1, int **cands,
			 int& numCands, int* set2, int level, 
			 SummarizedInfo& summary, int attr, int thruAttr);
  void ExtendDistSets(SummarizedInfo& summary, int attr, int thruAttr, 
		      ThruAttrClusters& clusterSet, 
		      int *prefix, int level, PrefixTree *pt);
  void IncrementClusterCounts(Set& xact, ClusterColln& attrClusters, int level,
			      bool subspace);

#endif

  friend class PrefixTree;
  friend class HashNode;  
  void output(const Set *set, ostream& os);

#ifdef INSTRUMENT
bool GetRandNegBItemset(Set **itemsetArr, int& numFilled, 
			int numItemsets, int level, int *arr);
#endif  
};

class PrefixTree {

  /* Hash function parameters */
  int h, a, b;

  /* The total number of items in the domain */
  int numItems;

  /* The (unhashed) array of prefix nodes of size numItems */
  PrefixNode * root;

  double relSupport;
  int numXacts;
  int minSupport;

#ifndef DIC_ALG
  /* pass number tells us what itemsets need to be counted. */
  int pass;
#endif DIC_ALG

  bool _LAZY;
  char *delItsFileName;
  ofstream delItsFile;
  int _numMaximalIts;
  int _numLargeIts;
  int _numNegBorderIts;

#ifdef DIC_ALG
  /* number of dashed nodes that are in the tree */
  int numDash;
#endif 

  /* 
   *Called from within join() only
   *The only criteria is count. Ignore dashed etc.
   */
  inline bool member(Set &setOfItems) const;

  /*
   *return true if inserted successfully, false otherwise
   *This insert assumes that count=0 && dashed =true
   */
  inline bool insert(Set &setOfItems
		#ifdef DIC_ALG
		     , int startTransaction
                #endif
		     , int count = 0, PrefixNodeStatus pnStatus=negBorder);


public:

  /*
   *This constructor is called if we count 1- and 2-itemsets outside the tree.
   *count1[numItems] and count2[numItems][numItems] contain the corresponding counts.
   *Assuming all items in the domain are 0..numItems-1 (LA: Revisit)
   *support is the minimux support for an itemset to be considered large
   */
  PrefixTree(int numItems, int *count1, int *count2, double relsup, int numTrans,
	     int h=DEF_HASH_H, int a=DEF_HASH_A, int b=DEF_HASH_B, bool lazy=FALSE,
	     char *fname=NULL);

  /*
   *The following constructor takes a set of itemsets (may be larger than 2)
   *and inserts them into a new prefix tree
   */
  PrefixTree(int nitems, Set **itemsets, int numItemsets, 
	     int H=DEF_HASH_H, int A=DEF_HASH_A, int B=DEF_HASH_B);

  ~PrefixTree();

  /*
   *Some utilitarian functions
   */
  bool Lazy() { return _LAZY;}
  inline void setMinSupport(int newMinSupport) {  minSupport = newMinSupport; }
  inline void SetNumXact(int numX) { numXacts = numX; }
  inline void IncrNumXacts() { numXacts++; }


  void SetNumMaxIts(int num) {_numMaximalIts = num;}
  void SetNumLargeIts(int num) {_numLargeIts = num;}

  inline void IncrNumMaximalIts() {_numMaximalIts++;}
  inline void IncrNumLargeIts() {_numLargeIts++;}
  inline void IncrNumNegBorderIts() {_numNegBorderIts++;}

  inline void DecrNumMaximalIts(int num=1) {_numMaximalIts -= num;}
  inline void DecrNumLargeIts(int num=1) {_numLargeIts -= num;}
  inline void DecrNumNegBorderIts(int num=1) {_numNegBorderIts -= num;}


  void SetItemCountStatus(int item, int cnt, PrefixNodeStatus pnStatus)
    {
      root[item].count = cnt;
      root[item].status = pnStatus;
    }

  /*
   *increment the count of an itemset by an amt ``cnt'' if it is supported by
   *transaction.
   */
#ifdef SAMPLING
  inline int incrementCounts(Set &transaction, int cnt=1, 
			     IncCountsAction action=passLevelOnly,
			     bool FirstBlk = FALSE, 
			     TIDListColln *newBlk = NULL, int tid=-1);
#else
  inline int incrementCounts(Set &transaction);
#endif

  /*
   *The function returns a pointer to its implicit argument.  This way the
   *prefix tree can be used in place of the hash tree.  LA: Revisit.
   *join large itemsets of length pass to generate candidates of length pass+1
   *implicitly assume monotonicJoin==true
   *numCandidates is the # of new candidates generated in this join
   *
   *This function should not be called when the DIC_FLAG is set;
   */
#ifndef DIC_ALG
#ifdef SAMPLING
  PrefixTree* join(int &numCandidates, JoinAction action = candidateOnly);
#else
  PrefixTree* join(int &numCandidates) ;
#endif SAMPLING
#endif DIC_ALG

#ifndef DIC_ALG
  /*
   *parent is the hash node above siblings, which contains the last item of the
   *itemset being joined at offset offset.  We join only with elements greater
   *than item if monotonicJoin is true.  Otherwise, we join with everything.
   *offset is the offset of the PrefixNode to be joined in the array
   *pre is the set of items that are common to all item sets hanging off 
   *parent. Need to remember this to generate new candidates;
   *returns # of new candidate set generated
   */
#ifdef SAMPLING
  int join(HashNode * parent, const Set *pre, PrefixNodeArray * siblings, 
	   int offset, bool monotonicJoin, JoinAction action = candidateOnly);
#else
  int join(HashNode * parent, const Set *pre, PrefixNodeArray * siblings, 
	   int offset, bool monotonicJoin);
#endif SAMPLING
#endif DIC_ALG

#ifdef DIC_ALG
  /*
   *similar to join in the non-DIC case
   *parent is the hash node above siblings, which contains the last item of the
   *seed itemset at offset offset. 
   *offset is the offset of the seed PrefixNode in the array
   *pre is the set of items that are common to all item sets hanging off
   *parent. Need to remember this to generate new candidates
   *return # of new candidate set generated.
   */
  int generateCandidates(HashNode *parent, const Set *pre,
			 PrefixNodeArray *siblings,
			 int offset, int startTransaction);
#endif

  /* 
   *VENKY: newIts is necessary in the dynamic case when we do not 
   *insert newly generated candidates before counting them.
   *Otherwise this fn is similar to generateCandidates above;
   */
  int generateCandidatesDyn(const Set *pre,
			    PrefixNodeArray *siblings,
			    int offset, Set **newIts);
#ifdef FANCYTID
  void CloseAllTIDLists(TIDListColln *tidlistColln);
#endif


  /*
   *VENKY: the following fns are used to find the maximal itemsets;
   *We first mark a superset of maximal itemsets and then unmark those
   *which are not maximal using the UnMark* functions.
   */
  
  int PrefixTree::MarkPossMaximalIts(int& totMaxItSup);
  int PrefixTree::UnMarkNonMaximalIts(int& totUnMarkedSup, PrefixTree *ptNew=NULL);
  int PrefixTree::UnMarkMaxItsContainedIn(int& numUnMarked, Set *maxIt);

  int NumItems() { return numItems;}
  int NumMaterializedLists()
    {
#ifdef FANCYTID
      return _numMaximalIts;
#else
      return numItems;
#endif
    }
  
  void AllocateBufSizes(int *bufSzArr, int maxBufSz);


  long TotSupForOneIts()
    {
      int i;
      long sup = 0;
      for (i = 0; i < numItems; i++)
	  sup += root[i].count;
      return sup;
    }

  /*venky
   *checks if a negBorder itemset is already part of the tree
   *uses function "member" with a minsup of -1
   */
  inline bool AlreadyInTree(Set &setOfItems) const;

  // This insert takes a count, but not a starTransaction since 
  // this is used only for the dynamic lits algorithm. 
  inline bool insertWithCount(Set &setOfItems, int count, PrefixNodeStatus pnStat)
    {
      return insert(setOfItems, count, pnStat); 
    }

  int UpdateAfterNewBlk(BlkSequence *blkSeq);

  /*
   *The following functions are required for dynamically maintaining 
   *frequent itemsets; 
   *Currently, these functions are not being used.
   */
  void OpenDelItsFile();
  void WriteOldIts(Set *itemset, int count);
  void CloseDelItsFile();

#if defined(SAMPLING) && !defined(DIC_ALG)
  /*
   *When pruneNegBorder is false, the return value reflects the number
   *of itemsets that were either pruned or marked as negBorder.
   */
  inline int pruneByCount(PruneAction action);
#elif !defined(DIC_ALG)
  inline int pruneByCount();
#endif SAMPLING

  //inline int count(Set &setOfItems) const;
  int count(Set &setOfItems) const;

  /*venky
   *For computing deviations between two trees;
   *If a maximal itemset is present in the tree, then it is
   *also marked PRESENT so that it need not be counted again 
   */
  int countForDevn(Set &set) const;

  // Sets counts of all itemsets in the tree to 0.
  void clearCounts();

#ifdef SAMPLING
  // Returns true iff there exists a large itemset marked negBorder.
  bool containsLargeNegBorder();
#endif

#ifdef DIC_ALG
  // Return true only if the count of this set is no less than minSupport AND
  // it is square
  inline bool squareCount(Set &setOfItems, int minSupport);
#endif

#ifndef DIC_ALG
  inline int returnPass() {return pass;}
  void SetPassLevel(int ps) { pass = ps;}
#endif

#ifdef DIC_ALG
  inline int returnNumDash() {return numDash;}
  inline void incNumDash(int amount) {numDash+=amount;}
  inline void decNumDash(int amount) {numDash-=amount;}
#endif

  //  inline int returnMinSupport() {return minSupport;}
  inline int returnMinSupport() {return (int)(relSupport * numXacts) ;}
  inline double RelSupport() { return relSupport;}
  inline int getNumXacts() { return numXacts; }

  /* choose one from the following
  int writeToFile(FILE* or fstream*, const PrefixTree &ht);
  friend int writeToFile(FILE* or fstream*, const PrefixTree &ht);
  */

  // need to choose one from the following
  /*
  PrefixTree* readFromFile(FILE* or fstream*);
  friend PrefixTree* readFromFile(FILE* or fstream*);
  */

  // Outputs large itemsets contained in ht
  friend ostream& operator<<(ostream& os, const PrefixTree& pt);
  friend ofstream& operator<<(ofstream& os, const PrefixTree& pt);

#ifdef DIC_ALG 
  // lastTransaction is the last transaction before this stop
  // return # of new candidate generated
  // numDashRemained is the number of dashed nodes that still remain in the
  //	tree after the processing. If it is 0, the caller knows that
  //	it is time to stop.
  int M_transaction_process(int lastTransaction);
#endif

#ifndef CATCLUS
  void CountItemsets(char *datafile);
#endif

  //int PrefixTree::AdjustITCounts();
  //void PrefixTree::DecrCountsForDevn(Set& itemset, int count);
  // the above two routines have not been tested;

  int PrefixTree::IncrCountsForDevn(Set& xact, Set **supportedItSets, int *arr);
  void PrefixTree::DiffwithPT2(PrefixTree *pt2, Set **itemsets, 
			       int& numIts, double& diff);
  void GetNonPresentMaxIts(Set **itemsets, int& numIts);
  void UnPRESENTIts();
  void ChangeMarking(PrefixNodeStatus from, PrefixNodeStatus to);


#ifdef CATCLUS  // VENKY
  void UpdateGraphCC(Graph *oldGraph, Graph *newGraph);
  PrefixTree* JoinCC(int& numNew, SummarizedInfo& summary, int attr, int thruAttr);
  void GenCandClustersCC(int *set1, int sizeOfSet1, int **cands, int& numCands,
			 SummarizedInfo& summary, int attr, int thruAttr);
  void ExtendDistSets(SummarizedInfo& summary, int attr, int thruAttr, 
		      ThruAttrClusters& clusterSet);
  void IncrementClusterCounts(Set& xact, ClusterColln& attrClusters, int baseAttrVal, 
			      bool subspace);
#endif  

#ifdef INSTRUMENT
  /*
   *Randomly selects a set of itemsets from the negative border
   */
  int GetRandNegBItemsets(Set **itemsetArr, int numItemsets);
#endif

  friend void testPrefixTree(int argc, char* argv[]);
};



#ifndef __INLINE_H__ 
#define __INLINE_H__

#include "PrefixNodeArray.inline.h"
#include "HashNode.inline.h"
#include "PrefixTree.inline.h"

#endif __INLINE_H__


#endif PREFIX_TREE_H
