/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef PREFIX_TREE_INLINE
#define PREFIX_TREE_INLINE

#ifdef DIC_ALG
bool PrefixTree::squareCount(Set &set, int minSupport)
{
  Item element0 = set.elements(0);

  if (set.size() == 1)
    return ((root[element0].count >= minSupport) &&
                (!root[element0].circle));
  else if (!root[element0].child)
    return false; // set not in the tree
  else
  {
    set.suffix(1);
    bool result = root[element0].child->squareCount(set, minSupport);
    set.suffix(-1);
    return result;
  }
}
#endif

#ifndef DIC_ALG
#ifdef SAMPLING
int PrefixTree::pruneByCount(PruneAction action)
#else
int PrefixTree::pruneByCount()
#endif
{

	int numPruned = 0;
	for (int i = 0; i < numItems; i++)
	{
		if ((root[i].count < minSupport))
		{
#ifdef SAMPLING
		  switch (action) {
		  case markNegBorder :
		    root[i].status = negBorder;
		    continue;
		  case markSmall :
		    root[i].status = small;

		    // Not impossible for sampling.
		    if (root[i].child)
		      delete root[i].child;
		    root[i].child = NULL;
		    continue;
		  default :  // prune
		    // Not impossible for sampling.
		    if (root[i].child)
		      delete root[i].child;
		    root[i].child = NULL;
		    break;
		  }
#endif
			if (root[i].child)
			{
				assert(0); // impossible
				// delete root[i].child;
				// root[i].child = 0;
			}
		}
		else
		{
			root[i].status = candidate; 
			//root[i].status = large; // VENKY

			if (!root[i].child)
				continue;
			bool allPruned;
			numPruned +=
#ifdef SAMPLING
			  root[i].child->pruneByCount(action, minSupport, allPruned);
#else
			  root[i].child->pruneByCount(minSupport, allPruned);
#endif
			if (allPruned)
			{
					delete root[i].child;
					root[i].child = 0;
			}
		}
	}
	return numPruned;
}
#endif
/****
int PrefixTree::incrementCounts(Set &transaction, IncCountsAction action = passLevelOnly,
				bool SampleBlk = FALSE, bool writeTIDLists = TRUE,
				TIDListColln *newBlk = NULL, int tid=-1)
***/


//BlkCounted: if TRUE then do not increment counts;
#ifdef SAMPLING
int PrefixTree::incrementCounts(Set &transaction, int cnt, 
				IncCountsAction action,//=passLevelOnly,
				bool BlkCounted, TIDListColln *newBlk, int tid)
#else
int PrefixTree::incrementCounts(Set &transaction)
#endif
{
  int incCount = 0;
  int arr[MAX_IT_LENGTH];

  int temp = transaction.size();

  for(int index = 0; index < temp; index++) 
    {
      Item indexElement = transaction.elements(index);
      arr[0] = indexElement;

#ifdef SAMPLING
      // For sampling, increment counts of even the root (singleton) itemsets.
      if ((FALSE == BlkCounted) && (action == allLevels))
	{
	  root[indexElement].count += cnt;
	}
#endif

      if(root[indexElement].child) 
	{
	  // The subtree is not empty
	  
	  transaction.suffix(index+1);
#ifdef SAMPLING
	  incCount += root[indexElement].child->incrementCounts(transaction, 2 , pass, 1, action, arr,
								BlkCounted,newBlk,tid);
#elif !defined(DIC_ALG)
	  incCount += root[indexElement].child->incrementCounts(transaction, 2 , pass);
#else
	  incCount += root[indexElement].child->incrementCounts(transaction, 2 );
#endif
	  
	  transaction.suffix(-index - 1);
	}
    }
  return incCount;
}


/****
     Junk from the above function;

	      if ((tid != -1) && (FALSE == SampleBlk) && (TRUE == writeTIDLists))
	    {
	    #ifdef FANCYTID
	      if ((maximal == root[indexElement].status) ||
		  (negBorder == root[indexElement].status)) // || small
		{
		  if (NULL == root[indexElement].tidListPtr)
		    {
		      // insert this itemset in the hash table and store the
		      // the pointer to the tid list in tidListPtr;
		      //root[indexElement].tidListPtr = newBlk->CreateTIDListToWrite(arr, 1);
		      //assert(NULL != root[indexElement].tidListPtr);
		    }
		  // directly access TID list;
		  root[indexElement].tidListPtr->WriteTID(tid); 
		}
#endif

//newBlk->WriteXact(&transaction, tid);
	    }
******/
// assuming count = 0 && dashed == true && circle = true
bool PrefixTree::insert(Set &set
#ifdef DIC_ALG
			, int startTransaction
#endif
			, int count, PrefixNodeStatus pnStatus)
{
  // since we start from 3-itemsets; Not always now 
  // Venky: Sometimes, we insert 2-itemsets as well.
  assert(set.size() >= 2);
  int index = set.elements(0);
  if (NULL == root[index].child)
    {
      // this should never happen if we start from 3-sets
      //assert(0); (VENKY - because neg border 2-itemsets are maintained in
      // an array. This situation is possible.)
      
      // but if we start from 2 or 1, we should:
      root[index].child = new HashNode(h, a, b);
    }
  
  set.suffix(1);
  bool inserted;
#ifndef DIC_ALG
#ifdef SAMPLING
  if (maximal == pnStatus)
    root[index].status = candidate;
  else
    root[index].status = pnStatus;
  
  inserted = root[index].child->insert(set, count, pnStatus);
#else
  inserted = root[index].child->insert(set, count);
#endif
#else
  inserted = root[index].child->insert(set, count, startTransaction, true, true);
#endif //#ifndef
  set.suffix(-1);

  return inserted;
}

// VENKY
// doesn't care for the count and minSupport as done by member.
bool PrefixTree::AlreadyInTree(Set &setOfItems) const
{
  PrefixNode *prefixNode;
  
  prefixNode = &(root[setOfItems.elements(0)]);
  if (!prefixNode->child)
    return false;

  if (setOfItems.size() == 1)
    return true;
  else
    {
      setOfItems.suffix(1);
      bool result = prefixNode->child->member(setOfItems, EXISTENCE_ONLY);
      setOfItems.suffix(-1);
      return result;
    }
}

bool PrefixTree::member(Set &setOfItems) const
{
	PrefixNode *prefixNode;

	prefixNode = &(root[setOfItems.elements(0)]);
	if ((prefixNode->count < minSupport) || (!prefixNode->child))
		return false;

	if (setOfItems.size() == 1)
		return true;
	else
	{
	  setOfItems.suffix(1);
	  bool result = prefixNode->child->member(setOfItems, minSupport);
	  setOfItems.suffix(-1);
	  return result;
	}
}

#endif





