/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include "PrefixTree.h"
#include "trans_utils.h"


int PrefixNodeArray::count(Set &set)
{
  int cnt = 0;

  for (int i=0; i<numNodes; i++)
  {
    if (prefixNodes[i].item == set.elements(0))
      {
	if (set.size() == 1)
	  cnt = prefixNodes[i].count;
	else if (NULL == prefixNodes[i].child)
	  cnt = 0;
	else
	  {
	    set.suffix(1);
	    cnt = prefixNodes[i].child->count(set);
	    set.suffix(-1);
	  }
	break; //VENKY
      }
  }
  return cnt;
}





int HashNode::count(Set &set)
{
  int bucket;
  
  bucket = HASH(a, b, h, set.elements(0));

  //cout << " bucket " << bucket << endl;
  if ((NULL == children) || (!children[bucket]))  // VENKY
    return 0;
  else
    return children[bucket]->count(set);
}

int PrefixTree::count(Set &set) const
{
  Item element0 = set.elements(0);

  if (set.size() == 1)
    return root[element0].count;
  else if (!root[element0].child)
    return 0; // set not in the tree
  else
  {
    set.suffix(1);
    int cnt = root[element0].child->count(set);
    set.suffix(-1);
    return cnt;
  }
  return 0;
}



/*
 *For computing deviations between two trees;
 *If a maximal itemset is present in the tree, then it is
 *also marked PRESENT so that it need not be counted again 
 */

int PrefixTree::countForDevn(Set &set) const
{
  Item element0 = set.elements(0);

  if (set.size() == 1)
    {
      if (maximal == root[element0].status)
	{
	  root[element0].status = PRESENT_MAXIT;
	  return root[element0].count;
	}
      else if (candidate == root[element0].status)
	{
	  root[element0].status = PRESENT_LARGEIT;
	  return root[element0].count;
	}
      else if (negBorder == root[element0].status)
	return root[element0].count;
    }
  else if (!root[element0].child)
    return SUPERSET; // set not in the tree
  else
    {
      set.suffix(1);
      int cnt = root[element0].child->countForDevn(set);
      set.suffix(-1);
      return cnt;
    }
  return 0;
}



int PrefixNodeArray::countForDevn(Set &set)
{
  int cnt = 0;
  int i;
  for (i=0; i<numNodes; i++)
    {
      if (prefixNodes[i].item == set.elements(0))
	{
	  if (set.size() == 1)
	    {
	      cnt = prefixNodes[i].count;
	      if (maximal == prefixNodes[i].status)
		{
		  prefixNodes[i].status = PRESENT_MAXIT;
		  return prefixNodes[i].count;
		}
	      else if (candidate == prefixNodes[i].status)
		{
		  prefixNodes[i].status = PRESENT_LARGEIT;
		  return prefixNodes[i].count;
		}
	      else if (negBorder == prefixNodes[i].status)
		return prefixNodes[i].count;
	    }
	  else if (NULL == prefixNodes[i].child)
	    return SUPERSET; // set not in the tree;
	  else
	    {
	      set.suffix(1);
	      cnt = prefixNodes[i].child->countForDevn(set);
	      set.suffix(-1);
	    }
	  break; //VENKY
	}
    }
  return cnt;
}





int HashNode::countForDevn(Set &set)
{
  int bucket;
  
  bucket = HASH(a, b, h, set.elements(0));

  //cout << " bucket " << bucket << endl;
  if ((NULL == children) || (!children[bucket]))  // VENKY
    return 0;
  else
    return children[bucket]->countForDevn(set);
}



// if set supports X, and no superset of X then the count
// of X is incremented. 
int PrefixTree::IncrCountsForDevn(Set& xact, Set **supportedItSets, int *arr)
{
  int incCount = 0;
  //int arr[MAX_IT_LENGTH];
  //Set **supportedItSets;
  int numItSupported = 0;

  bool incFlag;
  int indx, i;
  Item indxElement;

  for (indx = 0; indx < xact.size(); indx++)
    {
      incFlag = FALSE;
      indxElement = xact.elements(indx);
      arr[0] = indxElement;
      
      if ((root[indxElement].child != NULL) && (root[indxElement].status != maximal))
	{
	  xact.suffix(indx+1);
	  incFlag = root[indxElement].child->IncrCountsForDevn(xact, 2, arr, supportedItSets, numItSupported);
	  xact.suffix(-indx-1);
	}

      if (FALSE == incFlag) // something below is already incremented or a subset;
	{
	  if (((root[indxElement].status == maximal) || (root[indxElement].status == candidate)) && (FALSE == IsSubsetOf(arr, 1, supportedItSets, numItSupported)))
	    {
	      root[indxElement].count++;
	      incCount++;
	    }
	}
      else
	incCount++;
    }
  assert(numItSupported < MAX_SUPCNT_PERXACT);

  for (i = 0; i < numItSupported; i++)
    delete supportedItSets[i];

  return incCount;
}

bool HashNode::IncrCountsForDevn(Set& xact, int level, int *arr, Set **supportedItSets,
				 int& numItSupported)
{
  int link, indx, incCount=0;

  for (indx = 0; indx < xact.size(); indx++)
    {
      link = HASH(a, b, h, xact.elements(indx));
      if (NULL != children[link])
	{
	  xact.suffix(indx);
	  incCount += children[link]->IncrCountsForDevn(xact, level, arr, 
						       supportedItSets, numItSupported);
	  xact.suffix(-indx);
	}
    }
  return ((incCount > 0) ? TRUE : FALSE);
}

int PrefixNodeArray::IncrCountsForDevn(Set& xact, int level, int *arr, 
				       Set **supportedItSets, int& numItSupported)
{
  assert(level < MAX_IT_LENGTH);
  int i;
  int incCount = 0;
  bool incFlag = FALSE;

  for (i = 0; i < numNodes; i++)
    if (prefixNodes[i].item == xact.elements(0))
      {
	incFlag = FALSE;
	arr[level-1] = prefixNodes[i].item;
	if ((prefixNodes[i].child != NULL) && (prefixNodes[i].status == candidate))
	  //&& (prefixNodes[i].status != negBorder))
	  {
	    xact.suffix(1);
	    incFlag = prefixNodes[i].child->IncrCountsForDevn(xact, level+1, arr, 
							      supportedItSets, numItSupported);
	    xact.suffix(-1);
	  }
	if (FALSE == incFlag)
	  {
	    if (((prefixNodes[i].status == maximal) || (prefixNodes[i].status == candidate)) && (FALSE == IsSubsetOf(arr, level, supportedItSets, numItSupported)))
	      {
		prefixNodes[i].count++;
		AppendNewIt(supportedItSets, numItSupported, arr, level);
		incCount++;
	      }
	  }
	else //just to indicate that something below has been updated;
	  incCount++;
      }
  return incCount;
}


/*********
// should be similar to increment counts;
void PrefixTree::DecrCountsForDevn(Set& set, int count)
{
  int indx;
  Item indxElement;

  for (indx = 0; indx < set.size(); indx++)
    {
      indxElement = set.elements(indx);
      if (negBorder == root[indxElement].status)
	continue;

      root[indxElement].count -= count;
      assert(root[indxElement].count >= 0);

      // the inserted itemset should not contain a maximal itemset
      // or something that has already been adjusted;
      assert(root[indxElement].status != maximal);
      assert(root[indxElement].status != ADJUSTED_MAXIT);

      if (root[indxElement].child != NULL)
	{
	  set.suffix(indx+1);
	  root[indxElement].child->DecrCountsForDevn(set, count, 2);
	  set.suffix(-indx-1);
	}
    }
}

void HashNode::DecrCountsForDevn(Set& set, int count, int level)
{
  int link, indx;
  
  for (indx = 0; indx < set.size(); indx++)
    {
      link = HASH(a, b, h, set.elements(indx));
      if (NULL != children[link])
	{
	  set.suffix(indx);
	  children[link]->DecrCountsForDevn(set, count, level);
	  set.suffix(-indx);
	}
    }
}

void PrefixNodeArray::DecrCountsForDevn(Set& set, int count, int level)
{
  assert(level < MAX_IT_LENGTH);
  int i;

  for (i = 0; i < numNodes; i++)
    if (prefixNodes[i].item == set.elements(0))
      {
	switch(prefixNodes[i].status)
	  {
	  case maximal:
	  case ADJUSTED_LARGEIT:
	    //assert(0);
	    break;

	  case candidate:
	    if (level == set.size()-1)
	      prefixNodes[i].status = TO_BE_ADJUSTED;
	    prefixNodes[i].count -= count;
	    break;

	  case TO_BE_ADJUSTED:
	    if (level < set.size()-1)
	      prefixNodes[i].status = UNADJUSTABLE_AS_YET;
	    prefixNodes[i].count -= count;
	    break;

	  case UNADJUSTABLE_AS_YET:
	    prefixNodes[i].count -= count;
	    break;
	  }
	
	if ((set.size() > 1) && (prefixNodes[i].child != NULL) && (prefixNodes[i].status != maximal) && (prefixNodes[i].status != ADJUSTED_LARGEIT))
	  {
	    set.suffix(1);
	    prefixNodes[i].child->DecrCountsForDevn(set, count, level+1);
	    set.suffix(-1);
	  }
      }
}

*********/
