/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <fstream.h>
#include <iostream.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "globals.h"
#include "Set.h"
#include "sample.h"
#include "trans_utils.h"

const int MAXBLKSIZE=100000;

static int longCompare(const void *i, const void *j)
{
  return (*(long*)i - *(long*)j);
}

int WriteSample(ofstream& ofile, Set *sample[], int sampleSize)
{
  int i, j, tmp = -1;
  for (i = 0; i < sampleSize; i++)
    {
      tmp = sample[i]->size();
#ifdef ASCII
      //ofile << tmp << " ";
#else
      ofile.write(&i, sizeof(int));
      ofile.write(&tmp, sizeof(int));
#endif
      for (j = 0; j < sample[i]->size(); j++)
	{
	  tmp = sample[i]->elements(j);
#ifdef ASCII
	  ofile << tmp << " ";
#else	  
	  ofile.write(&tmp, sizeof(int));
#endif	  
	}
      tmp = -1;
#ifdef ASCII
      ofile << endl;
#else     
      ofile.write(&tmp, sizeof(int));
#endif
      //cout << sample[i]->size() << " ";
      //cout << *(sample[i]) << endl;
    }

  return i;
}


// the following function generates only one sample without replacement.
int SampleDataFileWithoutRepl(char *datafile, char *outFilePrefix, int sampleSize,
			      int seed, int numTrans)
{
  ifstream ifile;
  

  double prob;
  double drandNumber;

  srand48(seed);
  prob = (double)(sampleSize)/(double)numTrans;
  
  char buffer[TRANS_BLOCK_SIZE];
  int tid = 0;
  Set xactIter(0);
  int numSampled = 0;
  Set **sample = new (Set*)[sampleSize*2];
  ofstream ofile;

  ofile.open(outFilePrefix);

  memset((void*)sample, 0, sizeof(Set*)*sampleSize);
  SetupXactScan(datafile, buffer);
  while (getNextXact(xactIter, tid, buffer) != -1)
    {
      drandNumber = drand48();
      if (drandNumber <= prob)
	{
	  sample[numSampled++] = new Set(xactIter);
	}
    }

  WriteSample(ofile, sample, numSampled);
  ofile.close();
  cout << "******** NumSampled ****** " << numSampled << " : " << prob << " : " << endl;

  int i;
  for (i = 0; i < numSampled; i++)
    delete sample[i];

  delete [] sample;
  xactIter.Nullify();

  return numSampled;
}

// the following function generates generate numSamples number of samples with replacement.
void SampleDataFileWithRepl(char *datafile, char *outFilePrefix, int sampleSize, 
			    int numSamples, int randNum)
{
  Set **sample = new (Set*)[MAXBLKSIZE];
  long *sampleOffsets = new long[sampleSize];

  char outFileName[MAX_NAME_LENGTH];
  ifstream ifile;
  int sampNum;
  int i;
  long blkSize;
  long temp = 0;
  int numXacts = 0;
  ofstream ofile;
  int fd;
  long fileSize;

  fd = open(datafile, O_RDONLY);
  fileSize = lseek(fd, 0, SEEK_END);
#ifdef DEBUG
  cout << " File Size : Sample size " << fileSize << " : " << sampleSize << endl;
#endif
  close(fd);


  srandom(randNum);
  for (sampNum = 0; sampNum < numSamples; sampNum++)
    {
      getRandomOffsets(sampleOffsets, sampleSize, fileSize);
      qsort((void*)sampleOffsets, sampleSize, sizeof(long), longCompare);

      if (1 == numSamples)
	strcpy(outFileName, outFilePrefix);
      else
	sprintf(outFileName, "%s.sample%d", outFilePrefix, sampNum);
      ofile.open(outFileName);

      ifile.open(datafile);
      if (sampleSize <= MAXBLKSIZE)
	{
	  numXacts = readSampleRandom(ifile, sampleOffsets, sample, sampleSize);
	  WriteSample(ofile, sample, numXacts);
	  for (i = 0; i < numXacts; i++)
	    delete sample[i];
	}
      else
 	{
	  //assert(0);

	  long *tempSampOffsets = sampleOffsets;

	  blkSize = MAXBLKSIZE;
	  temp = 0;
	  while (temp < sampleSize)
	    {
	      numXacts = readSampleRandom(ifile, tempSampOffsets, sample, blkSize);
	      WriteSample(ofile, sample, numXacts);
	      for (i = 0; i < numXacts; i++)
		delete sample[i];
	      
	      temp += blkSize;
	      tempSampOffsets += blkSize;
	      
	      if (temp + blkSize > sampleSize)
		blkSize = sampleSize-temp;
	      else
		blkSize = MAXBLKSIZE;
	    }
	}
      ifile.close();
      ofile.close();
    }

  delete [] sample;
  delete [] sampleOffsets;
}


/****
     cout << temp << " " << blkSize << " " << tempSampOffsets[blkSize] << endl;
     cout << "Read numxacts " << numXacts << endl;
     cout << tempSampOffsets[blkSize] << " " << tempSampOffsets[blkSize+1]<< endl;
****/






