/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <stdlib.h>
#include <fstream.h>

#include "Set.h"
#include "macros.h"
#include "srcFunctions.h"
#include "trans_utils.h"

#include "PrefixTree.h"
//#include "HashTree.h"

int apriori(Tree * T[],
	    //#ifdef SAMPLING
	    //	    Set * sample[], int sampleSize) {
	    //#else
	    char * fileName, int numItems) {
  //#endif

  ifstream xfile;
  int setLength = 2;

  int curTree = setLength;

  while (1) {
		/* loop and in each iteration, do
		   1. join
		   2. incrementCount
		   3. pruneByCount

		   exit when join doesn't produce anything
		*/

    int numNew = 0;

#ifdef HASH_TREE
    // Compute the join using the tree for the currently processed
    // set length.  Write the pointer for the following tree into
    // the next location in the array.

    numNew = T[curTree]->join(T[curTree + 1]);
    curTree = setLength + 1;


#else
    // For sampling the default for the 2nd argument is candidateOnly.

    T[curTree] = T[curTree]->join(numNew);

#endif
#ifdef DEBUG
    cout << " Number of new candidates " << numNew << endl;
#endif

    if (!numNew) {
#ifdef HASH_TREE
      // Delete the just created (empty tree).
      delete T[curTree];
#endif
      break;
    }

    /**
#ifdef SAMPLING
    for (int index = 0; index < sampleSize; index++) {
	Set transaction(*sample[index]);
	T[curTree]->incrementCounts(transaction);
#else
    **/
    xfile.close();

#ifdef ASCII
    xfile.open(fileName);
    if (!xfile) error(false, "apriori", "cannot open input file", fileName);
    
    int strSize = MAX_TRANS_BYTES(numItems);
    char buf[strSize];

    xfile.getline(buf, strSize);

    while (xfile) {
      Set transaction(buf);
      //T[curTree]->incrementCounts(transaction);
      xfile.getline(buf, strSize);
#else
      int tid = 0;
      char buffer[TRANS_BLOCK_SIZE];
      Set transaction(0);
      
      SetupXactScan(fileName, buffer);
      while (getNextXact(transaction, tid, buffer) != -1) {
#endif  // ASCII
#ifdef SAMPLING
	T[curTree]->incrementCounts(transaction, passLevelOnly); //VENKY
#else
	T[curTree]->incrementCounts(transaction);
#endif
      }
#ifndef ASCII
      transaction.Nullify();
#endif

#if !defined(SAMPLING) && defined(ASCII)
    if (buf[0] != '\0')
      cerr << "no end of line before end of file in xfile" << endl
	   << "transaction " << buf << " ignored!" << endl;
#endif  // !SAMPLING && ASCII

#ifdef SAMPLING
    // Don't prune negative border itemsets, just mark them, so they are not used for
    // new candidate generation.  When the argument is markNegBorder, the return value
    // reflects the number of itemsets that were either pruned or marked as negBorder.
    // This way we can avoid doing one extra join to find out that we're done.
    int numPruned = T[curTree]->pruneByCount(markNegBorder);
#else
    int numPruned = T[curTree]->pruneByCount();
#endif

    if (numNew == numPruned) {
#ifdef HASH_TREE
      // Delete the pruned empty tree.
      delete T[curTree];
#endif
      break;
    }

    // There are large sets of length up to setLength+1, so increment setLength.
    setLength++;

  }  // end of while (1) loop.

    return setLength;
}

