/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <math.h>
#include <stdlib.h>
#include "Set.h"
#include "macros.h"
#include "srcFunctions.h"

#include "PrefixTree.h"

// Reads sampleSize number of transactions from the beginning of the file, and
// saves pointers to the sets corresponding to the transactions in sample[].
int readSampleStart(ifstream & xfile, Set * sample[], int sampleSize) {
    int numRead;

    int xNumber = 0;
    //xfile.read(&xNumber, sizeof(xNumber));

    for (numRead = 0; numRead < sampleSize && xfile; numRead++) {
	xfile.read(&xNumber, sizeof(xNumber));
	sample[numRead] = new Set(xfile);
	xfile.read(&xNumber, sizeof(xNumber)); // -1;
	xfile.read(&xNumber, sizeof(xNumber)); // next tid;
    }
    return numRead;
}

// Puts sampleSize random nonnegative integers up to fileSize into
// the array sampleOffsets[].  The integers are multiples of 4.
void getRandomOffsets(long sampleOffsets[], int sampleSize, long fileSize) {
  //srandom(1);

  // Make the random offsets obtained below a multiple of 4 bytes.
  // I use srandom() and random() instead of srand() and rand()
  // because the former have better randomization characteristics.
  // The speed of the the former is about 2/3 of the latter.
  for (int index = 0; index < sampleSize; index++)
    sampleOffsets[index] = (random() % fileSize) & (long)~0x3;
}

// Reads up to sampleSize number of transactions from the file at offsets
// specified in array sampleOffsets[].  The offsets in the array are sorted.
// Some offsets at the end may point past the start of the last transaction.
// The return value specifies the actual number of transactions read.
// Pointers to the sets corresponding to the transactions read are saved in
// array sample[].
int readSampleRandom(ifstream & xfile, long sampleOffsets[],
                     Set * sample[], int sampleSize) {
  int numRead;

  for (numRead = 0; numRead < sampleSize; numRead++) {

    // Seek to this offset relative to the beginning of the file.
    xfile.seekg(sampleOffsets[numRead], ios::beg);

    // Look for the -1 transaction separator.
    int xNumber = 0;
    while ((xNumber != -1) && xfile)
      xfile.read(&xNumber, sizeof(int));

    if (!xfile) {
	cerr << "sampling: Seeked past the start of the last xaction" << endl;
	// The remaining offsets point too far.  Stop collecting the sample.
	// Just adjust lowMinSupport to reflect the lower sampleSize.
	break;
      }

    // We found the separator.  There must be another transaction.
    // Read in the transaction number.  Then input the set.
    
    xfile.read(&xNumber, sizeof(xNumber)); // tid
    if (xfile)
      sample[numRead] = new Set(xfile);
    else
      {
	numRead--;
	assert(numRead > 0);
      }
  }
  return numRead;
}

/***
// The Tree passed in below must always be a PrefixTree.
int countAndPrune(PrefixTree * pt, char * fileName) {

    // Clear all counts before the first counting over the whole transaction file.
    pt->clearCounts();

    ifstream xfile(fileName);

    if (!xfile) error(false, "sampling", "cannot open input file", fileName);

    int xNumber = 0;
    xfile.read(&xNumber, sizeof(xNumber));

    while (xfile) {
      xfile.read(&xNumber, sizeof(xNumber));
      Set transaction(xfile);
      pt->incrementCounts(transaction, allLevels);
      xfile.read(&xNumber, sizeof(xNumber));
    }

    if (!pt->containsLargeNegBorder()) {
      // Happy, happy, joy, joy!!!
      // No negative border itemsets were found to be large (frequent), so no
      // more counting over the whole transaction files are needed.  Clean up
      // the tree and return.
      pt->pruneByCount(prune);
      return 0;
    }

    // A negative border itemset was found to be frequent.  Now we know we have to
    // go the second pass over the transaction file.  During pruning mark certain
    // prefixes as dead (PrefixNodeStatus of small), and delete everything below them.
    // Then loop S := S U NegBorder(S) will finish faster.
    cerr << "sampling: algorithm 2 kicking in.  Pass 2 begins." << endl;
    pt->pruneByCount(markSmall);

    int numNew;
    do
      pt->join(numNew, nonSmallNegBorder);
    while (numNew);


    // Clear all counts before the second counting over the whole transaction file.
    pt->clearCounts();

    // Start the second pass over the transaction file.
    xfile.close();
    xfile.open(fileName);

    if (!xfile) error(false, "sampling", "cannot open input file", fileName);

    xfile.read(&xNumber, sizeof(xNumber));

    while (xfile) {
      xfile.read(&xNumber, sizeof(xNumber));
      Set transaction(xfile);
      pt->incrementCounts(transaction, allLevels);
      xfile.read(&xNumber, sizeof(xNumber));
    }

    // Now we couldn't have missed anything.  Clean up.
    pt->pruneByCount(prune);

    return 1;  // Had to go to the second complete transaction file reading.
}
****/
