/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include <fstream.h>

#include "PrefixTree.h"
#include "blockSequence.h"

// PT1 and PT2;
// a maximal itemset in PT1: 
// can be maximal in PT2: obtain the difference. -- symmetric.
// the next two are anti-symmetric, i.e., if (a) happens from PT1, then 
// (b) happens if we look at the same itemset from PT2;
// (a) can be large but not maximal; - ignore;
// (b) can be small; - count later;


// deviation between two blocks: blkSeq1 and blkSeq2 through PT1 and PT2;
// 
double FindDeviation(PrefixTree *pt1, PrefixTree *pt2, BlkSeq *blkSeq1, BlkSeq *blkSeq2)
{
  double diff = 0;
  Set **itSetsOf1in2 = new (Set*)[MAX_SETCNT];
  int *countsOf1in2 = new int[MAX_SETCNT];
  Set **itSetsOf2in1 = new (Set*)[MAX_SETCNT];
  int *countsOf2in1 = new int[MAX_SETCNT];
  int num1in2, num2in1;
  int i;
  
  diff += ComparePT1withPT2(pt1, pt2, itSetsOf1in2, num1in2);
  blkSeq2->Count(itSetsOf1in2, countsOf1in2, num1in2);
  
  num2in1 = PT2->GetUnMarkedMaxIts(itSetsOf2in1, num2in1);
  blkSeq1->Count(itSetsOf2in1, countsOf2in1, num2in1);

  double val = 0;
  for (i = 0; i < num1in2; i++)
    {
      val = pt1->Count(itSetsOf1in2[i]);
      diff += (val > countsOf1in2[i])? (val-countsOf1in2[i]) : (countsOf1in2[i]-val);
    }
  for (i = 0; i < num2in1; i++)
    {
      val = pt2->Count(itSetsOf2in1[i]);
      diff += (val > countsOf2in1[i])? (val-countsOf2in1[i]) : (countsOf2in1[i]-val);
    }
  return diff;
}

// why?  A maximal itemset in PT1: Is it a maximal itemset in PT2? If
// so, obtain the difference in their counts; else put it in itSetsOf1in2;
// mark some of these as present in pt2;
double ComparePT1withPT2(PrefixTree *pt1, PrefixTree *pt2, Set **itSetsOf1in2,
			 int& num1in2)
{
  double diff = 0;
  diff += pt1->DiffwithPT(pt2, itSetsOf1in2, num1in2);
  return diff;
}

//GetUnMarkedMaxIts;
//DiffwithPT
//UnMarkMaxIts;

//main 
// construct a prefix tree over the entire dataset;
// take samples and then construct trees over each of them, 
// and then measure deviations with each of these; 
// aside: write a sample.sh to generate a number of samples
// from the same datafile;
void DeviationsFromSamples(char *prefixName, int numSamples, char *datafile)
{

  PrefixTree *pt1, *pt2;
  char sampleFileName[MAX_NAME_LENGTH];
  int i;

  apriori(datafile, pt1);
  for (i = 0; i < numSamples; i++)
    {
      sprintf(sampleFileName, "%s.sample%d", prefixName, i);
      // call apriori on each of the samples; 
      apriori(sampleFileName, pt2);
      FindDeviation(pt1, pt2);
    }
}
