/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <assert.h>
#include <iomanip.h>
#include "globals.h"

const int formatInput = 1;
const int wilcoxTest = 2;

static int doubleCompare(double *i, double *j)
{
  return ((*i >= *j) ? 1 : 0);
}

double get(int i, int j, int rowSz, double *arr)
{
  return arr[i*rowSz + j];
}

void set(int i, int j, int rowSz, double *arr, double val)
{
  arr[i*rowSz+j] = val;
}

// read the avg at the end of the sample array;
void ReadNextSampleNums(ifstream& ifile, int sampleNum, int sampleSize, 
			double *sample, 
			double &sampleFrac, int &dataSize)
{
  int i;
  double val;

  ifile >> sampleFrac;
  ifile >> dataSize;

  for (i = 0; i < sampleSize; i++)
    {
      ifile >> val;
      set(sampleNum, i, sampleSize, sample, val);
      assert(ifile != NULL);
    }
  ifile >> val;
  //set(sampleNum, sampleSize, sampleSize+1, sample, val);

  qsort((void*)(sample+sampleNum*sampleSize), sampleSize, sizeof(double), doubleCompare);

  /***
  cout.precision(4);
  cout.setf(ios::fixed, ios::floatfield);
  for (i = 0; i < sampleSize; i++)
    { 
      cout << get(sampleNum, i, sampleSize, sample) << " " ;
    }
  cout << endl << endl;
  ***/
}


void 
FormattedOutput(ifstream& ifile,
		int numSamples, 
		int sampleSize, 
		double *samples,
		ofstream& ofile,
		int precision)
{
  int i, j;
  double sampleFrac; // 1st arg in each line;
  int dataSize;  // second arg in each line;
  double avg;    // last arg in each line;
  double val;  // value at the end of each line;

  ofile.precision(precision);
  ofile.setf(ios::fixed, ios::floatfield);
  for (i = 0; i < numSamples; i++)
    {
      assert(ifile != NULL);
      ifile >> sampleFrac;
      ofile << sampleFrac << " ";

      ifile >> dataSize;
      ofile << dataSize << " ";

      for (j = 0; j < sampleSize; j++)
	{
	  ifile >> val;
	  ofile << val << " ";
	}
      ifile >> avg;
      ofile << avg;
      ofile << endl;      
    }
}



double WilcoxTest(double *samples, int samp1, int samp2, int sampleSize)
{
  double mean = (double)(sampleSize*sampleSize)/2;
  double variance = (double)(sampleSize*sampleSize)*(2*sampleSize+1)/12;

  int i, j;
  int u = 0;
  double val1, val2;

  for (i = 0; i < sampleSize; i++)
    {
      val1 = get(samp1, i, sampleSize, samples);
      for (j = 0; j < sampleSize; j++)
	{
	  val2 = get(samp2, j, sampleSize, samples);
	  //	  cout << val1 << " " << val2 << endl;
	  if (val1 < val2)
	    u++;
	}
    }

  cout << " " << u << " : " << mean << " " << ((double)u-mean)/(sqrt(variance)) <<  endl;


  return ((double)u-mean)/(sqrt(variance));
}


void Wilcoxon(double *samples, int numSamples, int sampleSize)
{
  double significance;
  int i;

  for (i = 0; i < numSamples-1; i++)
    {
      significance = WilcoxTest(samples, i, i+1, sampleSize);
      cout << i << " " << i+1 << " : " << significance << endl;

      /***
      for (j = 0; j < sampleSize; j++)
	{
	  diffVals[j] = get(i, j, sampleSize, samples) - get(i+1, j, sampleSize, samples);

	  //cout << diffVals[j] << " ";
	}
      cout << endl;
      ***/
    }
}



int main(int argc, char **argv)
{
  if (argc < 5)
    {
      cout << "Usage: wilcox <numSamples> <sampleSz> <inputfile> <outputfile> <option=[1:format, 2:wilcoxTest]>, [precision]" << endl;
      exit(1);
    }
  
  int  numSamples=0;
  int  sampleSize=0;
  char inputFileName[MAX_NAME_LENGTH];
  char outputFileName[MAX_NAME_LENGTH];
  int  option=2;
  int  precision=-1;

  ifstream ifile;
  ofstream ofile;

  numSamples = atoi(argv[1]);
  sampleSize = atoi(argv[2]);
  strcpy(inputFileName, argv[3]);
  strcpy(outputFileName, argv[4]);
  option = atoi(argv[5]);

  if (option == formatInput)
    {
      assert(argc == 6);
      precision = atoi(argv[6]);
    }

  ifile.open(inputFileName);
  assert(ifile != NULL);


  double *samples = new (double)[numSamples*sampleSize];

  int i;
  int dataSz;
  double frac;

  cout << numSamples << " " << sampleSize << endl;
  for (i = 0; i < numSamples; i++)
    {
      ReadNextSampleNums(ifile, i, sampleSize, samples, frac, dataSz);
    }
  ifile.close();

  if (option == formatInput)
    {
      ifile.open(inputFileName);
      ofile.open(outputFileName);
      FormattedOutput(ifile, numSamples, sampleSize, samples, ofile, precision);
      ifile.close();
      ofile.close();
    }
  if (option == wilcoxTest)
    Wilcoxon(samples, numSamples, sampleSize);
}
