/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#include "Set.h"
#include "PrefixTree.h"
#include "blockSequence.h"
#include "tidlist-colln.h"
#include "tid-list.h"
#include "trans_utils.h"
#include "timeutil.h"

BlkSequence::BlkSequence(bool WriteListFlag, int *counts2, int numItems)
{
  _firstBlk = NULL;
  _lastBlk = NULL;
  _numBlks = 0;
  _writeTIDLists = WriteListFlag;

  _itemsetTree = NULL;
  _twoItemsetCounts = counts2;
  _bufSizes = new int[numItems];

#ifdef FANCYTID
  _numMaxTIDLists = (int)(0.5 * numItems * (numItems+1));
  //_relevantTIDLists = new (TIDList*)[_numMaxTIDLists];
  _tids = new TIDType[_numMaxTIDLists];
  _indxToTIDLists = new TwoDArray(MAX_NUMITS, MAX_IT_LENGTH);

  //memset((void*)_relevantTIDLists, 0, sizeof(TIDList*)*_numMaxTIDLists);
  memset((void*)_tids, 0, sizeof(TIDType)*_numMaxTIDLists);
  //for (i = 0; i < numMaxTIDLists; i++)
  // relevantTIDLists[i] = NULL;
#endif

}

BlkSequence::~BlkSequence()
{
  TIDListColln *temp1 = _firstBlk, *temp2 = NULL;

  delete [] _bufSizes;
  delete [] _twoItemsetCounts;
  while (temp1 != NULL)
    {
      cout << "Space occupied by FancyTIDs for BlkNum " << temp1->NumBlk() << " = " << temp1->Space() << endl;
      temp2 = temp1->NextBlk();
      delete temp1;
      temp1 = temp2;
    }
  delete _itemsetTree;

#ifdef FANCYTID
  delete [] _tids;
  delete _indxToTIDLists;
#endif
}

void BlkSequence::_FormFileNamePrefix(char *fname)
{
  cout << "BlkSequence::ToDo--hardcoding the prefix name" << endl;
  sprintf(fname, "/local.blondie.1/testVenky/");
}

void BlkSequence::AllocateSpaceForFiles(TIDListColln *tmp)
{
  int *supports = new int[_itemsetTree->NumItems()];
  int i;
  int arr[1];
  Set tempSet(0);
  tempSet.newArr(1, arr);
  
  for (i = 0; i < _itemsetTree->NumItems(); i++)
    {
      arr[0] = i;
      supports[i] = _itemsetTree->count(tempSet);
    }
  tempSet.Nullify();
  tmp->AllocateSpaceForFiles(supports);

  delete [] supports;
}

void BlkSequence::AllocateBufSizes(int maxBufSz)
{
  _itemsetTree->AllocateBufSizes(_bufSizes, maxBufSz);

#ifdef DEBUG
  int i;
  cout << "BlkSequence::AllocateBufSizes" << endl;
  for (i = 0; i < _itemsetTree->NumItems(); i++)
    cout << _bufSizes[i] << " " ;
  cout << endl << endl;
#endif
}

void BlkSequence::AssimilateNewBlk(char *datafile, bool BlkCounted, Timer *T=NULL)
{
  // read xacts one by one; increment counts of itemsets;
  // also writes the TID lists, if necessary;
  int tid = 0;

#ifdef ASCII
  char *buffer = NULL;
#else
  char buffer[TRANS_BLOCK_SIZE];
#endif
  char fileNamePrefix[MAX_NAME_LENGTH];
  TIDListColln *newBlk;
  int numL;

  //to be used only for the last block. That is, only if T != NULL;
  Timer timer1("Time for Updating counts of old itemsets");
  Timer timer2("Counting time using TID-lists");

  _FormFileNamePrefix(fileNamePrefix);
  numL = _itemsetTree->NumItems();
  //numL = _itemsetTree->NumMaterializedLists();

  newBlk = new TIDListColln(numL, _numBlks, fileNamePrefix, _bufSizes, 
			    _writeTIDLists, datafile);

#ifdef DEBUG
  cout << " Created the list of TID's for block " << _numBlks;
  cout << " for file " << datafile << endl;
#endif

  if (NULL != T)
    {
      T->Clear();
      T->Start();
      timer1.Start();
    }

  if ((FALSE == BlkCounted) || (TRUE == _writeTIDLists))
    {
      Set xactIter(0); // creating a dummy variable.
      SetupXactScan(datafile, buffer);
      while (getNextXact(xactIter, tid, buffer) != -1)
	{
	  if (TRUE == _writeTIDLists)
	    newBlk->WriteXact(&xactIter, tid);

	  IncrementCounts(xactIter, newBlk, BlkCounted, tid, _writeTIDLists);
	  if (FALSE == BlkCounted)
	    _itemsetTree->IncrNumXacts();
	}
      xactIter.Nullify();
    }
  if (NULL != T)
    timer1.Stop();

  if (TRUE == _writeTIDLists)
    newBlk->WriteInfoToDisk(_itemsetTree);
  
  AppendNewBlk(newBlk); // sample blocks should not be sent in.

#ifdef DEBUG
  cout << " Before pt->UpdateAfterNewBlk " << endl;
#endif

  if (FALSE == BlkCounted)
    {
      cout << " Updates while assimilating a new block " << endl;
      InsertNewLarge2Its();

      if (NULL != T) timer2.Start();
      _itemsetTree->UpdateAfterNewBlk(this);
      if (NULL != T) timer2.Stop();

    }
#ifdef DEBUG
  cout << " After pt->UpdateAfterNewBlk " << endl;
#endif

  if (NULL != T)
    {
      T->Stop();
      cout << "Block " << _numBlks << " : " << endl;
      cout << timer1 << endl;
      cout << timer2 << endl << endl;
    }

  newBlk = NULL; // we do not want to delete blocks;
}


void BlkSequence::IncrementCounts(Set& xact, TIDListColln* newBlk, bool CountedBlk,
				  int tid, bool writeTIDLists)
{
  int i, j, k;

  if (FALSE == CountedBlk)
    {
      for (i = 0; i < xact.size(); i++)
	{
	  for (j = i+1; j < xact.size(); j++)
	    {
	      k = xact.elements(i)*_itemsetTree->NumItems() + xact.elements(j);
	      if (NOT_COUNTED != _twoItemsetCounts[k])
		_twoItemsetCounts[k]++;
	    }
	}
    }

  if (TRUE == writeTIDLists)
    _itemsetTree->incrementCounts(xact, 1, allLevels, CountedBlk, newBlk, tid);
  else
    _itemsetTree->incrementCounts(xact, 1, allLevels, CountedBlk, NULL, -1);
}


void BlkSequence::InsertNewLarge2Its()
{
  int i, j;
  int nItems = _itemsetTree->NumItems();
  Item arr[2];
  Set tempSet(0);
  tempSet.newArr(2,arr);

  ofstream ofile("/tmp/inserted2Its");
  for (i = 0; i < nItems; i++)
    {
      for (j = i+1; j < nItems; j++)
	{
	  if (_twoItemsetCounts[i*nItems+j] >= _itemsetTree->returnMinSupport())
	    {
	      arr[0] = i;
	      arr[1] = j;
	      ofile << i << "," << j << "   " << _twoItemsetCounts[i*nItems+j] << endl;
	      _itemsetTree->insertWithCount(tempSet, _twoItemsetCounts[i*nItems+j], 
					    negBorder);
	      _twoItemsetCounts[i*nItems+j] = NOT_COUNTED; // not necessary to count this.
	      // negBorder because new candidates have to be generated in the
	      // prefix tree when it is updated immediately after this.
	    }
	}
    }
  tempSet.Nullify();
  ofile.close();
}


void BlkSequence::Clear2ItCounts()
{
  int i, j;
  int nItems = _itemsetTree->NumItems();
  
  for (i = 0; i < nItems; i++)
    for (j = i+1; j < nItems; j++)
      if (_twoItemsetCounts[i*nItems+j] != NOT_COUNTED)
	_twoItemsetCounts[i*nItems+j] = 0;

}


void BlkSequence::Print2ItCounts()
{
  int i, j;
  int nItems = _itemsetTree->NumItems();
  int minSup = _itemsetTree->returnMinSupport();

  ofstream ofile("/tmp/2Its");
  for (i = 0; i < nItems; i++)
    for (j = i+1; j < nItems; j++)
      if (_twoItemsetCounts[i*nItems+j] >= minSup)
	ofile << i << "," << j << "   " << _twoItemsetCounts[i*nItems + j] << endl;
  ofile.close();
}







void BlkSequence::AppendNewBlk(TIDListColln *newBlk)
{
  newBlk->SetPrevBlk(_lastBlk);
  newBlk->SetNextBlk(NULL);
  if (NULL == _firstBlk)
    {
      assert(_lastBlk == NULL);
      _firstBlk = newBlk;
      _lastBlk = newBlk;
    }
  else {
    assert(NULL != _lastBlk);
    _lastBlk->SetNextBlk(newBlk);
    _lastBlk = newBlk;
  }
  _numBlks++;
}

void BlkSequence::Start2Count(int i, int j, int cnt)
{
  int nItems = _itemsetTree->NumItems();
  assert(_twoItemsetCounts[i*nItems+j] == NOT_COUNTED);
  _twoItemsetCounts[i*nItems+j] = cnt;
}


int BlkSequence::TwoItCount(int i, int j)
{
  assert(i < j);
  return (_twoItemsetCounts[i*_itemsetTree->NumItems() + j]);
}

void BlkSequence::DeleteFirstBlk()
{
  TIDListColln *temp = _firstBlk;
  assert(temp != NULL);

  _firstBlk = temp->NextBlk();
  _firstBlk->SetPrevBlk(NULL);

  delete temp;
  _numBlks--;
}

void BlkSequence::Count(Set **itemsets, int *countArr, int numItemsets)
{
  int i;
  bool *countFlags = new bool[numItemsets];
  TIDListColln *temp = _lastBlk;
  
  Timer countTime("Time for counting new itemsets");

#ifdef USUAL_DYNAMIC
  countTime.Start();
  PrefixTree *tempPT = new PrefixTree(_itemsetTree->NumItems(), itemsets,
				      numItemsets);
  countTime.Stop();
#endif

  for (i = 0; i < numItemsets; i++)
    {
#ifndef USUAL_DYNAMIC
      if (itemsets[i]->size() == 2)
	{
	  countArr[i] = TwoItCount(itemsets[i]->elements(0), itemsets[i]->elements(1));
	  countFlags[i] = FALSE;
	}
      else
	{
	  countArr[i] = 0;
	  countFlags[i] = TRUE;
	}
#else
  countFlags[i] = TRUE;
  countArr[i] = 0;
#endif     
    }

  countTime.Start();
  while (temp != NULL)
    {
#ifdef USUAL_DYNAMIC
      cout << " Counting " << numItemsets << " itemsets using PT" << endl;
      if (numItemsets > 0)
	temp->CountUsingPT(tempPT);
#else
#ifdef FANCYTID
      cout << " Counting " <<numItemsets<<" itemsets using Fancy TID lists" << endl;
      temp->CountFancy(itemsets, countArr, countFlags, numItemsets, 
		       _itemsetTree->NumItems(), _numMaxTIDLists, 
		       _tids, _indxToTIDLists);
#else
      cout << " Counting " << numItemsets << " itemsets using TID lists" << endl;
      temp->Count(itemsets, countArr, countFlags, numItemsets);
#endif
#endif USUAL_DYNAMIC
      temp = temp->PrevBlk();
    }

#ifdef USUAL_DYNAMIC
  // copy the itemset counts into the count Array;
  for (i = 0; i < numItemsets; i++)
    countArr[i] = tempPT->count(*(itemsets[i]));
  delete tempPT;
#endif

  countTime.Stop();
  cout << "***********Counting Time*********" << endl;
  cout << countTime << endl;

#ifdef DEBUG
  for (i = 0; i < numItemsets; i++)
    cout << *itemsets[i] << " count= " << countArr[i] << endl;
#endif

  delete [] countFlags;
}

#ifdef INSTRUMENT
void BlkSequence::CountRandomIts(int numIts)
{
  int totMaxItSup = 0, numMaxIts = 0;
  int totUnMarkedSup = 0, numUnMaxed = 0;

  numMaxIts = _itemsetTree->MarkPossMaximalIts(totMaxItSup);
  numUnMaxed = _itemsetTree->UnMarkNonMaximalIts(totUnMarkedSup);

#ifdef FANCYTID
  _firstBlk->CountRandomItemsets(_itemsetTree,numIts,_itemsetTree->NumItems(),
				 _numMaxTIDLists, _tids, _indxToTIDLists);
#else
  _firstBlk->CountRandomItemsets(_itemsetTree, numIts);
#endif
}
#endif





