/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef __PARTNSEQUENCE_H__
#define __PARTNSEQUENCE_H__

#include "globals.h"

class Set;
class PrefixTree;
class TIDListColln;
class TIDList;
class TwoDArray;
class Timer;

class BlkSequence
{
 private:
  int _numBlks;
  bool _writeTIDLists;

  // a doubly linked list.
  TIDListColln *_firstBlk;
  TIDListColln *_lastBlk;

  int *_bufSizes;
  int *_twoItemsetCounts;
  PrefixTree *_itemsetTree;

#ifdef FANCYTID
  int _numMaxTIDLists;
  //TIDList **_relevantTIDLists;
  TIDType *_tids;
  TwoDArray *_indxToTIDLists;
#endif  

  void _FormFileNamePrefix(char *fname);

 public:
  BlkSequence(bool writeTIDListFlag, int *counts2, int nitems);
  ~BlkSequence();

  void AllocateSpaceForFiles(TIDListColln *tmp);
  void AllocateBufSizes(int maxBufSz);
  PrefixTree* ItemsetTree() { return _itemsetTree;}
  void SetPrefixTree(PrefixTree* pt) { _itemsetTree = pt; }
  void AssimilateNewBlk(char *datafile, bool FirstBlk=FALSE, Timer *T=NULL);

  void AppendNewBlk(TIDListColln* newBlk);
  void DeleteFirstBlk(); // window model.
      
  void Count(Set **itemsets, int *countArray, int numItemsets);

  // return the count of itemset (i,j)
  int TwoItCount(int i, int j);
  void Start2Count(int i, int j, int cnt);
	
  // void Consolidate(); // reduce the space for the prefix tree.

  /*
   *The following function maintains the twoItemsetCounts and also 
   *increments the larger itemsets in the prefix tree;
   *It should not be called for the first block, but only for the 
   *subsequent blocks.
   *
   *xact: transaction whose subset itemsets are to be incremented.
   *newBlk: information for the newly inserted block of data.
   *BlkCounted: whether the new block has already been counted.
   *tid: tid of the xact;
   *writeTIDLists: whether TIDlists should be written for the block.
   */
  void IncrementCounts(Set& xact, TIDListColln* newBlk, bool BlkCounted = FALSE, 
		       int tid = -1, bool writeTIDLists=FALSE);
  void InsertNewLarge2Its();
  void Clear2ItCounts();
  void Print2ItCounts();

#ifdef INSTRUMENT
  /*
   *Randomly pick numIts number of itemsets from the negative border and
   *count them.
   */
  void CountRandomIts(int numIts);
#endif
};

#endif

