#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include "PrefixTree.h"
#include "blockSequence.h"
#include "srcFunctions.h"
#include "trans_utils.h"
#include "timeutil.h"

extern int count1and2Itemsets(char *datafile, int * counts1, int * counts2,
			      int numItems);



class OneDArray
{
private:
  int _sz;
  int _maxSz;
  int *_arr;

public:
  OneDArray(int size, int maxSize)
    {
      _sz = size;
      _maxSz = maxSize;
      _arr = new int[_maxSz];
    }

  ~OneDArray()
    {
      delete [] _arr;
    }

  int Size() { return _sz; }

  bool Member(int attrVal)
    {
      int i;
      for (i = 0; i < _sz; i++)
	if (attrVal == _arr[i])
	  return TRUE;
      return FALSE;
    }

  int Get(int indx)
    {
      assert(indx < _sz);
      return _arr[indx];
    }

  void Set(int indx, int val)
    {
      assert(indx < _sz);
      _arr[indx] = val;
    }

  void Append(int val)
    {
      //cout << "Append " << _sz << " " << _maxSz << endl;
      assert(_sz < _maxSz);
      /*****
      if (FALSE == Member(val))
	{
	  _arr[_sz] = val;
	  _sz++;
	}
      ***/

      int insertPos=0, indx=0;
      if (0 == _sz)
	{
	  _arr[_sz] = val;
	  _sz++;
	}
      else
	{
	  for (insertPos = 0; insertPos < _sz && _arr[insertPos] < val; insertPos++);
	  if (_arr[insertPos] != val)
	    {
	      for (indx = _sz; indx > insertPos; indx--)
		_arr[indx] = _arr[indx-1];
	      _arr[insertPos] = val;
	      _sz++;
	    }
	}
    }

  void Delete(int attrVal)
    {
      int i;
      int indx = -1;

      for (i = 0; i < _sz; i++)
	{
	  if (_arr[i] == attrVal)
	    {
	      indx = i;
	      break;
	    }
	}
      if (-1 == indx)
	assert(0);
      assert(indx < _sz);

      for (i = indx; i < _sz-1; i++)
	_arr[i] = _arr[i+1];
      _sz--;

      assert(_sz > 0);
    }

  void Sort()
    {
    }

};



void InitSizes(int *sizes, int numSizes, int maxSize)
{
  assert(numSizes <= 10);

  /****
  switch(numSizes)
    {
    case 3:
      sizes[2] = (int)(0.40 * maxSize);
    case 2:
      sizes[1] = (int)(0.30 * maxSize);
    case 1:
      sizes[0] = (int)(0.20 * maxSize);
      break;
    default:
      assert(0);
    }
  ******/

  switch(numSizes)
    {
    case 10:
      sizes[9] = (int)(0.80 * maxSize);
    case 9:
      sizes[8] = (int)(0.70 * maxSize);
    case 8:
      sizes[7] = (int)(0.60 * maxSize);
    case 7:
      sizes[6] = (int)(0.50 * maxSize);
    case 6:
      sizes[5] = (int)(0.40 * maxSize);
    case 5:
      sizes[4] = (int)(0.30 * maxSize);
    case 4:
      sizes[3] = (int)(0.20 * maxSize);
    case 3:
      sizes[2] = (int)(0.10 * maxSize);
    case 2:
      sizes[1] = (int)(0.05 * maxSize);
    case 1:
      //sizes[0] = (int)(0.01 * maxSize);
			// A HACK
			sizes[0] = (int)(0.20 * maxSize);
      break;
    default:
      assert(0);
    }
}

void ScanDataFileToCount(char *datafile, PrefixTree *pt)
{
  char buffer[TRANS_BLOCK_SIZE];
  Set xactIter(0);
  int tid = 0;
  
  int arr[MAX_IT_LENGTH];
  Set **supportedItSets = new (Set*)[MAX_SUPCNT_PERXACT];

  SetupXactScan(datafile, buffer);
  while (getNextXact(xactIter, tid, buffer) != -1)
    pt->IncrCountsForDevn(xactIter, supportedItSets, arr);

  delete [] supportedItSets;
  xactIter.Nullify();
}


// PT1 and PT2;
// a maximal itemset in PT1: 
// can be maximal in PT2: obtain the difference. -- symmetric.
// the next two are anti-symmetric, i.e., if (a) happens from PT1, then 
// (b) happens if we look at the same itemset from PT2;
// (a) can be large but not maximal; - ignore;
// (b) can be small; - count later;


// A maximal itemset in PT1: Is it a maximal itemset in PT2? If
// so, obtain the difference in their counts; else put it in itSetsOf1in2;
// mark some of these as present in pt2;
double ComparePT1withPT2(PrefixTree *pt1, PrefixTree *pt2, Set **itSetsOf1in2,
			 int& num1in2)
{
  double diff = 0.0;
  pt1->DiffwithPT2(pt2, itSetsOf1in2, num1in2, diff);
  return diff;
}


void CountItemsets(Set **itemsets, int *countArr, int numItemsets, int nitems, 
		   char *datafile)
{
  PrefixTree *pt = new PrefixTree(nitems, itemsets, numItemsets);
  pt->CountItemsets(datafile);
  
  int i;

#ifdef DEBUG
  //cout << " Devns -- Counted " << numItemsets << " itemsets" << endl;
#endif

  for (i = 0; i < numItemsets; i++)
    countArr[i] = pt->count(*(itemsets[i]));
  delete pt;
}

//Both pt1 and pt2 are changed to represent
//the least common model.
double FindDeviationWithAdjustedCounts(PrefixTree *pt1, PrefixTree *pt2, 
				       char *datafile1, char *datafile2)
{
  double diff = 0.0;

  Set **itSetsOf1in2 = new (Set*)[MAX_SETCNT];
  Set **itSetsOf2in1 = new (Set*)[MAX_SETCNT];
  //int *countsOf1in2;
  //int *countsOf2in1;

  int num1in2=0, num2in1=0;
  int ntrans1, ntrans2;
  int i;
 
  ntrans1 = pt1->getNumXacts();
  ntrans2 = pt2->getNumXacts();

  pt1->DiffwithPT2(pt2, itSetsOf1in2, num1in2, diff);
  assert(num1in2 < MAX_SETCNT);
  //countsOf1in2 = new int[num1in2];
  //memset((void*)countsOf1in2, 0, sizeof(int)*num1in2);
  //assert(countsOf1in2 != NULL);

  pt2->DiffwithPT2(pt1, itSetsOf2in1, num2in1, diff); 
  //pt2->GetNonPresentMaxIts(itSetsOf2in1, num2in1);
  assert(num2in1 < MAX_SETCNT);
  //countsOf2in1 = new int[num2in1];
  //memset((void*)countsOf2in1, 0, sizeof(int)*num2in1);
  //assert(countsOf2in1 != NULL);


  pt1->UnPRESENTIts();
  pt2->UnPRESENTIts();

  if (num1in2 > 0)
    {
      //CountItemsets(itSetsOf1in2, countsOf1in2, num1in2, nitems, datafile2);
      for (i = 0; i < num1in2; i++)
	pt2->insertWithCount(*(itSetsOf1in2[i]), 0, maximal);
    }

  if (num2in1 > 0)
    {
      //CountItemsets(itSetsOf2in1, countsOf2in1, num2in1, nitems, datafile1);
      for (i = 0; i < num2in1; i++)
	pt1->insertWithCount(*(itSetsOf2in1[i]), 0, maximal);
    }

  for (i = 0; i < num1in2; i++)
    delete itSetsOf1in2[i];
  for (i = 0; i < num2in1; i++)
    delete itSetsOf2in1[i];


  // now, both pt1 and pt2 conform to the least common model;
  pt1->clearCounts();
  pt2->clearCounts();

  ScanDataFileToCount(datafile1, pt1);
  ScanDataFileToCount(datafile2, pt2);
  
  int numDiff = 0;
  double diff1=0.0, diff2 = 0.0;

  pt1->DiffwithPT2(pt2, itSetsOf1in2, numDiff, diff1);
  assert(numDiff == 0);

  //diff2 = pt2->DiffwithPT2(pt1, itSetsOf1in2, numDiff);
  //assert(numDiff == 0);
  //assert(diff1 == diff2);
  

  pt1->UnPRESENTIts();
  pt2->UnPRESENTIts();

  delete [] itSetsOf1in2;
  delete [] itSetsOf2in1;
  //delete [] countsOf1in2;
  //delete [] countsOf2in1;

  return diff1;

  // repeatedly call AdjustItCounts while numAdjusted > 0;
}


// deviation between two blocks: blkSeq1 and blkSeq2 through PT1 and PT2;
// 
double FindDeviation(PrefixTree *pt1, PrefixTree *pt2, char *datafile1, 
		     char *datafile2)
{
  double diff = 0.0;
  Set **itSetsOf1in2 = new (Set*)[MAX_SETCNT];
  int *countsOf1in2;
  Set **itSetsOf2in1 = new (Set*)[MAX_SETCNT];
  int *countsOf2in1;
  int num1in2=0, num2in1=0;

  int nitems = pt1->NumItems();
  int ntrans1, ntrans2;
  int i;

  assert(nitems == pt2->NumItems());
  ntrans1 = pt1->getNumXacts();
  ntrans2 = pt2->getNumXacts();

  pt1->DiffwithPT2(pt2, itSetsOf1in2, num1in2, diff);
  assert(num1in2 < MAX_SETCNT);

  countsOf1in2 = new int[num1in2];
  memset((void*)countsOf1in2, 0, sizeof(int)*num1in2);
  assert(countsOf1in2 != NULL);

#ifndef LITS_UPBND
  if (num1in2 > 0)
    CountItemsets(itSetsOf1in2, countsOf1in2, num1in2, nitems, datafile2);
#endif

  pt2->DiffwithPT2(pt1, itSetsOf2in1, num2in1, diff); 
  //pt2->GetNonPresentMaxIts(itSetsOf2in1, num2in1);
  assert(num2in1 < MAX_SETCNT);
  countsOf2in1 = new int[num2in1];
  memset((void*)countsOf2in1, 0, sizeof(int)*num2in1);
  assert(countsOf2in1 != NULL);

#ifndef LITS_UPBND
  if (num2in1 > 0)
    CountItemsets(itSetsOf2in1, countsOf2in1, num2in1, nitems, datafile1);
#endif

  double val1 = 0.0, val2=0.0;

  for (i = 0; i < num1in2; i++)
    {
      val1 = ((double)(pt1->count(*(itSetsOf1in2[i]))))/(double)ntrans1;
      val2 = ((double)countsOf1in2[i])/(double)ntrans2;
      AggregateFn(DeviationFn(val1, val2), diff);
    }

  for (i = 0; i < num2in1; i++)
    {
      val1 = ((double)(pt2->count(*(itSetsOf2in1[i]))))/(double)ntrans2;
      val2 = ((double)countsOf2in1[i])/(double)ntrans1;
#ifdef DEBUG
      //cout << *itSetsOf2in1[i] << " :: " << val1 << " " << val2 << endl;
#endif
      AggregateFn(DeviationFn(val1, val2), diff);
    }

  pt1->UnPRESENTIts();
  pt2->UnPRESENTIts();

  for (i = 0; i < num1in2; i++)
    delete itSetsOf1in2[i];
  for (i = 0; i < num2in1; i++)
    delete itSetsOf2in1[i];

  delete [] itSetsOf1in2;
  delete [] itSetsOf2in1;
  delete [] countsOf1in2;
  delete [] countsOf2in1;

  return diff;
}


//GetUnMarkedMaxIts;
//DiffwithPT
//UnMarkMaxIts;

//main 
// construct a prefix tree over the entire dataset;
// take samples and then construct trees over each of them, 
// and then measure deviations with each of these; 
// aside: write a sample.sh to generate a number of samples
// from the same datafile;


void MarkMaximalIts(PrefixTree *pt)
{
  int numMaxIts=0, numUnMaxed=0;
  int totMaxItSup=0, totUnMarkedSup=0;

  numMaxIts = pt->MarkPossMaximalIts(totMaxItSup);
  numUnMaxed = pt->UnMarkNonMaximalIts(totUnMarkedSup);
  pt->SetNumMaxIts(numMaxIts - numUnMaxed);
}

PrefixTree* ConstructPT(char *datafile, int numItems, double relSupport)
{
  PrefixTree *pt;
  int *counts1 = new int[numItems];
  int *counts2 = new int[numItems * numItems];
  int numXacts;

  memset((void*)counts1, 0, sizeof(int)*numItems);
  memset((void*)counts2, 0, sizeof(int)*numItems*numItems);

  numXacts = count1and2Itemsets(datafile, counts1, counts2, numItems);
	
  pt = new PrefixTree(numItems, counts1, counts2, relSupport, numXacts);
  
  delete [] counts1;
  delete [] counts2;

  PrefixTree **T = new (PrefixTree*)[5];
  T[0] = NULL;
  T[1] = NULL;
  T[2] = pt;

  apriori(T, datafile, numItems);
  MarkMaximalIts(pt);
	
  T[2] = NULL;
  delete [] T;

  return pt;
}

// numSizes: number of different sample sizes;
// numSamples: number of samples per size;
// The older routine which uses FindDeviationsWithAdjustedCounts can be found
// in /u/v/g/vganti/db/dynamic/assoc/working.12Oct98/dynamic/deviations.cpp
void DeviationsFromSamples(char *prefixName, int numSizes, int numSamplesPerSz, 
													 int numItems, double relSupport, int numXacts, int blkNum)
{
  PrefixTree *pt1, *pt2;
  char sampleFileName1[MAX_NAME_LENGTH];
  char sampleFileName2[MAX_NAME_LENGTH];
  char datafile[MAX_NAME_LENGTH];
  char outputfile[MAX_NAME_LENGTH];
  char command[MAX_NAME_LENGTH];

  int i,j;
  double deviation = 0.0, sumDevn = 0.0;
  int *sizes = new int[numSizes];

  //cout << "Inside Deviations from samples " << endl;
  sprintf(datafile, "%s/d.blk%d", prefixName, blkNum);
  

  //pt1 = ConstructPT(datafile, numItems, relSupport);
  //InitSizes(sizes, numSizes, numXacts);

	// A HACK
	assert(1 == numSizes);
	sizes[0] = (int)(0.15*numXacts);

#ifndef SOL  
#ifdef LITS_UPBND
  sprintf(outputfile, "%s/devns/upbnd/devnsFromSamples.upbnd.%d.%f.%d.%d", 
	  prefixName, numXacts, relSupport, 
	  DevnFnType(), AggrFnType());
#else
  sprintf(outputfile, "%s/devns/devnsFromSamples.%d.%f.%d.%d", 
	  prefixName, numXacts, relSupport, 
	  DevnFnType(), AggrFnType());
#endif

#else
#ifdef LITS_UPBND
  sprintf(outputfile, "/u/v/g/vganti/db/dynamic/assoc/code/devns/upbnd/devnsFromSamples.upbnd.%dK.%f.%d.%d", 
	  numXacts, relSupport, 
	  DevnFnType(), AggrFnType());
#else
  sprintf(outputfile, "/tmp/devns/devnsFromSamples.%dK.%f.%d.%d", 
	  numXacts, relSupport, 
	  DevnFnType(), AggrFnType());

#endif
#endif

  //delete pt1; // To provide independence.

  ofstream ofile;
  ofile.open(outputfile);


  // call apriori on each of the samples; 
  SetUpSampling();

  cout << numXacts << endl;
  for (j = 0; j < numSizes; j++)
    {
      deviation = 0.0;
      sumDevn = 0.0;
      ofile << ((double)sizes[j]/numXacts) << "\t" << sizes[j] << "\t";

      for (i = 0; i < numSamplesPerSz; i++)
	{
#ifndef SOL
	  sprintf(sampleFileName1, "/local.blondie.1/testVenky/sampledFile1");
	  sprintf(sampleFileName2, "/local.blondie.1/testVenky/sampledFile2");
#else
	  sprintf(sampleFileName1, "/tmp/sampledFile1");
	  sprintf(sampleFileName2, "/tmp/sampledFile2");
#endif
	  //sprintf(sampleFileName, "%s/samples.%dK/d.sample%d", prefixName, sizes[j], i);
	  GetNextSample(datafile, sizes[j], sampleFileName1, numXacts);
	  pt1 = ConstructPT(sampleFileName1, numItems, relSupport);

	  GetNextSample(datafile, sizes[j], sampleFileName2, numXacts);
	  pt2 = ConstructPT(sampleFileName2, numItems, relSupport);



	  deviation = FindDeviation(pt1, pt2, sampleFileName1, sampleFileName2);
	  sumDevn += deviation;
	  
	  ofile << deviation << " ";
	  cout << "Deviation for sample " << i << " of size " << sizes[j] << " = " << deviation << endl;

	  delete pt1;
	  delete pt2;
	  sprintf(command, "/bin/rm %s", sampleFileName1);
	  system(command);

	  sprintf(command, "/bin/rm %s", sampleFileName2);
	  system(command);
	}

      ofile << sumDevn/numSamplesPerSz << endl;
      cout << "Average deviation for size " << sizes[j] << " = " << sumDevn/numSamplesPerSz << endl;
      cout << endl << endl; 
    }
  ofile.close();
  //delete pt1;
  delete [] sizes;
}

	

void DeviationBetweenBlks(char *prefixName, int numItems, double relSupport,
													int numBlks)
{
  PrefixTree **pt = new (PrefixTree*)[numBlks];
	char **datafiles = new (char*)[numBlks];
  char outputFileName[MAX_NAME_LENGTH];
  double deviation=0.0;
	int i, j;

	Timer T("Time for computing Deviations");
	for (i = 0; i < numBlks; i++)
		{
			T.Clear();
			T.Start();
			datafiles[i] = new char[MAX_NAME_LENGTH];
			sprintf(datafiles[i], "%s/d.blk%d", prefixName, i);
			cout << datafiles[i] << endl;

			pt[i] = ConstructPT(datafiles[i], numItems, relSupport);
			T.Start();
			cout << "Tree building time for block " << i << " ::: " << T << endl;
		}
  sprintf(outputFileName, "/tmp/deviations.actual");

  ofstream of;
  of.open(outputFileName, ios::app);
	assert(of != NULL);

	for (i = 0; i < numBlks; i++)
		for (j = i+1; j < numBlks; j++)
			{
				of << "Deviation between Blocks " << i << " " << j << " ";
				
				T.Clear();

				T.Start();
				deviation = FindDeviation(pt[i], pt[j], datafiles[i], datafiles[j]);
				T.Stop();
  
				of << deviation << endl;
				of << T << endl << endl;
			}
	for (i = 0; i < numBlks; i++)
		{
			delete pt[i];
			delete [] datafiles[i];
		}
	delete [] pt;
	delete [] datafiles;
  of.close();
}

const double maxDevnInGroup=0.11;
//const double maxDevnInGroup=0.13;

bool Appendable(OneDArray* group, int numGroups, int blkNum, char *prefixName, int numItems, 
								double relSupport, double *devns, int numBlks)
{
	int i;
	double deviation;
	bool result = TRUE;
	/**
	char **datafiles = new (char*)[group->Size()];
	char datafileNew[MAX_NAME_LENGTH];
	PrefixTree **pt = new (PrefixTree*)[group->Size()];
	PrefixTree *ptNew;

	sprintf(datafileNew, "%s/d.blk%d", prefixName, blkNum);
	ptNew = ConstructPT(datafileNew, numItems, relSupport);
	**/
	for (i = 0; i < group->Size(); i++)
		{
			assert(group->Get(i) < blkNum);
			/***
			assert(group->Get(i) != blkNum);
			datafiles[i] = new char[MAX_NAME_LENGTH];
			sprintf(datafiles[i], "%s/d.blk%d", prefixName, group->Get(i));
			pt[i] = ConstructPT(datafiles[i], numItems, relSupport);
			
			deviation = FindDeviation(ptNew, pt[i], datafileNew, datafiles[i]);
			**/
			deviation = devns[group->Get(i)*numBlks + blkNum];
			cout << "Deviation between " << group->Get(i) << " and " << blkNum << " is " << deviation << endl;

			/*  CleanUp 
			delete pt[i];
			delete datafiles[i];
			*/

			if (deviation > maxDevnInGroup)
				{
					result = FALSE;
					break;
				}
		}
	/*
	delete ptNew;
	delete [] datafiles;
	*/
	return result;
}

void ComputePairwiseDevns(char *prefixName, int numItems, double relSupport, int numBlks,
													double *devns)
{
	int i, j;
	char **datafiles = new (char*)[numBlks];
	double *times = new double[numBlks];
	char datafileNew[MAX_NAME_LENGTH];
	PrefixTree **pt = new (PrefixTree*)[numBlks];
	Timer timer("Assimilation time for a new block");

	for (i = 0; i < numBlks; i++)
		{ 
			times[i]=0;
			timer.Clear();
			timer.Start();
			datafiles[i] = new char[MAX_NAME_LENGTH];
			sprintf(datafiles[i], "%s/d.blk%d", prefixName, i);
			pt[i] = ConstructPT(datafiles[i], numItems, relSupport);
			timer.Stop();
			times[i] = timer.WallClockTime();
			cout << " Time for constructing PT for block " << i << " ::: " << timer << endl;
		}

	ofstream ofile1("/tmp/trend-times.1");
	for (i = numBlks-1; i >= 0; i--)
		ofile1 << i << "  " << times[i] << endl;
	ofile1.close();

	for (i = numBlks-1; i >= 0; i--)
		{
			timer.Clear();
			timer.Start();
			for (j = 0; j < i; j++)
			{
				devns[j*numBlks+i] = FindDeviation(pt[i], pt[j], datafiles[i], datafiles[j]);
			}
			timer.Stop();
			times[i] += timer.WallClockTime();
			cout << "Finding deviation " << i << " ::: " << timer << endl;
		}
	/*
	for (i = 0; i < numBlks; i++)
	 {
	   timer.Clear();
		 timer.Start();
		for (j = i+1; j < numBlks; j++)
			{
				devns[i*numBlks+j] = FindDeviation(pt[i], pt[j], datafiles[i], datafiles[j]);
			}
	   timer.Stop();
		 times[i] += timer.WallClockTime();
	   //cout << "Finding deviation " << i << " ::: " << timer << endl;
	}
			*/


	ofstream ofile("/tmp/trend-times");
	for (i = 0; i < numBlks; i++)
		{
			//cout << "Time for assimilating block " << i << " is " << times[i] << endl;
			//ofile << i << "\t" << times[numBlks-1-i] << endl;
			ofile << i << "\t" << times[i] << endl;
			/*  CleanUp */
			delete pt[i];
			delete datafiles[i];
		}
	ofile.close();
	delete [] pt;
	delete [] datafiles;
	delete [] times;
}



void DeviationTrends(char *prefixName, int numItems, double relSupport, int numBlks)
{
  OneDArray **groups = new (OneDArray*)[numBlks];
	double *devns = new double[numBlks*numBlks];
	int i, j;
	int numGroups = 0;
	bool result=FALSE;

	cout << "Inside DeviationTrends " << numBlks << endl;
	groups[0] = new OneDArray(0, numBlks);
	(groups[0])->Append(0);
	numGroups++;
	
	ComputePairwiseDevns(prefixName, numItems, relSupport, numBlks, devns);

	for (i = 1; i < numBlks; i++)
		{
			result = FALSE;
			for (j = 0; j < numGroups; j++)
				{
					result = Appendable(groups[j], numGroups, i, prefixName, numItems, relSupport,
															devns, numBlks);
					if (TRUE == result)
						{
							groups[j]->Append(i);
							//break;
						}
				}
			//if (FALSE == result)
				{
					groups[numGroups] = new OneDArray(0, numBlks);
					groups[numGroups]->Append(i);
					numGroups++;
				}
		}
	
	cout << "Number of Groups " << numGroups << "; Number of Blks " << numBlks << endl;
	for (i = 0; i < numGroups; i++)
		{
			cout << "Group " << i << " :: ";
			for (j = 0; j < groups[i]->Size(); j++)
				cout << groups[i]->Get(j)+1 << " "; // 6 hrs
			cout << endl;
		}

	/*** CleanUp ***/
	for (i = 0; i < numGroups; i++)
		delete groups[i];
	delete [] devns;
	delete [] groups;
}



