#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "globals.h"

#include "Set.h"
#include "macros.h"
#include "srcFunctions.h"
#include "PrefixTree.h"
#include "trans_utils.h"

#include "tid-list.h"
#include "tidlist-colln.h"
#include "timeutil.h"

void InputParameters(int argc, char **argv, char *prefixName, int& h, int& a, 
		     int& b, int& numItems, double& relSupport, bool& writeFlag,
		     int& numBlks, int& sampleSize, bool& lazy,
		     ALGORITHM& algo, int& numSampleSizes, int& numSampsPerSz,
		     DISTANCE_TYPE& devType, AGGFN_TYPE& aggfnType);

int count1and2Itemsets(char *datafile, int * counts1, int * counts2,
		       int numItems);

BlkSequence* ProcessFirstBlk(PrefixTree*& pt, char *datafile, int numItems, 
			     double relSupport, bool writeTIDListFlag, int h, 
			     int a, int b, bool SampleBlk, bool lazy);

void ProcessNewBlk(char *datafile, BlkSequence *blkSeq, Timer *T=NULL);
void CountItemsets(BlkSequence *blkSeq);

extern void DeviationsFromSamples(char *pfName, int numSzs, int numSamps, 
																	int nItems, double relSup, int numXacts, int blkNum=0);

extern void DeviationBetweenBlks(char *prefixName, int numItems, 
																 double relSupport, int numBlks);
extern void DeviationTrends(char *prefixName, int numItems, double relSupport, int numBlks);

// Processing the first block is slightly different because the 
// prefix tree is constructed for the first time. This involves
// counting 1-and 2-itemsets separately.
BlkSequence* ProcessFirstBlk(PrefixTree*& pt, char *datafile, int numItems, 
			     double relSupport, bool writeTIDListFlag, int h, 
			     int a, int b, bool SampleBlk, bool lazy)
{
  
  int numXacts;
  
  int *counts1 = new int[numItems];             // Counts for the 1-itemsets.
  int *counts2 = new int[numItems * numItems];  // Counts for the 2-itemsets.

  /**  // Clear the counts for the 1- and 2-itemsets.
  for (i = 0; i < numItems; i++) {
    counts1[i] = 0;
    for (int j = 0; j < numItems; j++)
      counts2[i * numItems + j] = 0;
  }
  ***/
  memset((void*)counts1, 0, sizeof(int)*numItems);
  memset((void*)counts2, 0, sizeof(int)*numItems*numItems);

  numXacts = count1and2Itemsets(datafile, counts1, counts2, numItems);

  pt = new PrefixTree(numItems, counts1, counts2, relSupport, numXacts, h, a, b, lazy, NULL);

#ifdef DEBUG
  cout << " NumXacts " << pt->getNumXacts();
#endif

  delete [] counts1;
  
  PrefixTree** T = new (PrefixTree*)[numItems];
  T[0] = NULL;
  T[1] = NULL;
  T[2] = pt;

  apriori(T, datafile, numItems);
  
  BlkSequence *blkSeq = new BlkSequence(writeTIDListFlag, counts2, numItems); 
  blkSeq->SetPrefixTree(pt);
  counts2 = NULL;

  if (FALSE == SampleBlk)
    {
      blkSeq->AllocateBufSizes(MAX_BUFSIZE);
      blkSeq->AssimilateNewBlk(datafile, TRUE);
    }

  T[2] = pt = NULL;
  delete [] T;
  return blkSeq;
}

void ProcessNewBlk(char *datafile, BlkSequence *blkSeq, Timer *T=NULL)
{
  blkSeq->AssimilateNewBlk(datafile, FALSE, T);
}

void testBlkSequence(char *prefixName, int h, int a, int b, int numItems,
		     double relSupport, bool writeFlag, int numBlks, bool SampleBlk, bool lazy)
{
  char datafile[MAX_NAME_LENGTH];
  BlkSequence *blkSeq=NULL;
  PrefixTree *pt=NULL;
  int i;

  Timer T1("Time for assimilating the last block");

#ifdef FANCYTID
  int totMaxItSup = 0, numMaxIts = 0;
  int totUnMarkedSup = 0, numUnMaxed = 0;
#endif

  if (TRUE == SampleBlk)
    {
      sprintf(datafile, "%s.sample", prefixName);
      cout << MAX_NAME_LENGTH << " " << strlen(datafile);
      blkSeq = ProcessFirstBlk(pt, datafile, numItems, relSupport, 
			       writeFlag, h, a, b, SampleBlk, lazy);

#ifdef FANCYTID      
      cout << " Just before marking maximal itemsets " << endl;
      numMaxIts = (blkSeq->ItemsetTree())->MarkPossMaximalIts(totMaxItSup);
      numUnMaxed = (blkSeq->ItemsetTree())->UnMarkNonMaximalIts(totUnMarkedSup);
      (blkSeq->ItemsetTree())->SetNumMaxIts(numMaxIts - numUnMaxed);
      cout << " Total number of maximal itemsets " << numMaxIts - numUnMaxed << endl;
      cout << " Total support for maximal itemsets + " << totMaxItSup - totUnMarkedSup << endl;
      cout << " Total sup for one itemsets " << (blkSeq->ItemsetTree())->TotSupForOneIts() << endl;
#endif
      
      cout << " Minimum Support value " << (blkSeq->ItemsetTree())->returnMinSupport() << " " << (blkSeq->ItemsetTree())->getNumXacts() << endl;

      blkSeq->AllocateBufSizes(MAX_BUFSIZE);

      (blkSeq->ItemsetTree())->clearCounts();
      blkSeq->Clear2ItCounts();
      (blkSeq->ItemsetTree())->SetNumXact(0);
    }

  for (i = 0; i < numBlks; i++)
    {
#ifdef DEBUG
      cout << " Processing block : " << i << endl;
      //cout << " Minimum Support value " << (blkSeq->ItemsetTree())->returnMinSupport() << " " << (blkSeq->ItemsetTree())->getNumXacts() << endl;
#endif

      if ((0 == i) && (FALSE == SampleBlk))
	{
	  sprintf(datafile, "%s.blk%d", prefixName, i);
	  cout << MAX_NAME_LENGTH << " " << strlen(datafile) << endl;
	  blkSeq = ProcessFirstBlk(pt, datafile, numItems, relSupport, 
				   writeFlag, h, a, b, SampleBlk, lazy);
#ifdef FANCYTID
	  cout << " Just before marking maximal itemsets " << endl;
	  numMaxIts = (blkSeq->ItemsetTree())->MarkPossMaximalIts(totMaxItSup);
	  numUnMaxed = (blkSeq->ItemsetTree())->UnMarkNonMaximalIts(totUnMarkedSup);
	  (blkSeq->ItemsetTree())->SetNumMaxIts(numMaxIts - numUnMaxed);
#endif	  
#ifdef DEBUG
	  cout << *(blkSeq->ItemsetTree());
#endif	  
	}
      else
	{
	  sprintf(datafile, "%s.blk%d", prefixName, i);
	  if (i == numBlks-1)
	    ProcessNewBlk(datafile, blkSeq, &T1);
	  else
	    ProcessNewBlk(datafile, blkSeq);
	}
      //if (i == numBlks-1)
      //T1.Stop();
    }

  cout << T1 << endl;

  pt = blkSeq->ItemsetTree();
  pt->setMinSupport(pt->returnMinSupport());
  pt = NULL;
#ifdef DEBUG
  cout << *(blkSeq->ItemsetTree());
#endif

#ifdef INSTRUMENT
  blkSeq->CountRandomIts(500);
#endif

#ifdef DEBUG
  //blkSeq->Print2ItCounts();
  //CountItemsets(blkSeq);
#endif

  delete blkSeq;
}

int main(int argc, char** argv)
{

  char *prefixName = new char[MAX_NAME_LENGTH];
  int h, a, b, numItems, numBlks;
  double relSupport;
  bool writeFlag;
  int sampleSz = -1;
  bool SampleBlk = FALSE;
  bool lazy = FALSE;
  ALGORITHM algo = MAINTENANCE;
  int numSamplesPerSz = 1;
  int numSampleSizes = 1;
  DISTANCE_TYPE devType = ABSOLUTE;
  AGGFN_TYPE aggType = L1;
  
  umask(0);
	cout << "calling inputparameters routine " << endl;
  InputParameters(argc, argv, prefixName, h, a, b, numItems, relSupport, 
		  writeFlag, numBlks, sampleSz, lazy, 
		  algo, numSampleSizes, numSamplesPerSz, devType, aggType);

  SetUpDevFns(devType, aggType);
  cout << " Function types : " << DevnFnType() << " " << AggrFnType() << endl;

  cout << sampleSz << endl;
  if (sampleSz > 0)
    SampleBlk = TRUE;
  else
    SampleBlk = FALSE;


  if (MAINTENANCE == algo)
    testBlkSequence(prefixName, h, a, b, numItems, relSupport, writeFlag, 
										numBlks, SampleBlk, lazy);
  else if (DEVIATIONS == algo)
    {
      int numXacts = 100000; // a hack just for the experiments.
      cout << " Computing deviations " << numXacts << " " << numSamplesPerSz << " " <<  numSampleSizes << endl;
			
			//YET another hack; overloading numBlks to bootstrap a particular block.
      DeviationsFromSamples(prefixName, numSampleSizes, numSamplesPerSz, numItems, 
														relSupport, numXacts, numBlks);
    }
  else if (DEVNS_ACTUAL == algo)
    {
      DeviationBetweenBlks(prefixName, numItems, relSupport, numBlks);
    }
	else if (TRENDS == algo)
		{
			cout << " Calling DeviationTrends " << endl;
			DeviationTrends(prefixName, numItems, relSupport, numBlks);
		}
  else assert(0);

  delete [] prefixName;
}





int count1and2Itemsets(char *datafile,
		       int * counts1, int * counts2,
		       int numItems) 
{
#ifdef ASCII
    char *buffer = NULL;
#else
    char buffer[TRANS_BLOCK_SIZE];
#endif

    int numXacts = 0;
    int tid = 0;
    int i,j,k;

    Set transaction(0);
    SetupXactScan(datafile, buffer);

    while (getNextXact(transaction, tid, buffer) != -1)
      {
	numXacts++;
	//	cout << " Transaction size " << transaction << " : " << transaction.size() << endl;
	for ( i = 0; i < transaction.size(); i++) 
	  {
	    // Increment the 1-itemset counts.
#ifdef DEBUG	  
	    assert(transaction.elements(i) < numItems);
#endif
	    counts1[transaction.elements(i)]++;
	    
	    // Increment the two itemset counts.
	    for ( j = i + 1; j < transaction.size(); j++) 
	      {
		// LA: Verify the computation below.
		k = transaction.elements(i) * numItems + transaction.elements(j);
		if (!(transaction.elements(j) < numItems))
		  cout << numXacts << " " << i << " " << j << " , " << numItems << " " << transaction << endl;
		assert(transaction.elements(j) < numItems);
		counts2[k]++;
		//if (counts2[k] == minSupport) large2Itemsets++;
	      }
	  }
      }
#ifndef ASCII
    transaction.Nullify();
    //delete [] buffer;
#endif

#if !defined(SAMPLING) && defined(ASCII)
    if (buf[0] != '\0')
      cerr << "no end of line before end of file in xfile" << endl
	   << "transaction " << buf << " ignored!" << endl;
#endif  // !SAMPLING && ASCII
    
    return numXacts;
}


#ifdef DEBUG
void CountItemsets(BlkSequence *blkSeq)
{
  int numItemsets = 4;

  Item items1[] = {202, 519};
  Item items2[] = {467, 860};
  Item items3[] = {137, 860};
  Item items4[] = {202};

  int counts[numItemsets];
  bool countFlags[numItemsets];
  Set **itemsets = new (Set*)[numItemsets];

  Set itemset1(2, items1);
  itemsets[0] = &itemset1;
  countFlags[0] = TRUE;

  Set itemset2(2, items2);
  itemsets[1] = &itemset2;
  countFlags[1] = TRUE;

  Set itemset3(2, items3);
  itemsets[2] = &itemset3;
  countFlags[2] = TRUE;

  Set itemset4(1, items4);
  itemsets[3] = &itemset4;
  countFlags[3] = TRUE;


  blkSeq->Count(itemsets, counts, 3);
  cout << " Counts for the itemsets " << endl;

  cout << itemset1 << " " << counts[0] << endl;
  cout << itemset2 << " " << counts[1] << endl;
  cout << itemset3 << " " << counts[2] << endl;
  cout << itemset4 << " " << counts[3] << endl;
  delete [] itemsets;
}
#endif




void error(bool usage, char * s, char * s1, char * s2 = "") {
  cerr << s << ": " << s1 << ' ' << s2 << endl;
  exit(1);
}




void InputParameters(int argc, char **argv, char *prefixName, int& h, int& a, int& b,
		     int& numItems, double& relSupport, bool& writeFlag, int& numBlks,
		     int& sampleSize, bool &lazy, 
		     ALGORITHM& algo, int& numSampleSizes, int& numSamplesPerSz,
		     DISTANCE_TYPE& devType, AGGFN_TYPE& aggType)
{
  // Optional arguments.
  //int setsPerLeaf = DEF_SETS_PER_LEAF;
  h = DEF_HASH_H;
  a = DEF_HASH_A;
  b = DEF_HASH_B;
  algo = MAINTENANCE;
  
  numItems = 0;
  relSupport = 0.0;
  numBlks = 0;

  int tmpFlag = 1;
  int tmpInt = 0;

    // Command line argument processing.

    if (argc > 1 && !strcmp(argv[1], "-help")) {
      //usage(argv[0]);
      cout << " usage error" << endl;
      exit(0);
    }

#ifdef STATIC
    if (argc < 9) error(true, argv[0], "wrong number of arguments");
#else
    if (argc < 7) error(true, argv[0], "wrong number of arguments");
#endif

    for (int i = 1; i < argc; i++) {
      switch (argv[i][1]) {
	
	// (Mostly) optiona largument input.
      case 's' :
#ifdef STATIC
	// Required argument for Sampling.
	sampleSize = atoi(argv[++i]);
	break;
#endif
      case 'l' :
	lazy = (atoi(argv[++i]) == 0) ? FALSE : TRUE;
	break;
      case 'd':
	tmpInt = atoi(argv[++i]);
	if (0 == tmpInt)
	  algo = MAINTENANCE;
	else if (1 == tmpInt)
	  algo = DEVIATIONS;
	else if (2 == tmpInt)
	  algo = BOOTSTRAP;
	else if (3 == tmpInt)
	  algo = DEVNS_ACTUAL;
	else if (4 == TRENDS)
	  algo = TRENDS;
	else
	  exit(1);
	break;

      case 'f':
	tmpInt = atoi(argv[++i]);
	if (0 == tmpInt)
	  devType = ABSOLUTE;
	else if (1 == tmpInt)
	  devType = RELATIVE;
	else
	  exit(1);
	break;

      case 'g':
	tmpInt = atoi(argv[++i]);
	if (0 == tmpInt)
	  aggType = L1;
	else if (1 == tmpInt)
	  aggType = L_INF;
	else
	  exit(1);
	break;

      case 'x':
	numSampleSizes = atoi(argv[++i]);
	break;

      case 'z':
	numSamplesPerSz = atoi(argv[++i]);
	break;

      case 'h' :
	h = atoi(argv[++i]);
	if (h <= 0)
	  error(false, argv[0], "hash_h must be greater than 0");
	break;
      case 'a' :
	a = atoi(argv[++i]);
	if (a <= 0)
	  error(false, argv[0], "hash_a must be greater than 0");
	break;
      case 'b' :
	b = atoi(argv[++i]);
	if (b < 0)
	  error(false, argv[0], "hash_b must be at least 0");
	break;
	
	// Required argument input.
      case 'i' :
	strcpy(prefixName, argv[++i]);
	break;
      case 'n' :
	numItems = atoi(argv[++i]);
	if (numItems <= 0)
	  error(false, argv[0], "num_items must be greater than 0");
	break;
      case 'm' :
	//minSupport = atoi(argv[++i]);
	relSupport = atof(argv[++i]);
	if (relSupport <= 0)
	  error(false, argv[0], "min_sup must be greater than 0");
	break;
      case 'k':
	numBlks = atoi(argv[++i]);
	if ((algo != DEVIATIONS) && (numBlks <= 0))
	  error(false, argv[0], "numBlks must be greater than 0");
	break;
      case 'w':
	tmpFlag = atoi(argv[++i]);
	writeFlag = (tmpFlag == 0) ? FALSE : TRUE;
	if (tmpFlag < 0)
	  error(false, argv[0], "flag must be 0 (False) or 1 (True)");
	break;
	/*
#ifdef SAMPLING
case 'l' :
lowMinSupport = atoi(argv[++i]);
if (lowMinSupport <= 0)
error(false, argv[0], "low_min_sup must be greater than 0");
break;
#endif
	*/
      default :
	error(true, argv[0], "unrecognized argument", argv[i]);
      }
    }
    
    //if (!xfile.is_open())
    //	error(true, argv[0], "transaction file (Xfile) not specified");
    if (!numItems)
	error(true, argv[0], "number of items (num_items) not specified");
    if (0.0 >= relSupport)
	error(true, argv[0], "minimum support (min_sup) not specified");
    /*
#ifdef SAMPLING
    if (!sampleSize)
	error(true, argv[0], "sample size (sample_size) not specified");
    if (!lowMinSupport)
	error(true, argv[0], "lowered minimum support (low_min_sup) not specified");
#endif
    */
}

