/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <assert.h>
#include <fstream.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "globals.h"
#include "tid-list.h"
#include "Set.h"

extern void formMaxItFileName(char* fileName, char *prefix, Set& it);

TIDList::TIDList(char *fname, int bufSize, bool readWrite)
{
  Init(fname, bufSize, readWrite);
}

TIDList::TIDList(char *fileNamePrefix, int bufSize, bool readWrite, int item1,
		 int item2=-1)
{
  char fname[MAX_NAME_LENGTH];

  if (-1 == item2)
    sprintf(fname, "%s.%d", fileNamePrefix, item1);
  else
    sprintf(fname, "%s.%d.%d", fileNamePrefix, item1, item2);

  Init(fname, bufSize, readWrite);
}

void TIDList::Init(char *fname, int bufSize, bool create)
{
  _fileName = new char[MAX_NAME_LENGTH];
  _fileSize = -1;
  strcpy(_fileName, fname);

#ifdef DEBUG
  //cout << " filename:  " << _fileName << endl;
#endif

  _beingUsed = TRUE;
  _readUse = (TRUE == create) ? FALSE : TRUE;
  // create a file for this TID-list;

  _maxCacheSize = bufSize;
  _curCacheOffset = 0;
  _cacheSize = 0;
  _cache = NULL;

  _nextTIDvalue = END_LIST;
  // initialize cache;
  if (TRUE == create)
    {
      _cache = new int[(int)(_maxCacheSize/sizeof(int))];
      assert(_cache != NULL);

      _fd = -1;
      _curFileOffset = 0;
    }
  //  else
  //_InitToRead();
}

void TIDList::_InitToRead()
{
  int nbytes;

  _fd = open(_fileName, O_RDONLY);
  assert(_fd != -1);
  // read a block from the file.


  if (_curFileOffset > 0)
    lseek(_fd, _curFileOffset, SEEK_SET);

  assert(_cache != NULL);
  nbytes = read(_fd, (void*)_cache, _maxCacheSize);
  assert(nbytes % sizeof(TIDType) == 0);

  close(_fd);
  _fd = -1;

  _cacheSize = nbytes;
  if (nbytes < _maxCacheSize)
    _lastBlock = TRUE;
  
  if (0 == nbytes)
    {
      _cache[0] = END_LIST;
      _nextTIDvalue = END_LIST;
    }
  else
    {
      _nextTIDvalue = _cache[0];
      _curCacheOffset = sizeof(TIDType); // offset in bytes.
      _curFileOffset += sizeof(TIDType);
    }
}

// We can repeatedly open and close a file without losing
// any information because _curFileOffset remembers the 
// location to be read or written.  

// TID lists are currently opened for counting itemsets
// and hence can be allocated memory independent of other TID lists
// as the number of such lists opened for reading is small;
// this part is commented out, for now. 
int TIDList::Open(bool FromStart, bool readStatus, int bufSize)
{
  _readUse = readStatus;
  if (NULL == _cache)
    {
      _maxCacheSize = PAGE_SIZE * INDIV_TIDLIST_BLKSZ;
      if ((_fileSize >= 0) && (_fileSize < _maxCacheSize))
	{
	  _maxCacheSize = _fileSize + sizeof(int);
	}
      _cache = new int[(int)(_maxCacheSize/sizeof(int))];
      //_cache = new int[(int)(_maxCacheSize/sizeof(int))];
      //cout << "TIDL:Open :: " << _maxCacheSize << endl;
    }
  
  if (TRUE == FromStart)
    {
      _curFileOffset = 0;
      _lastBlock = FALSE;
      _cacheSize = 0;
      _curCacheOffset = 0;
    }

  /***
  if (TRUE == _beingUsed)
    {
      if (BeingRead())
	_InitToRead();
      return SUCCESS;
    }
  ***/
  _beingUsed = TRUE;

  if (BeingRead())
    _InitToRead();
  else
    _curCacheOffset = 0;
  
  return SUCCESS;
}

void TIDList::Reset()
{
  Open(TRUE,TRUE);
}


// can close some of the TID-lists if there is not enough
// main memory, or if the number of open files is too high.
int TIDList::Close()
{
  int nbytes = 0;

  if (FALSE == BeingUsed())
    return SUCCESS;

  // write the last cache block, if any, and close.

  _beingUsed = FALSE;
  if ((!BeingRead()) && (_curCacheOffset > 0))
    {
      _fd = open(_fileName, O_RDWR | O_CREAT, 511);
      assert(_fd != -1);
      //cout << " closing " << _fileName << endl;
      nbytes = pwrite(_fd, (void*)_cache, _curCacheOffset, _curFileOffset);
      assert(nbytes == _curCacheOffset);
      close(_fd);
      _curFileOffset += _curCacheOffset;
      if (-1 == _fileSize)
	_fileSize = _curFileOffset;
      else
	{
	  assert(0); // for now, this should not be done.
	  _fileSize += _curFileOffset;
	}
    }
  _fd = -1;
  
  _curCacheOffset = 0;
  delete [] _cache;
  _cache = NULL;
    
  return SUCCESS;
}
 



