/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef __TIDLIST_H__
#define __TIDLIST_H__

#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <fstream.h>

#include "globals.h"
#include "Set.h"

class TIDList 
{
 private:
  char *_fileName;
  int _fd;
  //int _size;     // number of tid's the list.
  int _curFileOffset;  // current place to be read or written in the file.
  int _curCacheOffset; // -do- in the cache.
  
  int *_cache;   // a block in memory.
  int _maxCacheSize;
  int _cacheSize; 
  int _fileSize;

  int _nextTIDvalue; // while reading in; (blocks)
  bool _lastBlock;
  bool _readUse;
  bool _beingUsed;

  void _InitToRead(); // to be called after the constructor or Init.

 public:
  TIDList():_fd(-1),_curFileOffset(-1),_curCacheOffset(-1), _maxCacheSize(-1), _cacheSize(0){};
  TIDList(char *fname, int bufSize, bool read = TRUE);
  TIDList(char *prefixName, int bufSize, bool read, int item1, int item2=-1);
  ~TIDList()
    {
      //assert(_fileName);
      //assert(_cache);
      delete [] _fileName;
      delete [] _cache;
    }
  
  void Init(char *fname, int bufSize, bool read = TRUE);

  // Is the file open or not?
  bool OpenStatus() { return ((_fd > 0) ? TRUE : FALSE);}
  bool BeingRead() { return (TRUE == _readUse);}
  char* FileName() { return _fileName;}
  int Space() { return _curFileOffset;}


  bool BeingUsed() {return _beingUsed;}
  int Open(bool fromStart, bool readStatus, int bufSize=-1);
  int Close();

  void Reset();
  TIDType PeekNextTID() { return _nextTIDvalue;}
  inline TIDType ReadNextTID();
  

  inline void WriteTID(TIDType tid);
  
  bool Matches(Set& itemset, char *prefixName)
    {
      char tmpName[MAX_NAME_LENGTH];
      strcpy(tmpName, prefixName);

      int sz = itemset.size();
      int i;
      
      for (i = 0; i < sz; i++)
	sprintf(tmpName+strlen(tmpName), ".%d", itemset.elements(0));

      if (0 == strcmp(tmpName, _fileName))
	return TRUE;
      return FALSE;
    }
};




TIDType TIDList::ReadNextTID()
{
  int nbytes;
  TIDType tidValue;
  assert(_cache != NULL);
  
  if (0 == _curCacheOffset) // have to read a new block.
    {
      _fd = open(_fileName, O_RDONLY);
      assert(_fd != -1);
      if (TRUE == _lastBlock)
	_cache[0] = END_LIST;
      else
	{
	  nbytes = pread(_fd, (void*)_cache, _maxCacheSize, _curFileOffset);
	  _cacheSize = nbytes;
	  if (nbytes < _maxCacheSize)
	    _lastBlock = TRUE;
	  if (0 == nbytes)
	    _cache[0] = END_LIST;
	  
	  //_nextTIDvalue = _cache[0];
	  //_curCacheOffset = sizeof(TIDType);
	}
      close(_fd);
      _fd = -1;
    }
  
  tidValue = _nextTIDvalue; 
  _nextTIDvalue = _cache[(int)(_curCacheOffset/sizeof(TIDType))];
  
  _curCacheOffset += sizeof(TIDType);
  _curFileOffset += sizeof(TIDType); // assuming TID is of type int.
  if (_curCacheOffset == _cacheSize)
    _curCacheOffset = 0;
  
  return tidValue;
}

void TIDList::WriteTID(int tidValue)
{
  int nbytes;

  assert((_curCacheOffset >= 0) && (!BeingRead()));
  assert(_cache != NULL);

  _cache[(int)(_curCacheOffset/sizeof(TIDType))] = tidValue;
  _curCacheOffset += sizeof(TIDType);
  if (_curCacheOffset == _maxCacheSize)
    {// write the cache block.
#ifdef DEBUG
      //cout << " Writing a cache block" << endl;
#endif

      //cout << " FileName " << _fileName << endl;
      _fd = open(_fileName, O_RDWR | O_CREAT, 511);
      assert(_fd != -1);

      nbytes = pwrite(_fd, (void*)_cache, _maxCacheSize, _curFileOffset);
      assert (nbytes == _maxCacheSize);
      _curFileOffset += _maxCacheSize;
      _curCacheOffset = 0;

      close(_fd);
      _fd = -1;
    }
}


#endif

