#include <assert.h>
#include <fstream.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "globals.h"
#include "tid-list.h"
#include "Set.h"

extern void formMaxItFileName(char* fileName, char *prefix, Set& it);

TIDList::TIDList(char *fname, int bufSize, bool readWrite)
{
  Init(fname, bufSize, readWrite);
}

TIDList::TIDList(char *fileNamePrefix, int bufSize, bool readWrite, int item1,
		 int item2=-1)
{
  char fname[MAX_NAME_LENGTH];

  if (-1 == item2)
    sprintf(fname, "%s.%d", fileNamePrefix, item1);
  else
    sprintf(fname, "%s.%d.%d", fileNamePrefix, item1, item2);

  Init(fname, bufSize, readWrite);
}

void TIDList::Init(char *fname, int bufSize, bool create)
{
  _fileName = new char[MAX_NAME_LENGTH];
  strcpy(_fileName, fname);

#ifdef DEBUG
  //cout << " filename:  " << _fileName << endl;
#endif

  _readUse = (TRUE == create) ? FALSE : TRUE;
  // create a file for this TID-list;

  _maxCacheSize = bufSize;
  _curCacheOffset = 0;
  _cacheSize = 0;
  
  _nextTIDvalue = -1;
  // initialize cache;
  _cache = new int[(int)(_maxCacheSize/sizeof(int))];
  assert(_cache != NULL);

  if (TRUE == create)
    {
#ifndef FANCYTID
      _fd = open(_fileName, O_RDWR | O_CREAT, 511);
      assert(_fd != -1);
#else
      _fd = -1;
#endif
      _curFileOffset = 0;

    }
  else
    _InitToRead();
}

void TIDList::_InitToRead()
{
  int nbytes;

#ifdef FANCYTID
  _fd = open(_fileName, O_RDONLY);
#endif
  assert(_fd != -1);
  // read a block from the file.


  if (_curFileOffset > 0)
    lseek(_fd, _curFileOffset, SEEK_SET);

  assert(_cache != NULL);
  nbytes = read(_fd, (void*)_cache, _maxCacheSize);
  assert(nbytes % sizeof(TIDType) == 0);

#ifdef FANCYTID
  close(_fd);
  _fd = -1;
#endif

  _cacheSize = nbytes;
  if (nbytes < _maxCacheSize)
    _lastBlock = TRUE;
  
  if (0 == nbytes)
    {
      _cache[0] = END_LIST;
      _nextTIDvalue = END_LIST;
    }
  else
    {
      _nextTIDvalue = _cache[0];
      _curCacheOffset = sizeof(TIDType); // offset in bytes.
    }
}

// We can repeatedly open and close a file without losing
// any information because _curFileOffset remembers the 
// location to be read or written.  
int TIDList::Open(bool FromStart, bool readStatus)
{
#ifndef FANCYTID
  _fd = open(_fileName, O_RDONLY);
  assert(_fd != -1);
#endif
  if (NULL == _cache)
    _cache = new int[(int)(_maxCacheSize/sizeof(int))];

  if (TRUE == FromStart)
    {
      _curFileOffset = 0;
      _lastBlock = FALSE;
      _cacheSize = 0;
      _curCacheOffset = 0;
    }

  _readUse = readStatus;
  
  if (BeingRead())
    _InitToRead();
  else
    _curCacheOffset = 0;
  
  return SUCCESS;
}

void TIDList::WriteTID(int tidValue)
{
  int nbytes;

#if 0  
  assert((_curCacheOffset >= 0) && (!BeingRead()));
  assert(_cache != NULL);
#endif

  _cache[(int)(_curCacheOffset/sizeof(TIDType))] = tidValue;
  _curCacheOffset += sizeof(TIDType);
  if (_curCacheOffset == _maxCacheSize)
    {// write the cache block.
#ifdef DEBUG
      //cout << " Writing a cache block" << endl;
#endif

#ifdef FANCYTID
      _fd = open(_fileName, O_RDWR | O_CREAT, 511);
#endif
      assert(_fd != -1);

      nbytes = pwrite(_fd, (void*)_cache, _maxCacheSize, _curFileOffset);
      assert (nbytes == _maxCacheSize);
      _curFileOffset += _maxCacheSize;
      _curCacheOffset = 0;

#ifdef FANCYTID
      close(_fd);
      _fd = -1;
#endif
    }
}


TIDType TIDList::ReadNextTID()
{
  int nbytes;
  TIDType tidValue;
  assert(_cache != NULL);
  
  if (0 == _curCacheOffset) // have to read a new block.
    {
#ifdef FANCYTID
      _fd = open(_fileName, O_RDONLY);
#endif
      assert(_fd != -1);
      if (TRUE == _lastBlock)
	_cache[0] = END_LIST;
      else
	{
	  nbytes = pread(_fd, (void*)_cache, _maxCacheSize, _curFileOffset);
	  _cacheSize = nbytes;
	  if (nbytes < _maxCacheSize)
	    _lastBlock = TRUE;
	  if (0 == nbytes)
	    _cache[0] = END_LIST;
#ifdef FANCYTID
	  close(_fd);
	  _fd = -1;
#endif	 
	  //_nextTIDvalue = _cache[0];
	  //_curCacheOffset = sizeof(TIDType);
	}
      close(_fd);
      _fd = -1;
    }
  
  tidValue = _nextTIDvalue; 
  _nextTIDvalue = _cache[(int)(_curCacheOffset/sizeof(TIDType))];
  
  _curCacheOffset += sizeof(TIDType);
  _curFileOffset += sizeof(TIDType); // assuming TID is of type int.
  if (_curCacheOffset == _cacheSize)
    _curCacheOffset = 0;
  
  return tidValue;
}


// can close some of the TID-lists if there is not enough
// main memory, or if the number of open files is too high.
int TIDList::Close()
{
  int nbytes = 0;

  // write the last cache block, if any, and close.
  if ((!BeingRead()) && (_curCacheOffset > 0))
    {
#ifdef FANCYTID
      //cout << " closing " << _fileName << endl;
      _fd = open(_fileName, O_RDWR | O_CREAT, 511);
      assert(_fd != -1);
#endif
      assert(_fd != -1);

      nbytes = pwrite(_fd, (void*)_cache, _curCacheOffset, _curFileOffset);
      assert(nbytes == _curCacheOffset);
      _curFileOffset += _curCacheOffset;
      
      close(_fd);
      _fd = -1;
    }

  _curCacheOffset = 0;
  delete [] _cache;
  _cache = NULL;
  
  return SUCCESS;
}
 



