/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "globals.h"
#include "tidlist-colln.h"
#include "tid-list.h"
#include "Set.h"
#include "PrefixTree.h"
//#include "hashtable_v.h"
#include "2DArray.h"
#include "trans_utils.h"
#include "timeutil.h"

extern int hash(Set& itemset, int size);
extern void  formMaxItFileName(char *fileName, char *_fileNamePrefix, Set& set);

TIDList_HTEntry::TIDList_HTEntry()
{
  tidListptr = new TIDList();
  next = NULL;
}

TIDList_HTEntry::~TIDList_HTEntry()
{
  if (NULL != next) delete next;
  delete tidListptr;
}

void TIDList_HTEntry::Init(char *tmpname, int bufSz, bool flag)
{
  tidListptr->Init(tmpname, bufSz, flag);
}

PrefixNode_HTEntry::PrefixNode_HTEntry()
{
  pfnptr = new PrefixNode();
  next = NULL;
}

PrefixNode_HTEntry::~PrefixNode_HTEntry()
{
  if (NULL != next) delete next;
  delete pfnptr;
}

TIDListColln::TIDListColln(int nl, int partN, char *prefixName, int *bufSize, 
			   bool create, char *datafile)
{
  _nextBlk = NULL;
  _prevBlk = NULL;
  _blockNum = partN;
  _numLists = nl;
  _fileNamePrefix = new char[MAX_NAME_LENGTH];
  _datafile = new char[MAX_NAME_LENGTH];
  strcpy(_datafile, datafile);

  sprintf(_fileNamePrefix,"%s/b%d.t", prefixName, _blockNum);

#ifdef DEBUG  
  cout << " Inside TLC: constructor " << _numLists << " " << _fileNamePrefix << endl;
#endif

  int i;
  char tmpName[MAX_NAME_LENGTH];

  _tidLists = new TIDList[_numLists];
  strcpy(tmpName, _fileNamePrefix);
  for (i = 0; i < nl; i++)
    {
      // REVISIT: try avoiding sprintf.
      sprintf(tmpName+strlen(_fileNamePrefix),".%d", i);
      _tidLists[i].Init(tmpName, PAGE_SIZE * bufSize[i], create);
    }

#ifdef DEBUG
  cout << " In TIDListC constr: constructed tidlist " << endl;
#endif

#ifdef FANCYTID
  _numMaxTIDLists = (int)(0.5 * nl * nl);
  cout << " TIDListConstructor: Maximum number of 2-itemset tid lists " << _numMaxTIDLists << endl;
  _relevantTIDLists = new (TIDList*)[_numMaxTIDLists];
  memset((void*)_relevantTIDLists, 0, _numMaxTIDLists * sizeof(TIDList*));
  // create the strucutures for 2-itemset TIDlists here;

  //_materializedMaxIts = new HashTable(_numLists, TIDLIST, _fileNamePrefix);
  //_deletedMaxIts = new HashTable(MAX_DEL_ITS, COUNTSET, _fileNamePrefix);
#endif

}


void TIDListColln::AllocateSpaceForFiles(int *supports)
{
  char tmpFileName[MAX_NAME_LENGTH];
  int *tmpArr=NULL;
  int tmpFd;
  int i;

  assert(0 == _blockNum);
  for (i = 0; i < _numLists; i++)
    {
      sprintf(tmpFileName, "%s/b%d.t%d", _fileNamePrefix, _blockNum, i);
      cout << "Allocating space for " << tmpFileName << " " << supports[i] << endl;
      tmpArr = new int[supports[i]];

      memset((void*)tmpArr, 0, sizeof(int)*supports[i]);
      tmpFd = open(tmpFileName, O_TRUNC | O_CREAT, 511);
      write(tmpFd, (void*)tmpArr, sizeof(int)*supports[i]);
      close(tmpFd);
      delete [] tmpArr;
    }
}

TIDListColln::~TIDListColln()
{
  delete [] _fileNamePrefix;
  delete [] _datafile;
  delete [] _tidLists;
#ifdef FANCYTID
  int i;
  for (i = 0; i < _numMaxTIDLists; i++)
    if (_relevantTIDLists[i] != NULL)
      delete _relevantTIDLists[i];
  delete [] _relevantTIDLists;
  //delete _materializedMaxIts;
  //delete _deletedMaxIts;
#endif
}

#ifdef FANCYTID
TIDList* TIDListColln::CreateTIDListToWrite(int *arr, int itemsetSz)
{
  Set set(0);
  set.newArr(itemsetSz, arr);
  char fileName[MAX_NAME_LENGTH];

  //TIDList* temp = _materializedMaxIts->SearchToWrite(set, -1);
  formMaxItFileName(fileName, _fileNamePrefix, set);
  TIDList* temp = new TIDList(fileName, SMALL_BLKSZ * PAGE_SIZE, TRUE);
  assert(temp != NULL);

  set.Nullify();
  return temp;
}
#endif

void TIDListColln::WriteXact(Set *xactn, int tid)
{
  int i;
  int xSz = xactn->size();

  for (i = 0; i < xSz; i++)
    {
      _tidLists[xactn->elements(i)].WriteTID(tid);
    }
}

int TIDListColln::WriteInfoToDisk(PrefixTree *pt=NULL)
{
  // close all the TIDLists;
#ifdef DEBUG
  cout << "Closing all TID lists " << endl;
#endif
#ifdef FANCYTID
  assert(pt != NULL);
  pt->CloseAllTIDLists(this); 
#endif
  _CloseAllTIDLists();

  // write the deleted frequent itemsets.
  // TO BE DONE
  return 0;
}

void TIDListColln::_OpenTIDLists(int *items, int numitems)
{
  int i;
  for (i = 0; i < numitems; i++)
    _tidLists[items[i]].Open(TRUE,TRUE);
}

void TIDListColln::_CloseTIDLists(int *items, int numItems)
{
  int i;
  for (i = 0; i < numItems; i++)
    _tidLists[items[i]].Close();
}

void TIDListColln::_CloseAllTIDLists()
{
  int i;
  for (i = 0; i < _numLists; i++)
    _tidLists[i].Close();

#ifdef FANCYTID
  int num = 0;
  int sp = 0; 
  for (i = 0; i < _numMaxTIDLists; i++)
    if (_relevantTIDLists[i] != NULL)
      {
	sp += _relevantTIDLists[i]->Space();
	_relevantTIDLists[i]->Close();
	num++;
      }
  cout << " ** Number of 2-itemset fancy lists " << num << endl;
  cout << " ** Space occupied by these fancy lists " << sp << endl;
#endif
}

int TIDListColln::_CountFromDelIts(Set **itemsets, int *countArr, bool *countFlgs, 
				   int numItemsets)
{
  return 0; // Have to implement this one. 
}

#ifdef FANCYTID
int TIDListColln::_FindRelItList(Set **itemsets, int numItemsets, 
				 bool *countFlags, int*& relItList, 
				 int totNumItems)
{
  int i,j;
  int numRelItems = 0;
  int *relevantItems = new int[totNumItems];

  for (i = 0; i < totNumItems; i++)
    relevantItems[i] = -1;

  for (i = 0; i < numItemsets; i++)
    {
      if (FALSE == countFlags[i])
	continue;
      for (j = 0; j < itemsets[i]->size(); j++)
	if (relevantItems[itemsets[i]->elements(j)] == -1)
	  {
	    relevantItems[itemsets[i]->elements(j)] = 1;
	    numRelItems++;
	  }
    }
  
  relItList = new int[numRelItems];

  j = 0;
  for (i = 0; i < totNumItems; i++)
    if (1 == relevantItems[i])
      relItList[j++] = i;

  delete [] relevantItems;
  return numRelItems;
}

int TIDListColln::_OpenRelTIDLists(Set **itemsets,
				   int numItemsets, bool *countFlags,
				   int totNumItems, TwoDArray& indxToTIDLists)
{
  int i, j, k;
  int indx;
  int numOpened = 0;
  int nextSubset;
  int item1, item2;
  bool covered[MAX_IT_LENGTH];
  int tmp;

  for (i = 0; i < numItemsets; i++)
    {
      if (FALSE == countFlags[i])
	continue;

      nextSubset = 1;
      for (j = 0; j < itemsets[i]->size(); j++)
	covered[j] = FALSE;

      for (j = 0; j < itemsets[i]->size(); j++)
	{
	  if (TRUE == covered[j])
	    continue;
	  item1 = itemsets[i]->elements(j);
	  tmp = totNumItems + item1*totNumItems;

	  for (k = j+1; k < itemsets[i]->size(); k++)
	    {
	      if (TRUE == covered[k])
		continue;

	      item2 = itemsets[i]->elements(k);
	      indx = _getIndx(item1, item2, totNumItems);
	      
	      if (TRUE == _MaterializedIt(item1, item2))
		{
		  covered[j] = covered[k] = TRUE;
		  assert(_relevantTIDLists[indx] != NULL);
		  _relevantTIDLists[indx]->Open(TRUE, TRUE);
		  numOpened++;
		  indxToTIDLists.set(i, nextSubset, indx);
		  nextSubset++;
		}
	    }
	  if (FALSE == covered[j])
	    {
	      covered[j] = TRUE;
	      _relevantTIDLists[item1] = _tidLists + item1;
	      (_relevantTIDLists[item1])->Open(TRUE, TRUE);
	      numOpened++;
	      indxToTIDLists.set(i, nextSubset, item1);
	      nextSubset++;
	    }
	}
      indxToTIDLists.set(i, 0, nextSubset-1);
    }
  cout << "_OpenRelTIDLists : " << *(itemsets[0]) << " " << numOpened << endl;
  return numOpened;
}




void 
TIDListColln::CountFancy(Set **itemsets, int *countArray, bool *countFlags,
			 int numItemsets, int totNumItems, int numMaxTIDLists,
			 TIDType *tids, TwoDArray *indxToTIDLists)
{
  int i;
  for (i = 0; i < numItemsets; i++)
    _CountFancySingle(itemsets+i, countArray+i, countFlags+i, 1, totNumItems, 
		      numMaxTIDLists, tids, indxToTIDLists);
}


#endif


void TIDListColln::CountUsingPT(PrefixTree *pt)
{
  pt->CountItemsets(_datafile);
}


void TIDListColln::Count(Set **itemsets, int *countArray, bool *countFlags,
			 int numItemsets)
{
  int i,j;
  int *tids = new int[_numLists];
  TIDType tidUsed = MAX_TID;

  if (0 == numItemsets)
    return;

  // TO BE DONE: check in the list of itemsets kicked out 
  // from this block;

  // here, we can check if any of the itemsets were part of the 
  // prefix tree before this block was appended.

  //int numLarge = 0;
  //numLarge = _CountFromDelIts(itemsets, countArray, countFlags, numItemsets);

  int matchCnt;
  bool flag;
  for (i = 0; i < numItemsets; i++)
    {
      if (FALSE == countFlags[i])
	continue;
      for (j = 0; j < itemsets[i]->size(); j++)
	{
	  tids[itemsets[i]->elements(j)] = -1;
	  _tidLists[itemsets[i]->elements(j)].Open(TRUE,TRUE);
	}

      tidUsed = 0;
      matchCnt = 0;
      j = 0;
      flag=TRUE;
      while (flag == TRUE)
	{
	  while (tids[itemsets[i]->elements(j)] < tidUsed)
	    {
	      tids[itemsets[i]->elements(j)] = _tidLists[itemsets[i]->elements(j)].ReadNextTID();
	    }

	  if (tidUsed == tids[itemsets[i]->elements(j)])
	    {
	      matchCnt++;
	      if (matchCnt == itemsets[i]->size())
		{
		  countArray[i]++;
		  tids[itemsets[i]->elements(j)] = _tidLists[itemsets[i]->elements(j)].ReadNextTID();
		  tidUsed++;

		  //if (tids[itemsets[i]->elements(j)] == END_LIST)
		  //flag = FALSE;
		  matchCnt = 0;
		}
	    }
	  else if (tids[itemsets[i]->elements(j)] == END_LIST)
	    flag = FALSE;
	  else
	    {
	      tidUsed = tids[itemsets[i]->elements(j)];
	      matchCnt = 1;
	    }
	  j = (j+1)%(itemsets[i]->size());
	}
    }
  // reset tid lists for possible later use;
  for (i = 0; i < numItemsets; i++)
    for (j = 0; j < itemsets[i]->size(); j++)
      _tidLists[itemsets[i]->elements(j)].Close();

#ifdef DEBUG
  for (i = 0; i < numItemsets; i++)
    cout << *(itemsets[i]) << " : " << countArray[i] << endl;
#endif

  delete [] tids;
}




#ifdef INSTRUMENT
void SelectRandIts(Set **itemsetArr, int numNegBIts, int num, Set **setArr)
{
  assert(num < numNegBIts);
  int randNum;
  srandom(1325);
  int i, j;
  bool flag;

  i = 0;
  while (i < num)
    {
      randNum = random() % numNegBIts;
      flag = FALSE;
      for (j = 0; j < i; j++)
	if (setArr[j] == itemsetArr[randNum])
	  {
	    flag = TRUE;
	    break;
	  }
      if (FALSE == flag)
	{
	  setArr[i] = itemsetArr[randNum];
	  i++;
	}
    }
}

static int staticTidCountTimes=1;
static int staticPTCountTimes=0;
void TIDListColln::CountRandomItemsets(PrefixTree *pt, int numItemsets,
				       int totNumItems=-1, int numMaxTIDLists=-1, 
				       TIDType *tids=NULL, TwoDArray *indxToTIDLists = NULL)
{
  int maxCount = 1000;
  Set **itemsetArr = new (Set*)[numItemsets];
  Set **setArr = new (Set*)[maxCount];
  int *countArr = new int[maxCount];
  bool *countFlags = new bool[maxCount];
  int i, num;
  int numNegBIts;
  int nitems = pt->NumItems();

  char tmpName1[MAX_NAME_LENGTH];
  char tmpName2[MAX_NAME_LENGTH];
  ofstream of1, of2;
  
  system("sleep 10");

#ifdef FANCYTID
  sprintf(tmpName1, "/tmp/tid-vs-pt.countTimes/numbers.tid.%3f.fancy", 
	  pt->RelSupport());
  sprintf(tmpName2, "/tmp/tid-vs-pt.countTimes/numbers.pt.%3f.fancy", 
	  pt->RelSupport());
#else
  sprintf(tmpName1, "/tmp/tid-vs-pt.countTimes/numbers.tid.%3f",
	  pt->RelSupport());
  sprintf(tmpName2, "/tmp/tid-vs-pt.countTimes/numbers.pt.%3f",
	  pt->RelSupport());
#endif

  cout << "Counting times written to " << tmpName1 << " " << tmpName2 << endl;

  if (staticTidCountTimes != 0)
    {
      of1.open(tmpName1);
      assert(of1 != NULL);
    }

  // We don't need to count using PT again here too. 
  // But to make sure that fancy lists are not in memory already, we do.
  if (staticPTCountTimes > 0)
    {
      of2.open(tmpName2); 
      assert(of2 != NULL);
    }

  Timer T1("Counting Time using TID-lists");
  Timer T2("Counting Time using PT and a data scan");

  numNegBIts = pt->GetRandNegBItemsets(itemsetArr, numItemsets);
  for (i = 0; i < numNegBIts; i++)
    {
      countArr[i] = 0;
      countFlags[i] = TRUE;
    }

  cout << " CountRand: numNegBIts " << numNegBIts << endl;
  for (i = 0; i < numNegBIts; i++)
    {
      cout << *itemsetArr[i] << endl;
    }

  for (num = 5; num < numNegBIts; num += 10)
    {
      SelectRandIts(itemsetArr, numNegBIts, num, setArr);
    
      if (staticPTCountTimes > 0)
	{
	  cout << " Counting using prefix tree " << endl;
	  T2.Clear();
	  T2.Start();
	  // count using the prefix tree;
	  PrefixTree *tempPt = new PrefixTree(nitems, setArr, num);
	  CountUsingPT(tempPt);
	  //tempPt->setMinSupport(1);
	  //cout << *tempPt << endl;
#ifdef DEBUG
	  for (i = 0; i < num; i++)
	    {
	      cout << *setArr[i] << " PTCount: " << tempPt->count(*setArr[i]);
	    }
#endif
	  delete tempPt;
	  T2.Stop();
	  of2 << num <<" ";
	  T2.Output(of2);
	}

      if (staticTidCountTimes > 0)
	{
	  cout << "Counting using TID lists " << endl;
	  T1.Clear();
	  T1.Start();
#ifdef FANCYTID
	  assert(_relevantTIDLists != NULL);
	  assert(tids != NULL);
	  assert(indxToTIDLists != NULL);
	  CountFancy(setArr, countArr, countFlags, num, totNumItems, 
		     numMaxTIDLists, tids, indxToTIDLists);
#else
	  Count(setArr, countArr, countFlags, num);
#endif
	  T1.Stop();
	  of1 << num << " ";
	  T1.Output(of1);
	}
 
      for (i = 0; i < num; i++)
	{
	  setArr[i] = NULL;
	  countArr[i] = 0;
	}
      
    }
  if (staticTidCountTimes > 0)
    of1.close();
  
  if (staticPTCountTimes > 0)
    of2.close();

  for (i = 0; i < numNegBIts; i++)
    delete itemsetArr[i];

  delete [] setArr;
  delete [] itemsetArr;
  delete [] countArr;
  delete [] countFlags;
}
#endif

// rel for relevant.


/****
void TIDListColln::_ReadFirstRow(TIDType *tids, TIDType& tidUsed, 
				 int **indxToTIDLists, int numItemsets, 
				 TIDList **relTIDs)
{
  int i;

  for (i = 0; i < numRelTIDs; i++)
    {
      if (NULL != relTIDs[i])
	{
	  relTIDs[i].Open(TRUE, TRUE);
	  tids[i] = relTIDs[i].ReadNextTID();
	  if ((tids[i] != END_LIST) && (tids[i] < tidUsed))
	    tidUsed = tids[i];
	}
    }
}
****/

