/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef __TIDLISTCOLLN_H__
#define __TIDLISTCOLLN_H__

#include <fcntl.h>
#include <sys/types.h>
#include <stdio.h>

#include "Set.h"
#include "2DArray.h"
#include "tid-list.h"

class TwoDArray;
//class Set;
//class TIDList;
struct PrefixNode;
class PrefixTree;
class HashTable;

struct TIDList_HTEntry
{
  TIDList *tidListptr;
  TIDList_HTEntry *next;

  TIDList_HTEntry();
  ~TIDList_HTEntry();
  void Init(char *tmpname, int bufSz, bool flag);
};

struct PrefixNode_HTEntry
{
  PrefixNode *pfnptr;
  PrefixNode_HTEntry *next;

  PrefixNode_HTEntry();
  ~PrefixNode_HTEntry();
};

class TIDListColln
{
 private:
  int _blockNum; // block number in the sequence.
  char *_fileNamePrefix;
  char *_datafile;

  /*
   *TIDList information
   *number of lists;
   *array of TIDLists;
   *a 2-d array indexing TIDLists of 2-itemsets;
   */
  int _numLists;
  TIDList *_tidLists; 
  TwoDArray *_indxToTIDLists;

#ifdef FANCYTID
  int _numMaxTIDLists;
  TIDList **_relevantTIDLists;
#endif

  //to maintain a doubly-linked structure.
  TIDListColln *_nextBlk;
  TIDListColln *_prevBlk;

  void _OpenTIDLists(int *items, int numitems);
  void _CloseTIDLists(int *items, int numitems);
  void _CloseAllTIDLists();

#ifdef FANCYTID
  int _getIndx(int item1, int item2, int tot)
    {
      assert(item1 < item2);
      return (tot + item1 * tot - item1*(item1+1)/2 + (item2-item1+1));
    }

  /*
   *checks if the TIDList of the pair (item1,item2) is materialized.
   *if so, returns TRUE;
   */

  int _MaterializedIt(int item1, int item2)
    {
      char fname[MAX_NAME_LENGTH];
      int fd;

      sprintf(fname, "%s.%d.%d", _fileNamePrefix, item1, item2);
      fd = open(fname, O_RDONLY);
      if (-1 == fd)
	return FALSE;
      close(fd);
      return TRUE;
    }

  /*
   *Find the set of all items involved in **itemsets
   *returns the number of items in the resulting set;
   */
  int _FindRelItList(Set **itemsets, int numItemsets, bool *countFlags, 
		     int*& relItList, int totNumItems);

  /*
   *Open the set of TIDLists that can be used to count all itemsets
   *in **itemsets;
   *Returns the number of lists opened.
   */
  int _OpenRelTIDLists(Set **itemsets,
		       int numItemsets, bool *countFlags,
		       int totNumItems, TwoDArray& indxToTIDLists);

  /*
   *Read the next row of tids in the relevant tidlists;
   *numItemsets is the number of itemsets;
   *indxToTIDLists is the index into the sets of TIDLists, one per
   *itemset to be counted. 
   */
  inline int _ReadNextRow_Fancy(TIDType *tids, TIDType& tidUsed, 
				TwoDArray& indxToTIDLists, int numItemsets,
				bool FIRSTROW=FALSE);

  /*
   *After reading a new row of tids while merging them, this function can be
   *used to increment the counts of all itemsets that should be;
   */
  inline void _CountFancySingle(Set **itemsets, int *countArray, bool *countFlags,
				int numItemsets,int totNumItems,int numMaxTIDLists, 
				TIDType *tids, TwoDArray *indxToTIDLists);

#endif

  //int _ReadNextRow(TIDType *tids, int& tidUsed, int* tidListNums, int numLists);
  // Can we improve the following routine?
  // Currently, it loops I times.
  inline int _ReadNextRow(TIDType *tids, TIDType& tidUsed, int *tidListNums, 
			  int numRelTidLists);

  int _CountFromDelIts(Set **itemsets=NULL, int *countArr=NULL, 
		       bool *countFlgs=NULL, int numItemsets=-1);

  inline void _CountOneRow(TIDType* tids, Set **itemsets, int *countArray,
			   int numItemsets, TIDType tidUsed, 
			   TwoDArray *indxToTIDLists = NULL);

 public:
  TIDListColln(int nl, int partition, char* prefixName, int *bufSz, bool create=TRUE, char *datafile = NULL);
  ~TIDListColln();

  int Space()
    {
      int sp = 0;
#ifdef FANCYTID
      int i;
      for (i = 0; i < _numMaxTIDLists; i++)
	{
	  if (_relevantTIDLists[i] != NULL)
	    sp += _relevantTIDLists[i]->Space();
	}
#endif
      return sp;
    }
   
  void AllocateSpaceForFiles(int *supports);
  void WriteXact(Set *xactn, int tid);
  // void UpdatePrefixTree(Set *newXactBlk); -PT
  // void FindL2SItemsets(); -PT

#ifdef FANCYTID
  TIDList* CreateTIDListToWrite(int *arr, int itemsetSz);

  void CountFancy(Set **itemsets, int *countArray, bool *countFlags,
		  int numItemsets, int totNumItems, int numMaxTIDLists, 
		  TIDType *tids, TwoDArray *indxToTIDLists);

  void StorePtrToTIDList(int *arr, int sz, TIDList *tidListPtr)
    {
      assert(sz == 2);
      int indx = _getIndx(arr[0], arr[1], _numLists);

#ifdef DEBUG
      cout << " fancy tidlist " << arr[0] << " " << arr[1] << " " << indx << endl;
#endif
      assert(_relevantTIDLists[indx] == NULL);

      _relevantTIDLists[indx] = tidListPtr;
    }
#endif

  /*
   *Count the supports of itemsets in pt against _datafile;
   */
  void CountUsingPT(PrefixTree *pt);

  int WriteInfoToDisk(PrefixTree *pt=NULL);
  void Count(Set **itemsets, int *countArray, bool *countFlags,int numItemsets);

  TIDListColln* NextBlk() { return _nextBlk; }
  TIDListColln* PrevBlk() { return _prevBlk; }

  void SetNextBlk(TIDListColln* newBlk) { _nextBlk = newBlk;}
  void SetPrevBlk(TIDListColln* newBlk) { _prevBlk = newBlk;}

  inline int NumBlk() { return _blockNum; }

#ifdef FANCYTID
  /*
   *Write the TIDLists of items and 2-itemsets for the datafile;
   */
  void WriteFancyTIDLists(char *datafile, PrefixTree *pt);
#endif

#ifdef INSTRUMENT
  void CountRandomItemsets(PrefixTree *pt, int numItemsets, int totNumItems=-1,
			   int numMaxTIDLists=-1,
			   TIDType *tids=NULL, TwoDArray *indxToTIDLists=NULL);
#endif
};

#ifdef FANCYTID
int TIDListColln::_ReadNextRow_Fancy(TIDType *tids, TIDType& tidUsed, 
				     TwoDArray& indxToTIDLists, int numItemsets,
				     bool FIRSTROW)
{
  int i, j;
  int numRead = 0;
  TIDType temp = MAX_TID;
  int indx;

  for (i = 0; i < numItemsets; i++)
    {
#if 0
      for (j = 1; j <= indxToTIDLists.get(i,0); j++)
	cout << indxToTIDLists.get(i,j) << " ";
      cout << endl;	  
#endif      
      for (j = 1; j <= indxToTIDLists.get(i,0); j++)
	{
	  indx = indxToTIDLists.get(i,j);

	  if (((END_LIST != tids[indx]) && (tids[indx] == tidUsed)) || (TRUE == FIRSTROW))
	    {
	      assert(NULL != _relevantTIDLists[indx]);
	      tids[indx] = _relevantTIDLists[indx]->ReadNextTID();
	    }
	  if ((END_LIST != tids[indx]) && (temp > tids[indx]))
	    {
	      temp = tids[indx];
	      numRead++;
	    }
	}
    }
  assert((temp >= tidUsed) || (0 == numRead));
  tidUsed = temp;
  return numRead;
}
#endif


int TIDListColln::_ReadNextRow(TIDType *tids, TIDType& tidUsed, int *tidListNums, 
				      int numRelTidLists)
{
  int i;
  int numRead = 0;
  TIDType temp = MAX_TID;
  int indx;
  
#if 0
  for (i = 0; i < numRelTidLists; i++)
    cout << tidListNums[i] << " ";
  cout << endl;     
#endif
  
  for (i = 0; i < numRelTidLists; i++)
    {
      indx = tidListNums[i];
	  if ((tids[indx] == tidUsed) && (END_LIST != tids[indx]))
	    {
	      tids[indx] = _tidLists[indx].ReadNextTID();
	    }
	  if ((tids[indx] != END_LIST) && (temp > tids[indx]))
	    {
	      temp = tids[indx];
	      numRead++;
	    }
    }
  assert((temp > tidUsed) || (0 == numRead));
  tidUsed = temp;
  return numRead;
}

void TIDListColln::_CountOneRow(TIDType* tids, Set **itemsets, int *countArray,
				int numItemsets, TIDType tidUsed, 
				TwoDArray *indxToTIDLists)
{
  int i, j;
  int sz, indx;
  bool supportedFlag;
  //  ofstream ofile("/tmp/debugging", ios::app);
  
  assert(END_LIST != tidUsed);
  
  // split into two large blocks to avoid repeatedly
  // checking if indxToTIDLists is NULL;
  if (NULL == indxToTIDLists)
    {// non-fancy version
      for (i = 0; i < numItemsets; i++)
	{
	  supportedFlag = TRUE;
	  //ofile << itemsets[i]->size() << " : " << tidUsed << " : ";
	  sz = itemsets[i]->size();
	  for (j = 0; j < sz; j++)
	    {
	      assert(tids[itemsets[i]->elements(j)] != -1);
	      //if ((tids[itemsets[i].elements(j)] != tidUsed) || (tids[itemsets[i].elements(j)] == END_LIST))
	      indx = itemsets[i]->elements(j);
	      
	      if (tids[indx] != tidUsed)
		{
		  supportedFlag = FALSE;
		  break;
		}
	    }
	  if (TRUE == supportedFlag)
	    countArray[i]++;	  
	  //ofile << " : " << countArray[i] << endl;
	  
	}
    }
  else
    {
      for (i = 0; i < numItemsets; i++)
	{
	  supportedFlag = TRUE;
	  //ofile << itemsets[i]->size() << " : " << tidUsed << " : ";
	  sz = indxToTIDLists->get(i,0);
	  
	  for (j = 0; j < sz; j++)
	    {
	      //ofile << tids[itemsets[i]->elements(j)] << " ";
	      //assert(tids[itemsets[i]->elements(j)] != -1);
	      //if ((tids[itemsets[i].elements(j)] != tidUsed) || (tids[itemsets[i].elements(j)] == END_LIST))
	      indx = indxToTIDLists->get(i,j+1);
	      if (tids[indx] != tidUsed)
		{
		  supportedFlag = FALSE;
		  break;
		}
	    }
	  if (TRUE == supportedFlag)
	    countArray[i]++;	  
	  //ofile << " : " << countArray[i] << endl;
	  
	}
    }
  //  ofile.close();
}


#ifdef FANCYTID
void TIDListColln::_CountFancySingle(Set **itemsets,int *countArray,bool *countFlags,
				    int numItemsets, int totNumItems, 
				    int numMaxTIDLists,
				    TIDType *tids, TwoDArray *indxToTIDLists)
{
  int i;
  int numRead;
  bool flag = TRUE;
  TIDType tidUsed = 0;

  if (0 == numItemsets)
    return;

  //int numMaxTIDLists = 0;
  //TIDList **relevantTIDLists; 
  //int *tids;
  //TwoDArray *indxToTIDLists;

  // find the max number of 2-itemset pairs;

  //for (i = 0; i < numItemsets; i++)
  //maxItSz = (maxItSz < itemsets[i]->size()) ? itemsets[i]->size() : maxItSz;
  //indxToTIDLists = new TwoDArray(numItemsets, maxItSz+1);

  _OpenRelTIDLists(itemsets, numItemsets, countFlags,
		   totNumItems, *indxToTIDLists);
  _ReadNextRow_Fancy(tids, tidUsed, *indxToTIDLists, numItemsets, TRUE);
  
  assert(tidUsed >= 0);
  while ((TRUE == flag) && (MAX_TID != tidUsed))
    {
      _CountOneRow(tids, itemsets, countArray, numItemsets, 
		   tidUsed, indxToTIDLists);
      numRead = _ReadNextRow_Fancy(tids, tidUsed, *indxToTIDLists, numItemsets);
      if (0 == numRead)
	flag = FALSE;
    }
  
#ifdef DEBUG
  cout << "TIDListColln:CountFancy :: after counting " << endl;
  for (i = 0; i < numItemsets; i++)
    cout << *itemsets[i] << " " << countArray[i] << endl;
  cout << endl;
#endif  
  
  for (i = 0; i < totNumItems; i++) // do not want to delete 1-item TID-lists here;
    {
      if (NULL != _relevantTIDLists[i])
	{
	  _relevantTIDLists[i]->Close();
	  _relevantTIDLists[i] = NULL; 
	}
    }
  
  //for (i = 0; i < numMaxTIDLists; i++)
  //if (NULL != relevantTIDLists[i]) delete relevantTIDLists[i];
  
  int j, sz, indx;
  for (i = 0; i < numItemsets; i++)
    {
      sz = indxToTIDLists->get(i,0);
      for (j = 0; j < sz; j++)
	{
	  indx = indxToTIDLists->get(i,j+1);
	  if (NULL != _relevantTIDLists[indx])
	    _relevantTIDLists[indx]->Close();
	}
    }
  
  //delete indxToTIDLists;
  
}
#endif FANCYTID

#endif





