/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#include <assert.h>
#include <iostream.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils.h"

/*
 * fd: file descriptor of the open file
 * currentOffset: offset in *chars* into the current file from which point the reading should start
 * bufSize: the size of the block to be read in *chars*
 * buffer: space for at least bufSize chars
 *
 * return value:
 * size of the block in *chars*
 */

int
getNextBlock(int fd, int currentOffset, int bufSize, char* buffer)
{
  int bytesRead, currentPos, endPos;
  char bufferHere[extraBufferSize];
  int* checkElement;
  bool done;



  assert((0 == (bufSize % sizeof(int))) &&
	 (0 == (extraBufferSize % sizeof(int))));

  done = true; // indicates whether the end of the file is reached


  if (0 == buffer) {
    // seek until a little bit before the end of the buffer
    endPos = lseek(fd, 0, SEEK_END);
    if (currentOffset + bufSize > endPos)
      bufSize = endPos - currentOffset;
    assert(bufSize >= 0);
    currentPos = lseek(fd, currentOffset+bufSize-extraBufferSize, SEEK_SET);
    assert(currentPos <= endPos - extraBufferSize);
    bytesRead = read(fd, (char*)bufferHere, extraBufferSize); // then read the last piece
    if (bytesRead < extraBufferSize) // if the end of the file is reached
      return bufSize - extraBufferSize + bytesRead;
    
    
    checkElement = (int*)&bufferHere[extraBufferSize-sizeof(int)];
  }
  else {
    lseek(fd, currentOffset, SEEK_SET);     // seek for the current offset
    bytesRead = read(fd, (char*)buffer, bufSize); // read the whole block into memory
    if (bytesRead < bufSize) { // end of file?
      return bytesRead;
    }

    checkElement = (int*)&buffer[bufSize-sizeof(int)];
  }

  while (*checkElement != -1) {
    bufSize -= sizeof(int);
    checkElement--;
  }
  assert(bufSize > 0);
  return bufSize;

}
