/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef __HASHTABLE_H__
#define __HASHTABLE_H__

#include <assert.h>

#include "globals.h"
#include "Set.h"
#include "tid-list.h"

extern int hash(Set& itemset, int maxSz);
extern void formMaxItFileName(char* fileName, char *prefix, Set& it);

struct CountSet
{
  Set *itemset;
  int count;

  CountSet()
    {
      itemset = NULL;
      count = 0;
    }

  ~CountSet()
    {
      delete itemset;
    }

  void Init(Set& it, int ct)
    {
      itemset = new Set(it);
      count = ct;
    }

  Matches(Set& it)
    {
      int sz1 = it.size();
      int sz2 = itemset->size();
      int i;

      if (sz1 != sz2)
	return FALSE;
      
      for (i = 0; i < sz1; i++)
	if (itemset->elements(i) != it.elements(i))
	  return FALSE;
      return TRUE;
    }
};


struct HTEntry
{
  void *contents;
  HTEntry *next;

  HTEntry(HTType flag)
    {
      if (TIDLIST == flag)
	contents = new TIDList();
      else if (COUNTSET == flag)
	contents = new CountSet();
    }
  ~HTEntry(){};
};



class HashTable
{
 private:
  HTType _flag;
  int _size;
  char *_prefixName;

  HTEntry **_table;

      
 public:
  HashTable(int sz, HTType flag, char *pfName=NULL)
    {
      _size = (int)(HT_OVHD * sz);
      _flag = flag;
      if (pfName != NULL)
	{
	  _prefixName = new char[MAX_NAME_LENGTH];
	  strcpy(_prefixName, pfName);
	}
      _table = new (HTEntry*)[_size];
      int i;
      for (i = 0; i < _size; i++)
	_table[i] = NULL;
    }

  ~HashTable()
    {
      int i;
      HTEntry *temp1, *temp2;
      
      if (_prefixName != NULL)
	delete [] _prefixName;

      for (i = 0; i < _size; i++)
	{
	  temp1 = _table[i];
	  while (temp1 != NULL)
	    {
	      temp2 = temp1->next;
	      if (COUNTSET == _flag)
		delete (CountSet*)(temp1->contents);
	      // else : they will be deleted earlier.
	      delete temp1;
	      temp1 = temp2;
	    }
	}
    }

  Size() { return _size; }

  // if there is none, a new entry is created.
  TIDList* SearchToWrite(Set &itemset, int count)
    {
      HTEntry *temp = NULL, *temp1 = NULL;
      char fileName[MAX_NAME_LENGTH];
      int hashKey = hash(itemset, _size);
      cout << itemset << " : hashKey " << hashKey << endl;

      formMaxItFileName(fileName, _prefixName, itemset);
      if (_flag == TIDLIST)
	assert(count == -1);

      if (NULL != _table[hashKey])
	{ 
	  temp = _table[hashKey];
	  if (TIDLIST == _flag)
	    while (temp != NULL)
	      {
		if (((TIDList*)(temp->contents))->Matches(itemset, _prefixName))
		  break;
		temp = temp->next;
	      }
	  else if (COUNTSET == _flag)
	    while (temp != NULL)
	      {
		if (((CountSet*)(temp->contents))->Matches(itemset))
		  break;
		temp = temp->next;
	      }
	}

      //if (_flag == COUNTSET)
      assert(NULL == temp);  // can do the same for TIDLIST after PrefixNode struct;

      if (NULL == temp) // did not succeed in the above if stmt;
	{
	  temp = new HTEntry(_flag);
	  assert(temp);
	  if (TIDLIST == _flag)
	    ((TIDList*)(temp->contents))->Init(fileName, PAGE_SIZE, TRUE);
	  else
	    ((CountSet*)(temp->contents))->Init(itemset, count);

	  // insert temp; temp1 may be NULL;
	  temp1 = _table[hashKey];
	  _table[hashKey] = temp;
	  temp->next = temp1;
	}

      assert(temp);

      if (_flag == TIDLIST)
	return (TIDList*)(temp->contents); 
      else 
	return NULL;
    }

  // if there is no entry, it returns NULL;
  HTEntry* SearchToRead(Set &itemset)
    {
      HTEntry *temp = NULL;
      int hashKey = hash(itemset, _size);
      
      if (NULL != _table[hashKey])
	{ 
	  temp = _table[hashKey];
	  if (TIDLIST == _flag)
	    while (temp != NULL)
	      {
		if (((TIDList*)(temp->contents))->Matches(itemset, _prefixName))
		  break;
		temp = temp->next;
	      }
	  else if (COUNTSET == _flag)
	    while (temp != NULL)
	      {
		if (((CountSet*)(temp->contents))->Matches(itemset))
		  break;
		temp = temp->next;
	      }
	}
      return temp;
    }
  
  HTEntry*& operator[](int key) 
    { 
      return _table[key];
    }
};

#endif












