/*
 * Copyright (c) 1998-1999
 * Venkatesh Ganti
 * University of Wisconsin -- Madison
 * All Rights Reserved.

 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.

 * THE DEMON TEAM AT THE UNIVERSITY OF WISCONSIN-MADISON ALLOWS FREE
 * USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION. WE DISCLAIM ANY
 * LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
 * THE USE OF THIS SOFTWARE. 
 * VENKATESH GANTI, JOHANNES GEHRKE, AND RAGHU RAMAKRISHNAN COMPRISE
 * THE DEMON TEAM
 */



#ifndef MACROS_H
#define MACROS_H

#include <math.h>

typedef int Item;

#ifdef HASH_TREE
class HashTree;
typedef HashTree Tree;
#else
class PrefixTree;
typedef PrefixTree Tree;
#endif

// Used by the Set class.  This is used for growing the dynamic array.
#define NUM_ITEM_STEP 10 // Assuming sets generally won't be larger than
			 // about 20 to 30

// Default hashing parameters for the hash
// function defined in macro HASH() below.
#define DEF_HASH_H        20
#define DEF_HASH_A         1
#define DEF_HASH_B         0

// The hash function macro
// #define HASH(a, b, h, item)  (((a) * (item) + (b)) % (h))
#define HASH(a, b, h, item)  ((item) % (h))

// The default number of itemsets in a leaf of a hash tree,
// or siblings in a prefix node array of a prefix tree.
#define DEF_SETS_PER_LEAF 20

// do not change. otherwise check entire PrefixTree
#define FIRST_TRANSACTION 0

#define NUM_TRANSACTIONS 29

// Maximum number of items in a transaction
#define MAX_TRANS_SIZE 1000

// number of transactions between each stop
#define NUM_BETWEEN_STOPS 5

// The maximum size of the character string containing a transaction.
// Given the maximum item maxItem, log10(maxItem) + 1 gives us the max
// number of bytes per item.  log10(maxItem) + 2 counts in a blank space
// between every pair of items.  MAX_TRANS_SIZE * (log10(maxItem) + 2)
// then gives us the number of bytes necessary for MAX_TRANS_SIZE items
// separated by one blank space.  The 1 added at the end is for \n.
#define MAX_TRANS_BYTES(maxItem) (MAX_TRANS_SIZE * ((int) log10(maxItem) + 2) + 1)

#endif 
